using System;
using System.Collections.Generic;
using System.Text;
using System.Net.Sockets;
using System.Net;
using System.IO;
using System.Collections;
using System.Text.RegularExpressions;
using System.Diagnostics;
using System.Threading;

namespace YPortMgr
{
	class YProcessAddress
	{
		public string m_strHostName;
		public int m_nPort;
		public YProcessAddress( string strHost, int nPort )
		{
			m_strHostName = strHost;
			m_nPort = nPort;
		}
	}
	class ProcPortMapping
	{
		// mapping between process and port
		private static Hashtable m_hProcPort = new Hashtable();	// process number and port
		private static object m_hProcPortLock = new object();	// lock for hash table
		public ProcPortMapping()
		{
		}
		public static int GetNextProcId( int nPort, string strHost )
		{
			int nextId = -1;
			lock( m_hProcPortLock ) {
				int maxId = -1;
				foreach( int id in m_hProcPort.Keys ) {
					if( id > maxId ) maxId = id;
				}
				nextId = maxId + 1;
				m_hProcPort.Add( nextId, new YProcessAddress( strHost, nPort ) );
			}
			return nextId;
		}
		public static bool AddProcId( int nProcId, string strHost, int nPort )
		{
			lock( m_hProcPortLock ) {
				if( m_hProcPort.ContainsKey( nProcId ) ) {
					return false;
				}
				m_hProcPort.Add( nProcId, new YProcessAddress( strHost, nPort ) );
				return true;
			}
		}
		public static void DeleteProcEntry( int nProc )
		{
			lock( m_hProcPortLock ) {
				if( m_hProcPort.ContainsKey( nProc ) ) m_hProcPort.Remove( nProc );
			}
		}
		public static bool UpdateProcEntry( int nProc, int nPort, string strHost )
		{
			bool retVal = false;
			lock( m_hProcPortLock ) {
				if( m_hProcPort.ContainsKey( nProc ) ) {
					m_hProcPort[nProc] = new YProcessAddress( strHost, nPort );
					retVal = true;
				}
			}
			return retVal;
		}
		public static string MapToString()
		{
			string strRet = "";
			strRet += " " + m_hProcPort.Count;
			lock( m_hProcPortLock ) {
				foreach( int nProcId in m_hProcPort.Keys ) {
					YProcessAddress yProcAddr = (YProcessAddress)m_hProcPort[nProcId];
					strRet += " " + nProcId;
					strRet += " " + yProcAddr.m_strHostName;
					strRet += " " + yProcAddr.m_nPort;
				}
			}
			return strRet;
		}
		public static void AnswerClient( object param  )
		{
			try {
				TcpClient tClient = (TcpClient)param;
				NetworkStream ns = tClient.GetStream();
				StreamReader sr = new StreamReader( ns );
				StreamWriter sw = new StreamWriter( ns );
				string strRequest, strRet = "";
				int nPort, nProc;
				//Regex regEx = new Regex( @"^YaccaiPort\s+[0-9]+\s+[0-9]+\s*[0-9]*" );
				while( true ) {

					strRequest = sr.ReadLine();
					Console.WriteLine( strRequest );
					/*
					if( !regEx.IsMatch( strRequest ) ) {
						strRet = "YaccaiPort 0";
						sw.WriteLine( strRet );
						continue;
					}
					 * */
					string[] arrRequest = Regex.Split( strRequest, @"\s+" );

					if( arrRequest.Length < 2 ) {
						strRet = "YaccaiPort 0";		// invalid format
						sw.WriteLine( strRet );
						continue;
					}

					int nCmd = Int32.Parse( arrRequest[1] );
					switch( nCmd ) {
						case 100:		// query port from proc
							if( arrRequest.Length < 3 ) {
								strRet = "YaccaiPort 1 0";	// invalid argument
							} else {
								nProc = Int32.Parse( arrRequest[2] );
								if( m_hProcPort.ContainsKey( nProc ) ) {
									// success
									YProcessAddress yProcAddr = (YProcessAddress)m_hProcPort[nProc];
									strRet = "YaccaiPort 100 " + yProcAddr.m_nPort + " " + yProcAddr.m_strHostName;
								} else {
									// not found
									strRet = "YaccaiPort 102 0";
								}
							}
							break;
						case 200:		// add proc-port entry (auto-gen nProc)
							if( arrRequest.Length < 4 ) {
								strRet = "YaccaiPort 1";	// invalid argument
							} else {
								nPort = Int32.Parse( arrRequest[2] );
								string strHost = arrRequest[3];
								int retId = -1;
								if( ( retId = GetNextProcId( nPort, strHost ) ) >= 0 ) {
									strRet = "YaccaiPort 200 " + retId;
								} else {
									strRet = "YaccaiPort 103";	// error
								}
							}
							break;
						case 210:
							if( arrRequest.Length < 4 ) {
								strRet = "YaccaiPort 1";	// invalid argument
							} else {
								nProc = Int32.Parse( arrRequest[2] );
								string strHost = arrRequest[3];
								if( arrRequest.Length >= 5 ) nPort = Int32.Parse( arrRequest[4] );
								else nPort = -1;
								Console.WriteLine( "Adding procid = {0}, host = {1}, port = {2}", nProc, strHost, nPort );
								if( AddProcId( nProc, strHost, nPort ) ) {
									strRet = "YaccaiPort 210 " + nProc;
								} else strRet = "YaccaiPort 211";		// already existed
							}
							break;
						case 250:		// update proc-port entry
							if( arrRequest.Length < 5 ) {
								strRet = "YaccaiPort 1";	// invalid argument
							} else {
								nProc = Int32.Parse( arrRequest[2] );
								nPort = Int32.Parse( arrRequest[3] );
								string strHost = arrRequest[4];
								if( UpdateProcEntry( nProc, nPort, strHost ) ) {
									strRet = "YaccaiPort 250";
								} else {
									strRet = "YaccaiPort 103";	// error
								}
							}
							break;
						case 300:		// delete proc entry
							if( arrRequest.Length < 3 ) {
								strRet = "YaccaiPort 1";	// invalid argument
							} else {
								nProc = Int32.Parse( arrRequest[2] );
								DeleteProcEntry( nProc );
								strRet = "YaccaiPort 300";
							}
							break;
						case 400:		// get all mappings
							strRet = "YaccaiPort 400 " + MapToString();
							break;
						case 10:
							// terminate
							tClient.Close();
							return;
						case 1000:
							// end process
							tClient.Close();
							Environment.Exit( 0 );
							break;
						default:
							strRet = "YaccaiPort 2 0";	// invalid command
							break;
					}
					sw.WriteLine( strRet );
					sw.Flush();
					Console.WriteLine( strRet );
				}
			} catch( Exception exc ) {
				Console.WriteLine( exc.Message );
			}
		}
	}
	class PortMgrMain
	{
		public static bool m_bContinue = true;
		static void ParseCmdLine( string[] argInput, Hashtable hOptions )
		{
			hOptions.Clear();
			int i = 0;
			if( argInput.Length <= 0 ) return;
			while( argInput[i].StartsWith( "-" ) ) {
				string strOption = argInput[i];
				i++;
				if( !hOptions.ContainsKey( strOption ) ) {
					hOptions.Add( strOption, null );
				} else hOptions[strOption] = null;
				if( i >= argInput.Length ) {
					break;
				}
				string strOptionVal = argInput[i];
				if( !strOptionVal.StartsWith( "-" ) ) {
					hOptions[strOption] = strOptionVal;
					i++;
				}
			}
		}

		static void Main( string[] args )
		{
			Hashtable hOptions = new Hashtable();
			ParseCmdLine( args, hOptions );
			int nPort = 31572;
			if( hOptions.ContainsKey( "-p" ) ) {
				nPort = (int)hOptions["-p"];
			}
			TcpListener tcpListener = new TcpListener( IPAddress.Any, nPort );
			try {
				tcpListener.Start();
				while( true ) {
					TcpClient tClient = tcpListener.AcceptTcpClient();
					Thread th = new Thread( new ParameterizedThreadStart( ProcPortMapping.AnswerClient ) );
					th.Start( tClient );
				}
			} catch( SocketException se ) {
				if( se.SocketErrorCode == SocketError.AddressAlreadyInUse ) {
					Environment.Exit( 0 );
				} else {
					Console.WriteLine( se.Message );
					Environment.Exit( 1 );
				}
			}

		}
	}
}
