using System;
using System.Collections.Generic;
using System.Collections;
namespace Yaccai
{
	enum ExprType
	{
		LiteralInt, LiteralDouble, LiteralString, LiteralChar,
		Var, Operator, Subscript, AttributeRef, FuncCall,
		Assignment, ObjectAllocation, BeliefExpr, Create, RegexExpr,
		StopEvalExpr, Expression
	}

	enum OperatorFixKind
	{
		InFix, PreFix, PostFix
	}

	enum OperatorType
	{
		PLUS, MINUS, MUL, DIV, MOD, UPLUS, UMINUS, LT, GT, LT_EQ, GT_EQ, 
		EQ_EQ, DIFF, NOT, AND, OR, PLUS_EQ, MINUS_EQ, MUL_EQ, DIV_EQ, MOD_EQ,
		GT_GT, LT_LT, GT_GT_EQ, LT_LT_EQ, BITWISE_OR, BITWISE_AND, BITWISE_XOR,
		BITWISE_NOT, BITWISE_OR_EQ, BITWISE_AND_EQ, BITWISE_XOR_EQ, EQ
	}

	enum YaccaiAccessType
	{
		Public, Private, Protected
	}

	enum BeliefExprType
	{
		Query, Join, Remove, AddFact
	}
	enum StmtKind
	{
		Expression, Return, If, While, For, Break, Continue,
		MethodDeclaration, ClassDeclaration, AgentClassDeclaration, Include,
		PlanDeclaration, DoWhile, Switch, ForEach, LabelDeclaration, Nop, Block,
		PlanPar, PlanAnd, PlanOr, MemberVarDeclaration, Match
	}
	enum RegexOpType
	{
		Star, Question, Plus, None
	}

	class YaccaiAccessTypeWrapper
	{
		public YaccaiAccessType m_accessType;
		public YaccaiAccessTypeWrapper( YaccaiAccessType y )
		{
			m_accessType = y;
		}
	}
	class OperatorTypeWrapper
	{
		public OperatorType m_type;
		public OperatorTypeWrapper( OperatorType t )
		{
			m_type = t;
		}
	}
	class LinkedList<T>
	{
		private ArrayList m_arrayList;
		public int Count
		{
			get
			{
				return m_arrayList.Count;
			}
		}
		public LinkedList()
		{
			m_arrayList = new ArrayList();
		}
		public void AddLast( T obj )
		{
			m_arrayList.Add( obj );
		}
		public void AddFirst( T obj )
		{
			m_arrayList.Insert( 0, obj );
		}
		public void Add( T obj )
		{
			m_arrayList.Add( obj );
		}
		public IEnumerator GetEnumerator()
		{
			return m_arrayList.GetEnumerator();
		}
		public void Clear()
		{
			m_arrayList.Clear();
		}
		public T this[int index]
		{
			get
			{
				return (T)m_arrayList[index];
			}
			set
			{
				m_arrayList[index] = (T)value;
			}
		}
		public void Sort()
		{
			m_arrayList.Sort();
		}
		public void RemoveFirst()
		{
			if( m_arrayList.Count < 1 ) return;
			m_arrayList.RemoveAt( 0 );
		}
		public T First
		{
			get
			{
				if( m_arrayList.Count < 1 ) {
					throw new Exception( "Array index out of range" );
				}
				return (T)m_arrayList[0];
			}
		}
	}
	class SrcPos
	{
		public string m_strFileName;
		public int m_nLine;
		public int m_nCol;

		public SrcPos( string fn, int nLine, int nCol )
		{
			m_strFileName = fn;
			m_nLine = nLine;
			m_nCol = nCol;
		}
		public override string ToString()
		{
			string strRet = m_strFileName;
			strRet += " : (" + m_nLine + ", " + m_nCol + ")";
			return strRet;
		}
	}
	class MemberVarExprPair
	{
		public string m_strName;
		public Expr m_expr;
		public MemberVarExprPair( string n, Expr val )
		{
			m_strName = n;
			m_expr = val;
		}
	}
	class Expr
	{

		private ExprType m_type;
		private Object m_contents;
		private SrcPos m_pos;

		public Expr( int n, SrcPos p )
		{
			m_contents = n;
			m_type = ExprType.LiteralInt;
			m_pos = p;
		}

		public Expr( double d, SrcPos p )
		{
			m_contents = d;
			m_type = ExprType.LiteralDouble;
			m_pos = p;
		}

		public Expr( char ch, SrcPos p )
		{
			m_contents = ch;
			m_type = ExprType.LiteralChar;
			m_pos = p;
		}

		public Expr( ExprType type, String str, SrcPos p )
		{
			// Var, LiteralString
			m_type = type;
			m_contents = str;
			m_pos = p;
		}

		public Expr( ExprType type, Expr exp, SrcPos p )
		{
			m_pos = p;
			m_contents = exp;
			m_type = type;
		}
		public Expr( ExprOperator e, SrcPos p )
		{
			m_contents = e;
			m_type = ExprType.Operator;
			m_pos = p;
		}
		public Expr( ExprSubscript e, SrcPos p )
		{
			m_contents = e;
			m_type = ExprType.Subscript;
			m_pos = p;
		}

		public Expr( ExprAttref e, SrcPos p )
		{
			m_contents = e;
			m_type = ExprType.AttributeRef;
			m_pos = p;
		}

		public Expr( ExprFuncCall e, SrcPos p )
		{
			m_contents = e;
			m_type = ExprType.FuncCall;
			m_pos = p;
		}

		public Expr( ExprAssignment a, SrcPos p )
		{
			m_contents = a;
			m_type = ExprType.Assignment;
			m_pos = p;
		}

		public Expr( ExprObjectAllocation oa, SrcPos p )
		{
			m_contents = oa;
			m_type = ExprType.ObjectAllocation;
			m_pos = p;
		}

		public Expr( ExprBelief b, SrcPos p )
		{
			m_contents = b;
			m_type = ExprType.BeliefExpr;
			m_pos = p;
		}
		public Expr( ExprCreate c, SrcPos p )
		{
			m_contents = c;
			m_type = ExprType.Create;
			m_pos = p;
		}
		public Expr( ExprRegexExpr regexExp, SrcPos p )
		{
			m_contents = regexExp;
			m_type = ExprType.RegexExpr;
			m_pos = p;
		}
		public ExprType Type
		{
			get
			{
				return m_type;
			}
		}

		public Object Contents
		{
			get
			{
				return m_contents;
			}
		}

		public SrcPos Pos
		{
			get
			{
				return m_pos;
			}
			set
			{
				m_pos = value;
			}
		}

		public override string ToString()
		{
			switch( m_type ) {
				case ExprType.LiteralChar:
					return ( (char)m_contents ).ToString();
				case ExprType.LiteralDouble:
					return ( (double)m_contents ).ToString();
				case ExprType.LiteralInt:
					return ( (int)m_contents ).ToString();
				case ExprType.LiteralString:
					return ( (string)m_contents );
				case ExprType.Var:
					return (string)m_contents;
				case ExprType.AttributeRef:
					return ( (ExprAttref)m_contents ).ToString();
				case ExprType.Operator:
					return ( (ExprOperator)m_contents ).ToString();
				case ExprType.Subscript:
					return ( (ExprSubscript)m_contents ).ToString();
				case ExprType.FuncCall:
					return ( (ExprFuncCall)m_contents ).ToString();
				case ExprType.Assignment:
					return ( (ExprAssignment)m_contents ).ToString();
				case ExprType.BeliefExpr:
					return ( (ExprBelief)m_contents ).ToString();
				case ExprType.Create:
					return ( (ExprCreate)m_contents ).ToString();
				case ExprType.ObjectAllocation:
					return ( (ExprObjectAllocation)m_contents ).ToString();
				case ExprType.Expression:
					return "(" + ( (Expr)m_contents ).ToString() + ")";
				case ExprType.StopEvalExpr:
					return "@{" + ( (Expr)m_contents ).ToString() + "}";
				case ExprType.RegexExpr:
					return ( (ExprRegexExpr)m_contents ).ToString();
				default:
					return "Unknown expr type: type = " + m_type + ", contents = " + m_contents;
			}
		}

	}

	class ExprRegexExpr
	{
		public Expr m_expr;
		public RegexOpType m_opType;
		public ExprRegexExpr( string strType, Expr exp )
		{
			m_expr = exp;
			switch( strType ) {
				case "+":
					m_opType = RegexOpType.Plus;
					break;
				case "*":
					m_opType = RegexOpType.Star;
					break;
				case "?":
					m_opType = RegexOpType.Question;
					break;
				default:
					m_opType = RegexOpType.None;
					break;
			}
		}
		public override string ToString()
		{
			string strRet = "";
			if( m_expr != null ) {
				strRet = m_expr.ToString();
			}
			switch( m_opType ) {
				case RegexOpType.Plus:
					strRet += "+";
					break;
				case RegexOpType.Question:
					strRet += "?";
					break;
				case RegexOpType.Star:
					strRet += "*";
					break;
				default:
					break;
			}
			return strRet;
		}
	}
	class ExprSubscript
	{
		public Expr m_base;
		public Expr m_index;
		public ExprSubscript( Expr e1, Expr e2 )
		{
			m_base = e1;
			m_index = e2;
		}
		public override String ToString()
		{
			return m_base.ToString() + "[" + m_index.ToString() + "]";
		}
	}

	class ExprAttref
	{
		public Expr m_base;
		public Expr m_field;
		public ExprAttref( Expr b, Expr f )
		{
			m_base = b;
			m_field = f;
		}
		public override String ToString()
		{
			return m_base.ToString() + "." + m_field.ToString();
		}
	}

	class ExprOperator
	{
		public OperatorFixKind m_fix;
		public OperatorType m_type;
		public ArrayList m_args;

		public ExprOperator( OperatorFixKind f, OperatorType t, ArrayList args )
		{
			m_fix = f;
			m_type = t;
			m_args = args;
		}
		public override String ToString()
		{
			Expr e1, e2;
			String strOp;
			switch( m_fix ) {
				case OperatorFixKind.InFix:
					e1 = (Expr)m_args[0];
					e2 = (Expr)m_args[1];
					switch( m_type ) {
						case OperatorType.BITWISE_AND:
							strOp = " & ";
							break;
						case OperatorType.BITWISE_NOT:
							strOp = " ~ ";
							break;
						case OperatorType.BITWISE_OR:
							strOp = " | ";
							break;
						case OperatorType.BITWISE_XOR:
							strOp = " ^ ";
							break;
						
						case OperatorType.AND:
							strOp = " && ";
							break;
						case OperatorType.OR:
							strOp = " || ";
							break;
						case OperatorType.LT:
							strOp = " < ";
							break;
						case OperatorType.GT:
							strOp = " > ";
							break;
						case OperatorType.GT_EQ:
							strOp = " >= ";
							break;
						case OperatorType.LT_EQ:
							strOp = " <= ";
							break;
						case OperatorType.EQ_EQ:
							strOp = " == ";
							break;
						case OperatorType.PLUS:
							strOp = " + ";
							break;
						case OperatorType.MINUS:
							strOp = " - ";
							break;
						case OperatorType.MUL:
							strOp = " * ";
							break;
						case OperatorType.MOD:
							strOp = " % ";
							break;
						case OperatorType.DIV:
							strOp = " / ";
							break;
						case OperatorType.DIFF:
							strOp = " != ";
							break;
						default:
							strOp = "Unknown operator";
							break;
					}
					return e1.ToString() + strOp + e2.ToString();
				case OperatorFixKind.PreFix:
					e1 = (Expr)m_args[0];
					switch( m_type ) {
						case OperatorType.UMINUS:
							strOp = "-";
							break;
						case OperatorType.UPLUS:
							strOp = "";
							break;
						case OperatorType.NOT:
							strOp = "!";
							break;
						default:
							strOp = "Unknown prefix op type";
							break;
					}
					return strOp + e1.ToString();
			}
			return "Unknown operator";
		}
	}

	class ExprAssignment
	{
		public LinkedList<Expr> m_primaryList;
		public LinkedList<OperatorType> m_opList;
		public Expr m_orTestExpr;
		
		public ExprAssignment( LinkedList<Expr> primaryList,
								LinkedList<OperatorType> opList,
								Expr orTest )
		{
			m_primaryList = primaryList;
			m_opList = opList;
			m_orTestExpr = orTest;
		}

		public override string ToString()
		{
			string strRet = "";
			if( m_primaryList != null ) {
				if( m_opList == null || m_primaryList.Count != m_opList.Count ) {
					Console.WriteLine( "Bug: invalid assignment expression" );
					Environment.Exit( 1 );
				}
				int i = 0;
				foreach( Expr ex1 in m_primaryList ) {
					strRet += ex1.ToString();
					strRet += PrintUtil.OpType2Str( m_opList[i] );
					i++;
				}
				strRet += m_orTestExpr.ToString();
			}
			return strRet;
		}

	}

	class ExprObjectAllocation
	{
		public string m_strClassName;
		public LinkedList<string> m_memberList;
		public LinkedList<Expr> m_exprList;
		public LinkedList<Expr> m_bodyLst;
		public ExprObjectAllocation( string strClassName, LinkedList<Expr> exprList, LinkedList<Expr> bodyLst )
		{
			// new constructor( expr, expr, ... ) { expr, expr, ...}?;
			m_strClassName = strClassName;
			m_exprList = exprList;
			m_memberList = null;
			m_bodyLst = bodyLst;
		}
		public ExprObjectAllocation( LinkedList<string> memberList,
										LinkedList<Expr> exprList )
		{
			// anonymous class: a = new { x = 500, y = 600 };
			m_memberList = memberList;
			m_exprList = exprList;
			m_strClassName = null;
		}
		/*
		public ExprObjectAllocation( string strClassName,
										LinkedList<string> memberList,
										LinkedList<Expr> exprList )
		{
			// new classname { member initialization }
			m_strClassName = strClassName;
			m_memberList = memberList;
			m_exprList = exprList;
		}
		 * */
		public override string ToString()
		{
			string strRet = "new ";
			if( m_strClassName != null ) {
				strRet += m_strClassName;
			}
			int i = 0;
			if( m_memberList != null ) {
				strRet += "{";
				if( m_exprList.Count > 0 && m_memberList.Count > 0 ) {
					strRet += m_memberList[i] + " = " + m_exprList[i].ToString();
					for( i = 1; i < m_memberList.Count && i < m_exprList.Count; i++ ) {
						strRet += ", " + m_memberList[i] + " = " + m_exprList[i].ToString();
					}
				}
				strRet += "}";
			} else {
				strRet += "(";
				strRet += PrintUtil.List2Str<Expr>( m_exprList );
				strRet += ")";
				if( m_bodyLst != null ) {
					strRet += " {";
					strRet += PrintUtil.List2Str<Expr>( m_bodyLst );
					strRet += "}";
				}
			}
			return strRet;
		}
	}
	class YaccaiClassSignature
	{
		public string m_strClassName;
		public LinkedList<string> m_memberList;
		public YaccaiClassSignature( string className )
		{
			m_strClassName = className;
			m_memberList = null;
		}
		public YaccaiClassSignature( LinkedList<string> memberList )
		{
			m_memberList = memberList;
			m_strClassName = null;
			memberList.Sort();
		}
		public YaccaiClassSignature( string className, LinkedList<string> memberList )
		{
			m_memberList = memberList;
			m_strClassName = className;
			m_memberList.Sort();
		}
		public override bool Equals( object obj )
		{
			if( obj == null ) return false;
			YaccaiClassSignature other = (YaccaiClassSignature)obj;
			if( other == null ) return false;
			bool nameE = false;
			if( this.m_strClassName == null ) {
				if( other.m_strClassName == null ) nameE = true;
				else nameE = false;
			} else {
				if( other.m_strClassName != null ) {
					if( other.m_strClassName.Equals( this.m_strClassName ) ) nameE = true;
				}
			}
			if( !nameE ) return false;
			if( this.m_memberList == null ) {
				if( other.m_memberList == null ) return true;
				return false;
			} else {
				if( other.m_memberList == null ) return false;
				if( this.m_memberList.Count != other.m_memberList.Count ) return false;
				int i = 0;
				string str1, str2;
				for( i = 0; i < this.m_memberList.Count; i++ ) {
					str1 = m_memberList[i];
					str2 = other.m_memberList[i];
					if( !str1.Equals( str2 ) ) return false;
				}
			}
			return true;			
		}

		public bool Equals( YaccaiClassSignature other )
		{
			if( other == null ) return false;
			bool nameE = false;
			if( this.m_strClassName == null ) {
				if( other.m_strClassName == null ) nameE = true;
				else nameE = false;
			} else {
				if( other.m_strClassName != null ) {
					if( other.m_strClassName.Equals( this.m_strClassName ) ) nameE = true;
				}
			}
			if( !nameE ) return false;
			if( this.m_memberList == null ) {
				if( other.m_memberList == null ) return true;
				return false;
			} else {
				if( other.m_memberList == null ) return false;
				if( this.m_memberList.Count != other.m_memberList.Count ) return false;
				int i = 0;
				string str1, str2;
				for( i = 0; i < this.m_memberList.Count; i++ ) {
					str1 = m_memberList[i];
					str2 = other.m_memberList[i];
					if( !str1.Equals( str2 ) ) return false;
				}
			}
			return true;
		}
		public override int GetHashCode()
		{
			string strRet = "";
			if( m_strClassName != null ) {
				strRet += m_strClassName;
			} else strRet += ";";
			if( m_memberList != null ) {
				foreach( string str in m_memberList ) strRet += str + " ";
			}
			return strRet.GetHashCode();
		}
		public override string ToString()
		{
			string strRet = "";
			if( m_strClassName != null ) strRet += m_strClassName;
			if( m_memberList != null ) {
				strRet += "{";
				int i = 0;
				if( m_memberList.Count > 0 ) {
					strRet += m_memberList[0];
					for( i = 1; i < m_memberList.Count; i++ )
						strRet += ", " + m_memberList[i];
				}
				strRet += "}";
			}
			return strRet;
		}
	}
	class BeliefProjection
	{
		public string m_strObjName;
		public LinkedList<string> m_memberNameList;
		public BeliefProjection( string objName )
		{
			m_strObjName = objName;
			m_memberNameList = null;
		}
		public BeliefProjection( string objName, LinkedList<string> memberList )
		{
			m_strObjName = objName;
			m_memberNameList = memberList;
		}
		public override string ToString()
		{
			string strRet = "";
			strRet += m_strObjName;
			if( m_memberNameList != null ) {
				
				if( m_memberNameList.Count > 0 ) {
					int i = 0;
					strRet += m_memberNameList[0];
					for( i = 1; i < m_memberNameList.Count; i++ ) {
						strRet += ", " + m_memberNameList[i];
					}
				}
			}
			return strRet;
		}
	}

	class ExprBelief
	{
		public BeliefExprType m_type;
		public Expr m_expr;
		public YaccaiClassSignature m_signature;
		public BeliefProjection m_projection;
		public Expr m_filterExpr;
		public Expr m_exprJoin1;	// in join
		public Expr m_exprJoin2;
		public ExprBelief( Expr e )
		{
			m_type = BeliefExprType.AddFact;
			m_expr = e;
		}
		public ExprBelief( BeliefExprType t, YaccaiClassSignature s, BeliefProjection p, Expr filter )
		{
			// for Remove & Query
			m_type = t;
			m_filterExpr = filter;
			m_signature = s;
			m_projection = p;
		}
		public ExprBelief( Expr expr1, Expr expr2,
							BeliefProjection p, Expr filter )
		{
			m_type = BeliefExprType.Join;
			m_exprJoin1 = expr1;
			m_exprJoin2 = expr2;
			m_projection = p;
			m_filterExpr = filter;
		}
		public override string ToString()
		{
			string strRet = "";
			switch( m_type ) {
				case BeliefExprType.AddFact:
					strRet += "belief fact ";
					strRet += m_expr;
					break;
				case BeliefExprType.Remove:
					strRet += "belief remove ";
					strRet += m_signature;
					if( m_projection != null ) {
						strRet += " as " + m_projection;
						if( m_filterExpr != null ) {
							strRet += " where " + m_filterExpr;
						}
					}
					break;
				case BeliefExprType.Query:
					strRet += "belief query ";
					strRet += m_signature;
					if( m_projection != null ) {
						strRet += " as " + m_projection;
						if( m_filterExpr != null ) {
							strRet += " where " + m_filterExpr;
						}
					}
					break;
				case BeliefExprType.Join:
					strRet += "belief join ";
					strRet += m_exprJoin1.ToString() + " join " + m_exprJoin2 + " on " +
								m_expr;
					if( m_projection != null ) {
						strRet += " as " + m_projection;
						if( m_filterExpr != null ) {
							strRet += " where " + m_filterExpr;
						}
					}					
					break;
				default:
					strRet = "Error: unknown belief expr type";
					break;
			}
			return strRet;
		}
	}
	class ExprCreate
	{
		public string m_agentClassName;
		public LinkedList<Expr> m_exprList;
		public Expr m_exprHostname;
		public ExprCreate( string n, LinkedList<Expr> exprList, Expr exprHost )
		{
			m_agentClassName = n;
			m_exprList = exprList;
			m_exprHostname = exprHost;
		}
		public override string ToString()
		{
			string strRet = "new ";
			strRet += m_agentClassName;
			strRet += "(";
			if( m_exprList != null && m_exprList.Count > 0 ) {
				int i = 0;
				strRet += m_exprList[i].ToString();
				for( i = 1; i < m_exprList.Count; i++ ) {
					strRet += ", " + m_exprList[i];
				}
			}
			strRet += ")";
			if( m_exprHostname != null ) {
				strRet += " at " + m_exprHostname.ToString();
			}
			return strRet;
		}
	}

	class ExprFuncCall
	{
		public Expr m_f;
		public ArrayList m_args;
		public ExprFuncCall( Expr f, ArrayList args )
		{
			m_f = f;
			m_args = args;
		}
		public override string ToString()
		{
			String strArgs = "(";
			if( m_args != null ) {
				if( m_args.Count > 0 ) {
					int i = 0;
					strArgs += m_args[0];
					for( i = 1; i < m_args.Count; i++ )
						strArgs += ", " + m_args[i].ToString();
				}
			}
			strArgs += ")";
			return m_f.ToString() + strArgs;
		}
	}
	class YaccaiMatchPair
	{
		public LinkedList<Expr> m_exprList;
		public Stmt m_stmt;
		public YaccaiMatchPair( LinkedList<Expr> expList, Stmt s )
		{
			m_exprList = expList;
			m_stmt = s;
		}
	}
	/*
	 * Trailer
	 * primary ::= atom (trailer)*
	 * trailer ::= '(' expr_list ')' | '[' expr ']' | '.' IDENTIFIER
	 * */
	enum TrailerType { Call, Subscript, AttributeRef };
	class ExprTrailer
	{

		public TrailerType m_type;
		public Object m_contents;
		public SrcPos m_pos;
		public ExprTrailer( TrailerType t, Object c )
		{
			m_type = t;
			m_contents = c;	// t = Call -> c = expr_list, Subscript -> expr, Attref -> IDENTIFIER
			m_pos = null;
		}
		public ExprTrailer( TrailerType t, Object c, SrcPos p )
		{
			m_type = t;
			m_contents = c;
			m_pos = p;
		}

	}

	class MatchStmt
	{
		public Expr m_subjectExpr;
		public LinkedList<YaccaiMatchPair> m_targetList;
		public MatchStmt( Expr subj, LinkedList<YaccaiMatchPair> targetList )
		{
			m_subjectExpr = subj;
			m_targetList = targetList;
		}
		public override string ToString()
		{
			string strRet = "match ";
			if( m_subjectExpr != null ) {
				strRet += m_subjectExpr;
			}
			strRet += " with {\n";
			foreach( YaccaiMatchPair p in m_targetList ) {
				if( p.m_exprList != null && p.m_exprList.Count > 0 ) {
					int i = 0;
					strRet += p.m_exprList[0].ToString();
					for( i = 1; i < p.m_exprList.Count; i++ )
						strRet += ", " + p.m_exprList[i];
				}
				strRet += " -> ";
				strRet += p.m_stmt;
				strRet += "\n";
			}
			strRet += "}\n";
			return strRet;
		}
	}

	class IfStmt
	{
		public Expr m_expr;
		public Stmt m_trueBranch;
		public Stmt m_falseBranch;
		public IfStmt( Expr e, Stmt trueBranch, Stmt falseBranch )
		{
			m_expr = e;
			m_trueBranch = trueBranch;
			m_falseBranch = falseBranch;
		}
		public override string ToString()
		{
			string strRet = "if( ";
			strRet += m_expr;
			strRet += " )";
			strRet += m_trueBranch;
			if( m_falseBranch != null ) {
				strRet += "\nelse ";
				strRet += m_falseBranch;
			}
			return strRet;
		}
	}

	class WhileStmt
	{
		public Expr m_expr;
		public Stmt m_stmt;
		public WhileStmt( Expr e, Stmt s )
		{
			m_expr = e;
			m_stmt = s;
		}
		public override string ToString()
		{
			string strRet = "while( ";
			strRet += m_expr;
			strRet += " )";
			strRet += m_stmt;
			return strRet;
		}
	}

	class PrintUtil
	{
		public static string List2Str<T>( LinkedList<T> inputList )
		{
			string strRet = "";
			if( inputList == null ) return strRet;
			if( inputList.Count > 0 ) {
				int i = 0;
				strRet += inputList[i].ToString();
				for( i = 1; i < inputList.Count; i++ )
					strRet += ", " + inputList[i].ToString();
			}
			return strRet;
		}
		public static string List2Str( ArrayList inputList )
		{
			string strRet = "";
			if( inputList == null ) return strRet;
			if( inputList.Count > 0 ) {
				int i = 0;
				strRet += inputList[i].ToString();
				for( i = 1; i < inputList.Count; i++ )
					strRet += ", " + inputList[i].ToString();
			}
			return strRet;

		}
		public static string AccessType2Str( YaccaiAccessType y )
		{
			switch( y ) {
				case YaccaiAccessType.Private:
					return "private ";
				case YaccaiAccessType.Protected:
					return "protected ";
				case YaccaiAccessType.Public:
					return "public ";
				default:
					return "unknown access type ";
			}
		}
		public static string OpType2Str( OperatorType t )
		{
			string strOp = "";
			switch( t ) {
				case OperatorType.BITWISE_OR_EQ:
					strOp = " |= ";
					break;
				case OperatorType.BITWISE_AND_EQ:
					strOp = " &= ";
					break;
				case OperatorType.BITWISE_XOR_EQ:
					strOp = " ^= ";
					break;
				case OperatorType.DIV_EQ:
					strOp = " /= ";
					break;
				case OperatorType.EQ:
					strOp = " = ";
					break;
				case OperatorType.GT_GT_EQ:
					strOp = " >>= ";
					break;
				case OperatorType.LT_LT_EQ:
					strOp = " <<= ";
					break;
				case OperatorType.MINUS_EQ:
					strOp = " -= ";
					break;
				case OperatorType.MUL_EQ:
					strOp = " *= ";
					break;
				case OperatorType.PLUS_EQ:
					strOp = " += ";
					break;
				case OperatorType.MOD_EQ:
					strOp = " %= ";
					break;
				default:
					strOp = " unknown assignment operator type ";
					break;
			}
			return strOp;
		}
		public static string PrettyPrint( Stmt st, int nTabLevel )
		{
			string strRet = "";
			string strTab = "";
			string strTabMinus1 = "";
			if( st == null ) return strRet;
			int i = 0;
			for( i = 0; i < nTabLevel; i++ ) strTab += "  ";
			for( i = 0; i < nTabLevel - 1; i++ ) strTabMinus1 += "  ";
			switch( st.Kind ) {
				case StmtKind.AgentClassDeclaration:
					AgentClassDeclStmt acd = (AgentClassDeclStmt)st.Contents;
					strRet += strTab + AccessType2Str( acd.m_accessType ) + "agentclass " + acd.m_agentClassName + " ";
					if( acd.m_inheritanceList != null && acd.m_inheritanceList.Count > 0 ) {
						strRet += "extends {" + PrintUtil.List2Str( acd.m_inheritanceList ) + "}";
					}
					strRet += "\n";
					strRet += strTab + "{\n";
					strRet += PrintUtil.PrettyPrint( acd.m_stmtList, nTabLevel + 1 );
					strRet += strTab + "}\n";
					break;
				case StmtKind.Block:
					strRet += strTabMinus1 + "{\n";
					strRet += PrintUtil.PrettyPrint( (LinkedList<Stmt>)st.Contents, nTabLevel  );
					strRet += strTabMinus1 + "}\n";
					break;
				case StmtKind.Break:
				case StmtKind.Continue:
				case StmtKind.Expression:
				case StmtKind.Include:
				case StmtKind.LabelDeclaration:
				case StmtKind.MemberVarDeclaration:
				case StmtKind.Nop:
				case StmtKind.PlanAnd:
				case StmtKind.PlanOr:
				case StmtKind.PlanPar:
				case StmtKind.Return:
					strRet += strTab + st.ToString() + "\n";
					break;
				case StmtKind.ClassDeclaration:
					ClassDeclStmt cds = (ClassDeclStmt)st.Contents;
					strRet += strTab + AccessType2Str( cds.m_accessType ) + "class " + cds.m_className + " ";
					if( cds.m_inheritanceList != null && cds.m_inheritanceList.Count > 0 ) {
						strRet += "extends {" + PrintUtil.List2Str( cds.m_inheritanceList );
					}
					strRet += "\n";
					strRet += strTab + "{\n";
					strRet += PrintUtil.PrettyPrint( cds.m_stmtList, nTabLevel + 1 );
					strRet += strTab + "}\n";
					break;
				case StmtKind.DoWhile:
					DoWhileStmt dws = (DoWhileStmt)st.Contents;
					strRet += strTab + "do\n";
					strRet += PrintUtil.PrettyPrint( dws.m_stmt, nTabLevel + 1 );
					strRet += strTab + "while( " + dws.m_expr + " );\n";
					break;
				case StmtKind.For:
					ForStmt fst = (ForStmt)st.Contents;
					strRet += strTab + "for( ";
					strRet += fst.m_initExpr + "; ";
					strRet += fst.m_conditionExpr + "; ";
					strRet += fst.m_updateExpr + " )\n";
					strRet += PrintUtil.PrettyPrint( fst.m_stmt, nTabLevel + 1 );
					break;
				case StmtKind.ForEach:
					ForEachStmt fes = (ForEachStmt)st.Contents;
					strRet += strTab + "foreach( " + fes.m_strIdentifier + " in " + fes.m_expr + " )\n";
					strRet += PrintUtil.PrettyPrint( fes.m_stmt, nTabLevel + 1 );
					break;
				case StmtKind.If:
					IfStmt ifs = (IfStmt)st.Contents;
					strRet += strTab + "if( " + ifs.m_expr + " )\n";
					strRet += PrintUtil.PrettyPrint( ifs.m_trueBranch, nTabLevel + 1 );
					if( ifs.m_falseBranch != null ) {
						strRet += PrintUtil.PrettyPrint( ifs.m_falseBranch, nTabLevel + 1 );
					}
					break;
				case StmtKind.MethodDeclaration:
					MethodDeclStmt mds = (MethodDeclStmt)st.Contents;
					strRet += strTab + AccessType2Str( mds.m_accessType ) + "function " + mds.m_strName + "( ";
					strRet += List2Str( mds.m_paramList );
					strRet += " )\n";
					strRet += strTab + "{\n";
					strRet += PrettyPrint( mds.m_stmtList, nTabLevel + 1 );
					strRet += strTab + "}\n";
					break;
				case StmtKind.PlanDeclaration:
					PlanDeclStmt pds = (PlanDeclStmt)st.Contents;
					strRet += strTab + AccessType2Str( pds.m_accessType ) + "plan " + pds.m_strName + "( ";
					strRet += List2Str( pds.m_paramList );
					strRet += " )\n";
					if( pds.m_preCond != null ) {
						strRet += strTab + "precond( " + pds.m_preCond + " )\n";
					}
					if( pds.m_maintainCond != null ) {
						strRet += strTab + "maintains( " + pds.m_maintainCond + " )\n";
					}
					strRet += strTab + "{\n";
					strRet += PrettyPrint( pds.m_stmtList, nTabLevel + 1 );
					strRet += strTab + "}\n";
					break;
				case StmtKind.Switch:
					SwitchStmt swstm = (SwitchStmt)st.Contents;
					strRet += strTab + "switch( " + swstm.m_expr + " ) {\n";
					strRet += PrettyPrint( swstm.m_stmtList, nTabLevel + 1 );
					strRet += strTab + "}\n";
					break;
				case StmtKind.While:
					WhileStmt wstmt = (WhileStmt)st.Contents;
					strRet += strTab + "while( " + wstmt.m_expr + " )\n";
					strRet += PrettyPrint( wstmt.m_stmt, nTabLevel + 1 );
					break;
				case StmtKind.Match:
					MatchStmt mstm = (MatchStmt)st.Contents;
					int nBranchIndex = 0;
					strRet += strTab + "match " + mstm.m_subjectExpr + " with{\n";
					foreach( YaccaiMatchPair p in mstm.m_targetList ) {
						strRet += strTab;
						if( nBranchIndex > 0 ) strRet += "| ";
						if( p.m_exprList != null && p.m_exprList.Count > 0 ) {
							strRet += List2Str<Expr>( p.m_exprList );
						}
						strRet += " -> \n";
						strRet += PrettyPrint( p.m_stmt, nTabLevel + 1 );
						nBranchIndex++;
					}
					strRet += strTab + "}\n";
					break;
				default:
					strRet += " unknown statement kind \n";
					break;
			}
			return strRet;

		}
		public static string PrettyPrint( LinkedList<Stmt> stmtList, int nTabLevel )
		{
			string strRet = "";
			foreach( Stmt st in stmtList ) {
				strRet += PrintUtil.PrettyPrint( st, nTabLevel + 1 );
			}
			return strRet;
		}
	}
	class ForStmt
	{
		public LinkedList<Expr> m_initExpr;
		public LinkedList<Expr> m_conditionExpr;
		public LinkedList<Expr> m_updateExpr;
		public Stmt m_stmt;
		public ForStmt( LinkedList<Expr> iE, LinkedList<Expr> cE, LinkedList<Expr> uE, Stmt s )
		{
			m_initExpr = iE;
			m_conditionExpr = cE;
			m_updateExpr = uE;
			m_stmt = s;
		}
		public override string ToString()
		{
			string strRet = "for( ";
			strRet += PrintUtil.List2Str<Expr>( m_initExpr );
			strRet += "; ";
			strRet += PrintUtil.List2Str<Expr>( m_conditionExpr );
			strRet += "; ";
			strRet += PrintUtil.List2Str<Expr>( m_updateExpr );
			strRet += " ) ";
			strRet += m_stmt;
			return strRet;
		}
	}

	class ForEachStmt
	{
		public string m_strIdentifier;
		public Expr m_expr;
		public Stmt m_stmt;
		public ForEachStmt( string id, Expr exp, Stmt s )
		{
			m_strIdentifier = id;
			m_expr = exp;
			m_stmt = s;
		}
		public override string ToString()
		{
			string strRet = "foreach( ";
			strRet += m_strIdentifier;
			strRet += " in ";
			strRet += m_expr;
			strRet += " ) ";
			strRet += m_stmt;
			return strRet;
		}
	}

	class LabelDeclStmt
	{
		public Expr m_expr;
		public LabelDeclStmt( Expr e )
		{
			m_expr = e;
		}
		public LabelDeclStmt()
		{
			// default:
			m_expr = null;
		}
		public override string ToString()
		{
			string strRet = "";
			if( m_expr != null ) {
				strRet += "case " + m_expr + ": ";
			} else strRet += "default: ";
			return strRet;
		}
	}
	class SwitchStmt
	{
		public LinkedList<Stmt> m_stmtList;
		public Expr m_expr;
		public SwitchStmt( Expr e, LinkedList<Stmt> stmtList )
		{
			m_expr = e;
			m_stmtList = stmtList;
		}
		public override string ToString()
		{
			string strRet = "switch( ";
			strRet += m_expr;
			strRet += " ) {\n";
			foreach( Stmt s in m_stmtList )
				strRet += s;
			strRet += "}";
			return strRet;
		}
	}

	class DoWhileStmt
	{
		public Expr m_expr;
		public Stmt m_stmt;
		public DoWhileStmt( Expr e, Stmt s )
		{
			m_expr = e;
			m_stmt = s;
		}
		public override string ToString()
		{
			string strRet = "do( ";
			strRet += m_expr;
			strRet += " ) ";
			strRet += m_stmt;
			strRet += "while( " + m_expr + "); ";
			return strRet;
		}
	}

	class MemberVarDeclStmt
	{
		public LinkedList<MemberVarExprPair> m_pairList;
		public bool m_bStatic;
		public YaccaiAccessType m_accessType;
		public MemberVarDeclStmt( LinkedList<MemberVarExprPair> pL, bool bStatic, YaccaiAccessType a )
		{
			m_pairList = pL;
			m_bStatic = bStatic;
			m_accessType = a;
		}
		public override string ToString()
		{
			string strRet = PrintUtil.AccessType2Str( m_accessType );
			if( m_bStatic ) strRet += "static ";
			if( m_pairList != null && m_pairList.Count > 0 ) {
				strRet += m_pairList[0].m_strName;
				if( m_pairList[0].m_expr != null ) {
					strRet += " = " + m_pairList[0].m_expr;
				}
				for( int i = 1; i < m_pairList.Count; i++ ) {
					strRet += ", " + m_pairList[i].m_strName;
					if( m_pairList[i].m_expr != null ) {
						strRet += " = " + m_pairList[i].m_expr;
					}
				}
			}
			strRet += ";";
			return strRet;
		}
	}
	class MethodDeclStmt
	{
		public string m_strName;
		public ArrayList m_paramList;
		public LinkedList<Stmt> m_stmtList;
		public YaccaiAccessType m_accessType;
		public bool m_bStatic;
		public MethodDeclStmt( string n, ArrayList p, LinkedList<Stmt> lst, YaccaiAccessType yat, bool bStatic )
		{
			m_strName = n;
			m_paramList = p;
			m_stmtList = lst;
			m_accessType = yat;
			m_bStatic = bStatic;
		}
		public override string ToString()
		{
			string strRet = PrintUtil.AccessType2Str( m_accessType ) + "function " ;
			strRet += m_strName + "( ";
			strRet += PrintUtil.List2Str( m_paramList );
			strRet += " ) {\n";
			foreach( Stmt s in m_stmtList )
				strRet += s;
			strRet += "}";
			return strRet;
		}
	}

	class PlanDeclStmt
	{
		public string m_strName;
		public ArrayList m_paramList;
		public LinkedList<Stmt> m_stmtList;
		public Expr m_preCond;
		public Expr m_maintainCond;
		public YaccaiAccessType m_accessType;
		public PlanDeclStmt( string n, ArrayList p, LinkedList<Stmt> lst, YaccaiAccessType yat )
		{
			m_strName = n;
			m_paramList = p;
			m_stmtList = lst;
			m_preCond = null;
			m_maintainCond = null;
			m_accessType = yat;
		}
		public PlanDeclStmt( string n, ArrayList p, Expr preCond,
								Expr maintainCond,
								LinkedList<Stmt> lst, YaccaiAccessType yat )
		{
			m_strName = n;
			m_paramList = p;
			m_preCond = preCond;
			m_maintainCond = maintainCond;
			m_stmtList = lst;
			m_accessType = yat;
		}
		public override string ToString()
		{
			string strRet = PrintUtil.AccessType2Str( m_accessType ) + "plan ";
			strRet += m_strName;
			strRet += "( ";
			strRet += PrintUtil.List2Str( m_paramList );
			strRet += ")\n";
			if( m_preCond != null ) {
				strRet += "precond (" + m_preCond + ")\n";
			}
			if( m_maintainCond != null ) {
				strRet += "maintains (" + m_maintainCond + ")\n";
			}
			strRet += "{\n";
			strRet += PrintUtil.List2Str<Stmt>( m_stmtList );
			strRet += "}\n";
			return strRet;
		}
	}

	class ClassDeclStmt
	{
		public string m_className;
		public ArrayList m_inheritanceList;
		public LinkedList<Stmt> m_stmtList;
		public YaccaiAccessType m_accessType;
		public ClassDeclStmt( string n, ArrayList h, LinkedList<Stmt> lst, YaccaiAccessType yat )
		{
			m_className = n;
			m_inheritanceList = h;
			m_stmtList = lst;
			m_accessType = yat;
		}
		public override string ToString()
		{
			string strRet = PrintUtil.AccessType2Str( m_accessType ) + "class ";
			strRet += m_className;
			if( m_inheritanceList != null ) {
				strRet += " extends { " + PrintUtil.List2Str( m_inheritanceList ) + " }\n";
			}
			strRet += "{\n";
			strRet += PrintUtil.List2Str<Stmt>( m_stmtList );
			strRet += "}\n";
			return strRet;
		}
	}

	class AgentClassDeclStmt
	{
		public string m_agentClassName;
		public ArrayList m_inheritanceList;
		public LinkedList<Stmt> m_stmtList;
		public YaccaiAccessType m_accessType;
		public AgentClassDeclStmt( string n, ArrayList h, LinkedList<Stmt> lst, YaccaiAccessType yat )
		{
			m_agentClassName = n;
			m_inheritanceList = h;
			m_stmtList = lst;
			m_accessType = yat;
		}
		public override string ToString()
		{
			string strRet = PrintUtil.AccessType2Str( m_accessType ) + "agentclass ";
			strRet += m_agentClassName;
			if( m_inheritanceList != null && m_inheritanceList.Count > 0 ) {
				strRet += " extends { " + PrintUtil.List2Str( m_inheritanceList ) + " }\n";
			}
			strRet += "{\n";
			foreach( Stmt st in m_stmtList ) {
				strRet += st;
			}
			strRet += "}\n";
			return strRet;
		}

	}
	class Stmt
	{
		private StmtKind m_kind;
		private SrcPos m_pos;
		private Object m_contents;

		public StmtKind Kind
		{
			get
			{
				return m_kind;
			}
		}

		public SrcPos Pos
		{
			get
			{
				return m_pos;
			}
			set
			{
				m_pos = value;
			}
		}

		public Object Contents
		{
			get
			{
				return m_contents;
			}
			set
			{
				m_contents = value;
			}
		}

		public Stmt( SrcPos p )
		{
			m_kind = StmtKind.Nop;
			m_pos = p;
		}
		public Stmt( Expr e )
		{
			// expression statement
			m_kind = StmtKind.Expression;
			m_contents = e;
			m_pos = e.Pos;
		}
		/*
		public Stmt( AssignmentStmt a, SrcPos p )
		{
			// assignment
			m_kind = StmtKind.Assignment;
			m_contents = a;
			m_pos = p;
		}
		*/
		public Stmt( IfStmt ifs, SrcPos p )
		{
			// If statement
			m_pos = p;
			m_kind = StmtKind.If;
			m_contents = ifs;
		}

		public Stmt( StmtKind k, SrcPos p )
		{
			// Continue, Break, Return (void)
			m_kind = k;
			m_pos = p;
			m_contents = null;
		}

		public Stmt( StmtKind k, SrcPos p, Expr e )
		{
			// Return expr
			m_kind = k;
			m_pos = p;
			m_contents = e;
		}

		/*
		public Stmt( StmtKind k, SrcPos p, LinkedList<Stmt> lst )
		{
			// FuncDeclaration, ClassDeclaration
			m_kind = k;
			m_pos = p;
			m_contents = lst;	// statements list
		}
		*/

		public Stmt( WhileStmt w, SrcPos p )
		{
			// While statement
			m_kind = StmtKind.While;
			m_contents = w;
			m_pos = p;

		}

		public Stmt( ForStmt f, SrcPos p )
		{
			// For statement
			m_kind = StmtKind.For;
			m_contents = f;
			m_pos = p;
		}

		public Stmt( ForEachStmt f, SrcPos p )
		{
			// Foreach statement
			m_kind = StmtKind.ForEach;
			m_contents = f;
			m_pos = p;
		}

		public Stmt( DoWhileStmt dw, SrcPos p )
		{
			// Do While statement
			m_kind = StmtKind.DoWhile;
			m_contents = dw;
			m_pos = p;
		}

		public Stmt( SwitchStmt sw, SrcPos p )
		{
			// Switch statement
			m_kind = StmtKind.Switch;
			m_contents = sw;
			m_pos = p;
		}

		public Stmt( LabelDeclStmt lds, SrcPos p )
		{
			// Label declaration
			m_kind = StmtKind.LabelDeclaration;
			m_contents = lds;
			m_pos = p;
		}

		public Stmt( string strFilePath, SrcPos p )
		{
			// Include statement
			m_kind = StmtKind.Include;
			m_contents = strFilePath;
			m_pos = p;
		}

		public Stmt( MethodDeclStmt s, SrcPos p )
		{
			m_pos = p;
			m_contents = s;
			m_kind = StmtKind.MethodDeclaration;
		}

		public Stmt( ClassDeclStmt c, SrcPos p )
		{
			m_pos = p;
			m_contents = c;
			m_kind = StmtKind.ClassDeclaration;
		}
		public Stmt( AgentClassDeclStmt a, SrcPos p )
		{
			m_pos = p;
			m_contents = a;
			m_kind = StmtKind.AgentClassDeclaration;
		}
		public Stmt( LinkedList<Stmt> b, SrcPos p )
		{
			m_pos = p;
			m_contents = b;
			m_kind = StmtKind.Block;
		}
		public Stmt( StmtKind k, LinkedList<Expr> eList, SrcPos p )
		{
			m_kind = k;
			m_pos = p;
			m_contents = eList;
			
		}
		public Stmt( PlanDeclStmt pl, SrcPos p )
		{
			m_contents = pl;
			m_pos = p;
			m_kind = StmtKind.PlanDeclaration;
		}
		public Stmt( MemberVarDeclStmt mds, SrcPos p )
		{
			m_contents = mds;
			m_pos = p;
			m_kind = StmtKind.MemberVarDeclaration;
		}
		public Stmt( MatchStmt m, SrcPos p )
		{
			m_contents = m;
			m_pos = p;
			m_kind = StmtKind.Match;
		}
		public override string ToString()
		{
			Expr ex1;
			LinkedList<Stmt> stmList;
			string strRet = "";
			switch( m_kind ) {
				case StmtKind.Break:
					strRet = "break;";
					break;
				case StmtKind.Continue:
					strRet = "continue;";
					break;
				case StmtKind.Return:
					ex1 = (Expr)this.m_contents;
					strRet = "return";
					if( ex1 == null ) strRet += ";";
					else strRet += " " + ex1.ToString() + ";";
					break;
				case StmtKind.Nop:
					strRet += ";";
					break;
				case StmtKind.Include:
					strRet += "include " + ((Expr)this.m_contents).ToString() ;
					break;
				case StmtKind.Block:
					strRet += "{\n";
					stmList = (LinkedList<Stmt>)m_contents;
					foreach( Stmt st in stmList )
						strRet += st.ToString();
					strRet += "}\n";
					break;
				case StmtKind.Expression:
					strRet = ((Expr)m_contents).ToString() + ";";
					break;
				case StmtKind.PlanAnd:
					strRet = "and (" + PrintUtil.List2Str<Expr>( (LinkedList<Expr>)m_contents ) + ");";
					break;
				case StmtKind.PlanOr:
					strRet = "or (" + PrintUtil.List2Str<Expr>( (LinkedList<Expr>)m_contents ) + ");";
					break;
				case StmtKind.PlanPar:
					strRet = "par (" + PrintUtil.List2Str<Expr>( (LinkedList<Expr>)m_contents ) + ");";
					break;
				default:
					strRet = m_contents.ToString();
					break;
			}
			return strRet;
		}
	}

}
