using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Net.Sockets;
using System.Net;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Diagnostics;
using System.Reflection;
using Antlr.Runtime;

namespace Yaccai
{
	class YProcAddress
	{
		public string m_strHostName;
		public int m_nPort;
		public YProcAddress( string strHost, int nPort )
		{
			m_strHostName = strHost;
			m_nPort = nPort;
		}
	}
	class YMpiMsg
	{
		public string m_strCommName;
		public int m_nSrc;
		public int m_nDst;
		public string m_strContents;
		public YMpiMsg( int nSrc, int nDst, string strCommName, string strContents )
		{
			m_nSrc = nSrc;
			m_nDst = nDst;
			m_strCommName = strCommName;
			m_strContents = strContents;
		}

	}

	class ReduceResult
	{
		public Collection<int> m_pendingProcIdSet;
		public Hashtable m_valSet;
		public YaccaiVal m_opVal;
		public AutoResetEvent m_readyEvent;
		public bool m_bAllRequestSent;
		// for debug
		public int m_nProcId;
		public ReduceResult( YaccaiVal opVal )
		{
			m_pendingProcIdSet = new Collection<int>();
			m_valSet = new Hashtable();
			m_opVal = opVal;
			m_readyEvent = new AutoResetEvent( false );
			m_bAllRequestSent = false;
		}
		public void AddValue( int nProcId, YaccaiVal val )
		{
			//Console.WriteLine( "{2}) ReduceRet: Adding value val = {0} from id = {1}", val, nProcId, m_nProcId );
			lock( m_valSet ) {
				m_valSet.Add( nProcId, val );
			}
			lock( m_pendingProcIdSet ) {
				m_pendingProcIdSet.Remove( nProcId );
			}
			//Console.WriteLine( m_pendingProcIdSet );
			if( m_bAllRequestSent && m_pendingProcIdSet.Count == 0 ) m_readyEvent.Set();
		}
		public void RemoveId( int nProcId )
		{
			lock( m_pendingProcIdSet ) {
				m_pendingProcIdSet.Remove( nProcId );
			}
			if( m_bAllRequestSent && m_pendingProcIdSet.Count == 0 ) m_readyEvent.Set();
		}
	}
	class RemoteEvalExpr
	{
		public string m_expr;
		public string m_op;
		public int m_nId;
		public Socket m_sock;
		public int m_nSrc;
		public string m_strCommName;
		public RemoteEvalExpr( string expr, string op, int nId, Socket s, int nSrc, string strCommName )
		{
			m_expr = expr;
			m_sock = s;
			m_nId = nId;
			m_nSrc = nSrc;
			m_strCommName = strCommName;
			m_op = op;
		}
	}

	partial class YEvaluator
	{
		private static Hashtable m_hProcSock = new Hashtable();
		private static Hashtable m_hProcAddress = new Hashtable();
		private string m_strPortMgrAddr;
		private int m_nPortMgrPort;
		private int m_nMyProcId;
		private int m_nMyPort;
		private string m_strMyHost;
		private Hashtable m_hProcMsgQueue;
		// private static Dictionary<int, LinkedList<RemoteEvalExpr>> m_hProcRemoteEvalExpr = new Dictionary<int,LinkedList<RemoteEvalExpr>>();
		// private static Dictionary<int, AutoResetEvent> m_hProcRemoteEvalExprEvent = new Dictionary<int, AutoResetEvent>();
		private AutoResetEvent m_listenerReadyEvent;
		private AutoResetEvent m_msgArrivedEvent;
		private Hashtable m_hCommMsgList;
		private object m_oCommMsgListLock;
		private object m_oProcConnectingLock;
		private object m_oProcSockLock;
		private Hashtable m_hProcConnecting;
		public int m_nIOConnectorPort;
		private Hashtable m_hCommYComObj;
		private bool m_bIdleWaitForMsg;

		private int m_nMaxCollectiveIdentifier;
		private Hashtable m_hCollectiveIdReduceObj;

		public void SetMyRank( int nRank )
		{
			m_nMyProcId = nRank;
			
		}
		private void SocketConnect( Socket sk, string strHost, int nPort )
		{
			Type socketT = sk.GetType();
			//Console.WriteLine( "***** SocketConnect: strHost = {0}, port = {1}", strHost, nPort );

//#if __LINUX__
			IPHostEntry ipHostEntry = Dns.GetHostEntry( strHost );
			IPAddress[] ipAddrs = ipHostEntry.AddressList;
			if( ipAddrs.Length <= 0 ) {
				Console.WriteLine( "Error: Host {0} not found", strHost );
				Environment.Exit( 1 );
			}
			IPAddress ipAddr = ipAddrs[0];
			sk.Connect( new IPEndPoint( ipAddr, nPort ) );
/*
#else	// Windows
			MethodInfo methInfo = null;
			try {
				methInfo = socketT.GetMethod( "Connect", new Type[] { strHost.GetType(), nPort.GetType() } );
			} catch( Exception ex ) {
				methInfo = null;
			}
			if( methInfo != null ) {
				sk.Connect( strHost, nPort );
			} else { // mono does not support Connect( string, int )
				IPHostEntry ipHostEntry = Dns.GetHostEntry( strHost );
				IPAddress[] ipAddrs = ipHostEntry.AddressList;
				if( ipAddrs.Length <= 0 ) {
					Console.WriteLine( "Error: Host {0} not found", strHost );
					Environment.Exit( 1 );
				}
				IPAddress ipAddr = ipAddrs[0];
				sk.Connect( new IPEndPoint( ipAddr, nPort ) );
			}
#endif
*/
		}
		private YaccaiVal ExprValFromString( string strExpr, YEnv env )
		{
			// Console.WriteLine( "ExprVal({0})", strExpr );
			YaccaiLexer lex = new YaccaiLexer( new Antlr.Runtime.ANTLRStringStream( strExpr ) );
			CommonTokenStream tks = new CommonTokenStream( lex );
			YaccaiParser yParser = new YaccaiParser( tks );
			yParser.StrInputFileName = "NativeFuncs.cs";
			Expr yExp = null;
			try {
				yExp = yParser.expression();
			} catch( Exception ex ) {
				Console.WriteLine( "In function ExprValFromString @ YMpi.cs:" );
				Console.WriteLine( ex.Message );
				RuntimeError( new SrcPos( "YMpi.cs:ExprValFromString", 1, 0 ), "Parse failed" );
			}
			if( yParser.IsSuccessParse() ) {
				if( yExp == null ) {
					RuntimeError( new SrcPos( "YMpi.cs:ExprValFromString", 2, 0 ), "Parse failed" );
				}
				YaccaiVal retVal = EvalExpr( yExp, env );
				//Console.WriteLine( "{0}", retVal );
				return retVal;
			}
			return new YaccaiVal( YaccaiType.NULL, null );
		}

		private string GetStrToSend( YaccaiVal obj )
		{
			switch( obj.m_type ) {
				case YaccaiType.INT:
					return "1 " + obj.m_contents.ToString();
				case YaccaiType.DOUBLE:
					return "2 " + obj.m_contents.ToString();
				case YaccaiType.CHAR:
					return "3 " + obj.m_contents.ToString();
				default:
					return "-1 ";
			}
		}
		private YaccaiVal GetValueFromStrReceived( string strInput, int[] arrId )
		{
			string strPattern = @"([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+(.*)";
			if( Regex.IsMatch( strInput, strPattern ) ) {
				Match g = Regex.Match( strInput, strPattern );
				int v1 = int.Parse( g.Groups[1].ToString() );
				int id = int.Parse( g.Groups[2].ToString() );
				arrId[0] = id;
				int objType = int.Parse( g.Groups[3].ToString() );
				string objStrVal = g.Groups[4].ToString();
				YaccaiVal retVal;
				switch( objType ) {
					case 1:
						retVal = new YaccaiVal( YaccaiType.INT, int.Parse( objStrVal ) );
						break;
					case 2:
						retVal = new YaccaiVal( YaccaiType.DOUBLE, double.Parse( objStrVal ) );
						break;
					case 3:
						retVal = new YaccaiVal( YaccaiType.CHAR, char.Parse( objStrVal ) );
						break;
					case 4:
						retVal = NativeObjects.YStringObject( objStrVal );
						break;
					default:
						retVal = new YaccaiVal( YaccaiType.NULL, null );
						break;
				}
				return retVal;
			} else {
				RuntimeError( new SrcPos( "YMpi.cs", 0, 0 ), "Invalid message format" );
			}
			return null;
		}

		private void ProcessReduceRequest( RemoteEvalExpr rmExpr )
		{
			Socket s = rmExpr.m_sock;
			string strCommName = rmExpr.m_strCommName;
			int nSrc = rmExpr.m_nSrc;
			int nReduceId = rmExpr.m_nId;

			//m_bSuspendDeliverMessage = true;
			//Console.WriteLine( "Agent{0}: Wait to evaluate {1}", m_nMyProcId, rmExpr.m_op );
			m_mutexExecutionTurn.WaitOne();
			m_bEvalRemote = true;
			YaccaiVal opVal = ExprValFromString( rmExpr.m_op, m_gYEnv );
			m_bEvalRemote = false;
			//m_bSuspendDeliverMessage = false;
			//Console.WriteLine( "Agent{0}: Restarted DeliverMsg", m_nMyProcId );
			m_mutexExecutionTurn.ReleaseMutex();
			if( opVal.m_type == YaccaiType.NOT_FOUND ) {
				RuntimeError( new SrcPos( "YMpi.cs", 0, 0 ), "Operator " + rmExpr.m_op + "not found" );
			}
			//Console.WriteLine( "Agent{0}: Wait to evaluate {1}", m_nMyProcId, rmExpr.m_expr );
			//m_bSuspendDeliverMessage = true;
			m_mutexExecutionTurn.WaitOne();
			m_bEvalRemote = true;
			YaccaiVal exprVal = ExprValFromString( rmExpr.m_expr, m_gYEnv );
			m_bEvalRemote = false;
			//Console.WriteLine( "Agent{0}: Evaluated", m_nMyProcId );
			//m_bSuspendDeliverMessage = false;
			m_mutexExecutionTurn.ReleaseMutex();
			//Console.WriteLine( "Agent{0}: Restarted DeliverMsg", m_nMyProcId );
			//Console.WriteLine( "exprVal = {0}", exprVal );
			string strToSend = strCommName + "!3!" + m_nMyProcId + "!" + nSrc + "!" + nReduceId + "!";
			switch( exprVal.m_type ) {
				case YaccaiType.INT:
					strToSend += "1 " + exprVal.ToString();
					break;
				case YaccaiType.DOUBLE:
					strToSend += "2 " + exprVal.ToString();
					break;
				case YaccaiType.CHAR:
					strToSend += "3 " + exprVal.ToString();
					break;
				case YaccaiType.OBJECT: {
						YObject yObj = (YObject)exprVal.m_contents;
						if( yObj.m_yclass.m_strName.Equals( "String" ) ) {
							if( yObj.GetMemberEx( "2contents", m_gYEnv ) != null ) {
								strToSend += "4 " + (string)yObj.GetMemberEx( "2contents", m_gYEnv );
							}
						} else {
							if( yObj.m_yclass.m_strName.Equals( "YValIdPair" ) ) { // 
								strToSend += "5 " + ( (YaccaiVal)yObj.m_instanceVars["Id"] ).Marshall().ToString() + " " + GetStrToSend( (YaccaiVal)yObj.m_instanceVars["Value"] );
							} else {
								// normal object
								RuntimeError( new SrcPos( "YMpi.cs", 0, 0 ), "Collective operations do not support this data type! (" + yObj.m_yclass.m_strName + ")" );
							}
						}
					}
					break;
				default:
					strToSend += "-1 " + exprVal.ToString();
					break;
			}
			strToSend += "!";
			//Console.WriteLine( "Agent{0}: acquiring socket[{1}] lock", m_nMyProcId, nSrc );
			lock( s ) {
				//Console.WriteLine( "Agent{0}: got socket[{1}] lock", m_nMyProcId, nSrc );
				NetworkStream ns = new NetworkStream( s );
				StreamWriter sw = new StreamWriter( ns );
				sw.WriteLine( strToSend );
				sw.Flush();
			}
			//Console.WriteLine( "Agent{0}: released socket[{1}] lock", m_nMyProcId, nSrc );
		}
		public void MessageListening( object nProcIdObj )
		{
			//Console.WriteLine( "MessageListening" );
			int nProcId = (int)nProcIdObj;
			Socket s = (Socket)m_hProcSock[nProcId];
			if( !m_hProcMsgQueue.ContainsKey( nProcId ) ) {
				m_hProcMsgQueue.Add( nProcId, new LinkedList<string>() );
			}
			/*
			if( !m_hProcRemoteEvalExpr.ContainsKey( nProcId ) ) {
				lock( m_hProcRemoteEvalExpr ) {
					m_hProcRemoteEvalExpr.Add( nProcId, new LinkedList<RemoteEvalExpr>() );
				}
			}
			*/
			string strLine = null;
			NetworkStream ns = new NetworkStream( s );
			StreamReader sr = new StreamReader( ns );
			string[] arrRecv;
			string strSplit = "!";
			string strCommName;
			string strContents;
			int nSrc;
			int nDst;
			int nCmd;
			
			while( true ) {
				try {
					strLine = sr.ReadLine();
					if( strLine == null ) {
						continue;
					}
				} catch {
					continue;
				}
				//Console.WriteLine( "{0}: Received: {1}", m_nMyProcId, strLine );
				arrRecv = Regex.Split( strLine, strSplit );
				if( arrRecv.Length < 5 ) {
					Console.WriteLine( "Invalid message received" );
					continue;
				}
				strCommName = arrRecv[0];
				nCmd = Int32.Parse( arrRecv[1] );
				nSrc = Int32.Parse( arrRecv[2] );
				nDst = Int32.Parse( arrRecv[3] );
				
				switch( nCmd ) {
					case 0: case 1:
						if( m_hCommYComObj == null ) break;
						//Console.WriteLine( "{0} msgCommName = {1}", m_nMyProcId, strCommName );
						if( !m_hCommYComObj.ContainsKey( strCommName ) ) break;
						strContents = arrRecv[4];
						LinkedList<YMpiMsg> msgList = null;
						//Console.WriteLine( "{0} locking m_oCommMsgListLock", m_nMyProcId );
						lock( m_oCommMsgListLock ) {
							if( !m_hCommMsgList.ContainsKey( strCommName ) ) {
								msgList = new LinkedList<YMpiMsg>();
								m_hCommMsgList.Add( strCommName, msgList );
							} else msgList = (LinkedList<YMpiMsg>)m_hCommMsgList[strCommName];
						}
						YMpiMsg msg = new YMpiMsg( nSrc, nDst, strCommName, strContents );
						//Console.WriteLine( "{0} locking msglist", m_nMyProcId );
						lock( msgList ) {
							//Console.WriteLine( "{0}: Adding msg {1} from {2}", m_nMyProcId, msg.m_strContents, msg.m_nSrc );
							msgList.Add( msg );
						}
						m_msgArrivedEvent.Set();
						break;
					case 2:	// reduce request
						{
							//Console.WriteLine( "{0}) received: {1}", m_nMyProcId, strLine );
							string nReduceId = arrRecv[4];
							string strTargetData = arrRecv[5];
							string strReduceOperator = arrRecv[6];
							if( m_hCommYComObj == null ) break;
							if( !m_hCommYComObj.ContainsKey( strCommName ) ) {
								string strReject = strCommName + "!4!" + m_nMyProcId + "!" + nSrc + "!" + nReduceId + "!" + "1 0!";
								//Console.WriteLine( "> Agent{0}: > acquiring socket[{1}] lock in processMsg", m_nMyProcId, nSrc );
								//lock( s ) {
								//	Console.WriteLine( "> Agent{0}: > got socket[{1}] lock in processMsg", m_nMyProcId, nSrc );
									NetworkStream ns1 = new NetworkStream( s );
									StreamWriter sw1 = new StreamWriter( ns1 );
									sw1.WriteLine( strReject );
									sw1.Flush();
								//}
								//Console.WriteLine( "> Agent{0}: > release socket[{1}]", m_nMyProcId, nSrc );
								break;
							}
							if( m_gYEnv == null ) m_gYEnvReadyEvent.WaitOne();
							ProcessReduceRequest( new RemoteEvalExpr( strTargetData, strReduceOperator, int.Parse( nReduceId ), s, nSrc, strCommName ) );
							//Console.WriteLine( "> Agent{0}: > locking RemoteList[{1}]", m_nMyProcId, nProcId );
							//lock( m_hProcRemoteEvalExpr[nProcId] ) {
							//	Console.WriteLine( "> Agent{0}: > locked RemoteList[{1}]", m_nMyProcId, nProcId );
								//Console.WriteLine( "Agent{0}: adding {1} for reduction", m_nMyProcId, strTargetData );
								//m_hProcRemoteEvalExpr[nProcId].Add( new RemoteEvalExpr( strTargetData, strReduceOperator, int.Parse( nReduceId ), s, nSrc, strCommName ) );
							//}
							//Console.WriteLine( "> Agent{0}: > unlocked RemoteList[{1}]", m_nMyProcId, nProcId );
							/*
							Console.WriteLine( "***** Agent{0}: Waiting for eval remote ...", m_nMyProcId );
							m_hProcRemoteEvalExprEvent[nProcId].WaitOne();
							Console.WriteLine( "Agent{0}: got remote value", m_nMyProcId );
							
							YaccaiVal opVal = m_hProcRemoteEvalExpr[nProcId][0].m_val;
							YaccaiVal exprVal = m_hProcRemoteEvalExpr[nProcId][1].m_val;
							lock( m_hProcRemoteEvalExpr[nProcId] ) {
								m_hProcRemoteEvalExpr[nProcId].Clear();
							}
							//Console.WriteLine( "exprVal = {0}", exprVal );
							if( opVal.m_type == YaccaiType.NOT_FOUND ) {
								RuntimeError( new SrcPos( "YMpi.cs", 0, 0 ), "Operator " + strReduceOperator + "not found" );
							}
							string strToSend = strCommName + "!3!" + m_nMyProcId + "!" + nSrc + "!" + nReduceId + "!";
							switch( exprVal.m_type ) {
								case YaccaiType.INT:
									strToSend += "1 " + exprVal.ToString();
									break;
								case YaccaiType.DOUBLE:
									strToSend += "2 " + exprVal.ToString();
									break;
								case YaccaiType.CHAR:
									strToSend += "3 " + exprVal.ToString();
									break;
								case YaccaiType.OBJECT: {
										YObject yObj = (YObject)exprVal.m_contents;
										if( yObj.m_yclass.m_strName.Equals( "String" ) ) {
											if( yObj.GetMemberEx( "2contents", m_gYEnv ) != null ) {
												strToSend += "4 " + (string)yObj.GetMemberEx( "2contents", m_gYEnv );
											}
										} else {
											if( yObj.m_yclass.m_strName.Equals( "YValIdPair" ) ) { // 
												strToSend += "5 " + ( (YaccaiVal)yObj.m_instanceVars["Id"] ).Marshall().ToString() + " " + GetStrToSend( (YaccaiVal)yObj.m_instanceVars["Value"] );
											} else {
												// normal object
												RuntimeError( new SrcPos( "YMpi.cs", 0, 0 ), "Collective operations do not support this data type! (" + yObj.m_yclass.m_strName + ")" );
											}
										}
									}
									break;
								default:
									strToSend += "-1 " + exprVal.ToString();
									break;
							}
							strToSend += "!";
							lock( s ) {
								StreamWriter sw = new StreamWriter( ns );
								sw.WriteLine( strToSend );
								sw.Flush();
							}
							*/
						}
						break;
					case 3: case 4:// reduce response
						{
							//Console.WriteLine( "{0}) received: {1}, nDst = {2}", m_nMyProcId, strLine, nDst );
							if( nDst == m_nMyProcId ) {
								int nReduceId = Int32.Parse( arrRecv[4] );
								if( nCmd == 4 ) {
									if( m_hCollectiveIdReduceObj.ContainsKey( nReduceId ) ) {
										ReduceResult reduceRsl = (ReduceResult)m_hCollectiveIdReduceObj[nReduceId];
										reduceRsl.RemoveId( nSrc );
									}
									break;
								}
								string strResult = arrRecv[5];
								
								YaccaiVal val = new YaccaiVal( YaccaiType.NULL );
								string[] arrReduceVal = null;
								try {
									arrReduceVal = Regex.Split( strResult, @"\s+" );
								} catch( Exception exc2 ) {
									Console.WriteLine( "In function MessageListening @ YMpi.cs:" );
									Console.WriteLine( exc2.Message );
									Console.WriteLine( "strResult = {0}", strResult );
								}
								//Console.WriteLine( "{2}) strResult = '{0}', len = {1}", strResult, arrReduceVal.Length, m_nMyProcId );
								//Console.WriteLine( "[0] = {0}", arrReduceVal[0] );
								//Console.WriteLine( "[1] = {0}", arrReduceVal[1] );

								if( arrReduceVal == null || arrReduceVal.Length < 2 ) {
									break;
								}
								int nType = Int32.Parse( arrReduceVal[0] );
								//Console.WriteLine( "{2}) nType = {0}, value = {1}", nType, arrReduceVal[1], m_nMyProcId );
								switch( nType ) {
									case 1:
										val = new YaccaiVal( YaccaiType.INT, Int32.Parse( arrReduceVal[1] ) );
										break;
									case 2:
										val = new YaccaiVal( YaccaiType.DOUBLE, Double.Parse( arrReduceVal[1] ) );
										break;
									case 3:
										val = new YaccaiVal( YaccaiType.CHAR, Char.Parse( arrReduceVal[1] ) );
										break;
									case 4: {
											string strReduceRetStr;
											strReduceRetStr = strResult.TrimStart( new char[] { ' ' } );
											strReduceRetStr = strResult.Remove( 0, arrReduceVal[0].Length );
											strReduceRetStr = strReduceRetStr.TrimStart( new char[] { ' ' } );
											val = NativeObjects.YStringObject( strReduceRetStr );
										}
										break;
									case 5: {
											int[] arrId = new int[1];
											val = GetValueFromStrReceived( strResult, arrId );
											ArrayList arrValList = new ArrayList();
											arrValList.Add( val );
											arrValList.Add( new YaccaiVal( YaccaiType.INT, arrId[0] ) );
											//Console.WriteLine( "{0}) creating YValIdPair object", m_nMyProcId );
											m_mutexExecutionTurn.WaitOne();
											m_bEvalRemote = true;
											val = CreateObjectFromYValues( "YValIdPair", arrValList, m_gYEnv );
											m_bEvalRemote = false;
											m_mutexExecutionTurn.ReleaseMutex();
										}
										break;
									default:
										break;
								}
								// Console.WriteLine( "{2}) reduceId = {0}, val = {1}", nReduceId, val, m_nMyProcId );
								if( m_hCollectiveIdReduceObj.ContainsKey( nReduceId ) ) {
									//Console.WriteLine( "{2}) reduceId = {0}, from = {1}", nReduceId, nSrc, m_nMyProcId );
									ReduceResult reduceRet = (ReduceResult)m_hCollectiveIdReduceObj[nReduceId];
									if( val == null ) {
										Console.WriteLine( "rank = {0}, strResult = {1}", m_nMyProcId, strResult );
										Environment.Exit( 0 );
									}
									reduceRet.AddValue( nSrc, val );
								}
								
							}
						}
						break;
				}
			}
		}
		private bool ConnectTo( int nProcId )
		{
			//Console.WriteLine( "ConnectTo" );
			lock( m_oProcSockLock ) {
				if( m_hProcSock.ContainsKey( nProcId ) ) return true;
			}
			int nPort = -1;
			string strHost = null;
			YProcAddress procAddr;
			bool bNeedQueryPortMgr = false;
			if( m_hProcAddress.ContainsKey( nProcId ) ) {
				procAddr = (YProcAddress)m_hProcAddress[nProcId];
				if( procAddr.m_nPort <= 0 ) {
					bNeedQueryPortMgr = true;
				} else {
					nPort = procAddr.m_nPort;
					strHost = procAddr.m_strHostName;
				}
			} else bNeedQueryPortMgr = true;
			if( bNeedQueryPortMgr ) {
				try {
					//Console.WriteLine( "Connect: PortMgrAddr = {0}", m_strPortMgrAddr );
					TcpClient tClient = new TcpClient( m_strPortMgrAddr, m_nPortMgrPort );
					NetworkStream ns = tClient.GetStream();
					StreamReader sr = new StreamReader( ns );
					StreamWriter sw = new StreamWriter( ns );
					sw.WriteLine( "YaccaiPort 100 " + nProcId );
					sw.Flush();
					string strRet = sr.ReadLine();
					string[] arrRet = Regex.Split( strRet, @"\s+" );
					if( arrRet.Length >= 2 ) {
						if( arrRet[1].Equals( "100" ) ) {
							if( arrRet.Length >= 4 ) {
								nPort = Int32.Parse( arrRet[2] );
								strHost = arrRet[3];
								m_hProcAddress.Add( nProcId, new YProcAddress( strHost, nPort ) );
							}
						} else {
							if( arrRet[1].Equals( "102" ) ) {
								strHost = null;
								nPort = -1;
							}
						}
					}

					sw.Write( "YaccaiPort 10" );
					sw.Flush();
					tClient.Close();
				} catch( Exception ex ) {
					Console.WriteLine( "In function ConnectTo @ YMpi.cs:" );

					Console.WriteLine( ex.Message );
				}
			}
			if( nPort > 0 ) {
				try {
					if( m_hProcSock.ContainsKey( nProcId ) ) return true;
					lock( m_oProcConnectingLock ) {
						if( !m_hProcConnecting.ContainsKey( nProcId ) ) 
							m_hProcConnecting.Add( nProcId, new Mutex() );
					}
					Mutex m = (Mutex)m_hProcConnecting[nProcId];
					if( m_hProcSock.ContainsKey( nProcId ) ) return true;
					m.WaitOne();
					if( m_hProcSock.ContainsKey( nProcId ) ) {
						m.ReleaseMutex();
						return true;
					}
					//Console.WriteLine( "{0}: ConnectTo: strHost = {1}, port = {2}", m_nMyProcId, strHost, nPort );
					
					Socket sk = new Socket( AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp );
					//sk.Connect( strHost, nPort );
					SocketConnect( sk, strHost, nPort );
					NetworkStream nsSock = new NetworkStream( sk );
					StreamWriter nsWrite = new StreamWriter( nsSock );
					StreamReader srSock = new StreamReader( nsSock );
					//Console.WriteLine( "{0}: sending YaccaiPort 50...", m_nMyProcId );
					nsWrite.WriteLine( "YaccaiPort 50 " + m_nMyProcId );
					nsWrite.Flush();
					string strResponse50 = srSock.ReadLine();
					//Console.WriteLine( "{0}(ConnectTo): received = {1}", m_nMyProcId, strResponse50 );
					if( strResponse50.Contains( "YaccaiPort 50" ) ) {
						lock( m_oProcSockLock ) {
							m_hProcSock.Add( nProcId, sk );

							Thread th_msgListening = new Thread( new ParameterizedThreadStart( this.MessageListening ) );
							th_msgListening.Name = "YaccaiMsgListening" + nProcId;
							th_msgListening.Start( nProcId );
						}
					}
					nsWrite.Close();
					srSock.Close();
					m.ReleaseMutex();
					return true;
				} catch( Exception ex ) {
					Console.WriteLine( "In function ConnectTo (2) @ YMpi.cs:" );
					Console.WriteLine( ex.Message );
					return false;
				}
			}
			return false;
		}
		public void ProcessConnection( object sk )
		{
			Socket s = (Socket)sk;
			NetworkStream ns = new NetworkStream( s );
			StreamReader sr = new StreamReader( ns );
			StreamWriter sw = new StreamWriter( ns );
			string strHello = sr.ReadLine();
			string[] arrRet = Regex.Split( strHello, @"\s+" );
			if( arrRet.Length < 3 ) {
				Console.WriteLine( "Invalid hello message" );
				sw.WriteLine( "YaccaiPort 1" );
				sw.Flush();
				s.Close();
			} else {
				int nCmd = Int32.Parse( arrRet[1] );
				if( nCmd == 50 ) {	// hello
					int nProc = Int32.Parse( arrRet[2] );
					bool bAccept = true;
					
					lock( m_oProcSockLock ) {
						if( m_hProcSock.ContainsKey( nProc ) ) bAccept = false;
					}
					if( bAccept ) {
						Mutex m = null;
						bool bAdded = false;
						bool bContinue = false;
						lock( m_oProcConnectingLock ) {
							if( m_hProcConnecting.ContainsKey( nProc ) ) {
								m = (Mutex)m_hProcConnecting[nProc];
							} else {
								m = new Mutex();
								m_hProcConnecting.Add( nProc, m );
							}
						}
						if( !m.WaitOne( 0, false ) ) {
							if( m_nMyProcId >= nProc ) {
								sw.WriteLine( "YaccaiPort 51" );
								sw.Flush();
								s.Close();
								while( true ) {
									if( m_hProcSock.ContainsKey( nProc ) ) break;
								}
								return;
							} else {
								if( !m_hProcSock.ContainsKey( nProc ) ) {
									m.WaitOne();
									bContinue = true;
									if( m_hProcSock.ContainsKey( nProc ) ) {
										m.ReleaseMutex();
										bContinue = false;
									}
								}
							}
						} else {
							if( !m_hProcSock.ContainsKey( nProc ) )
								bContinue = true;
							else m.ReleaseMutex();
						}
						if( !bContinue ) {
							sw.WriteLine( "YaccaiPort 51" );
							sw.Flush();
							s.Close();
							m.ReleaseMutex();
							return;
						}
						lock( s ) {
							lock( m_oProcSockLock ) {
								if( !m_hProcSock.ContainsKey( nProc ) ) {
									m_hProcSock.Add( nProc, s );
									bAdded = true;
									Thread th_msgListening = new Thread( new ParameterizedThreadStart( this.MessageListening ) );
									th_msgListening.Name = "YaccaiMsgWaiting" + DateTime.Now.ToLongTimeString();
									th_msgListening.Start( nProc );
								}
							}
							if( bAdded ) {
								sw.WriteLine( "YaccaiPort 50" );
								sw.Flush();
							} else {
								sw.WriteLine( "YaccaiPort 51" );
								sw.Flush();
								s.Close();
							}
						}
						m.ReleaseMutex();
					} else {
						sw.WriteLine( "YaccaiPort 51" );
						sw.Flush();
						s.Close();
					}
				}
			}
		}

		private bool WaitForConnections()
		{
			try {
				//Console.WriteLine( "***** Debug: WaitForConnection" );
				//Thread.Sleep( 3000 );
				Socket listenSock = new Socket( AddressFamily.InterNetwork, SocketType.Stream,
												ProtocolType.Tcp );
				//Console.WriteLine( "***** Debug: create listening socket" );
				listenSock.Bind( new IPEndPoint( IPAddress.Any, 0 ) );
				listenSock.Listen( 10 );
				IPEndPoint localAddr = (IPEndPoint)listenSock.LocalEndPoint;
				m_nMyPort = localAddr.Port;
				m_strMyHost = Environment.UserDomainName;
				//Console.WriteLine( "Debug: MyHost = {0}, MyPort = {1}", m_strMyHost, m_nMyPort );
				m_listenerReadyEvent.Set();
				while( true ) {
					Socket s = listenSock.Accept();
					Thread th_processConnection = new Thread( new ParameterizedThreadStart( this.ProcessConnection ) );
					th_processConnection.Name = "YaccaiProcessConnection" + DateTime.Now.ToLongTimeString();
					th_processConnection.Start( s );
				}
			} catch( SocketException se ) {
				Console.WriteLine( "In function WaitForConnection @ YMpi.cs:" );

				Console.WriteLine( se.Message );
				Environment.Exit( 0 );
				return false;
			}
		}
		public void ConnectionListeningThread()
		{
			WaitForConnections();
		}
		private bool EstablishConnections()
		{
			//Console.WriteLine( "--->>>>> EstablishConnection" );
			try {
				//Console.WriteLine( "Debug: EstablishConnection" );
				//Console.WriteLine( "Debug: portMgrAddr = {0}", m_strPortMgrAddr );
				//Thread.Sleep( 3000 );
				Console.Out.Flush();
				TcpClient tcpClient = new TcpClient( m_strPortMgrAddr, m_nPortMgrPort );
				NetworkStream stream = tcpClient.GetStream();
				StreamReader sr = new StreamReader( stream );
				StreamWriter sw = new StreamWriter( stream );
				
				if( m_listenerReadyEvent.WaitOne() ) {
					sw.WriteLine( "YaccaiPort 210 " + m_nMyProcId + " " + m_strMyHost + " " + m_nMyPort );
					sw.Flush();
				}
				sr.ReadLine();
				//Console.WriteLine( "Debug: sending YaccaiPort 400 to server ..." );
				sw.WriteLine( "YaccaiPort 400" );
				sw.Flush();
				string strRet = sr.ReadLine();
				//Console.WriteLine( "{0}) Debug: received " + strRet, m_nMyProcId );
				string[] arrRet = Regex.Split( strRet, @"\s+" );
				if( arrRet.Length < 3 ) {
					Console.WriteLine( "Invalid proc-port list from server" );
					return false;
				}
				int nResponseCode = Int32.Parse( arrRet[1] );
				if( nResponseCode == 400 ) {
					int nLen = Int32.Parse( arrRet[2] );
					int i = 3;
					int nProc, nPort;
					string strHost;
					while( i < arrRet.Length - 2 ) {
						nProc = Int32.Parse( arrRet[i] );
						strHost = arrRet[i + 1];
						nPort = Int32.Parse( arrRet[i + 2] );
						m_hProcAddress.Add( nProc, new YProcAddress( strHost, nPort ) );
						i += 3;
					}
				} else return false;
				foreach( int nProcId in m_hProcAddress.Keys ) {
					if( nProcId >= m_nMyProcId ) {
						continue;
					}
					YProcAddress yProcAddr = (YProcAddress)m_hProcAddress[nProcId];
					Socket sk = new Socket( AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp );
					
					if( yProcAddr.m_nPort > 0 ) {
						lock( m_oProcSockLock ) {
							if( m_hProcSock.ContainsKey( nProcId ) ) continue;
						}
						lock( m_oProcConnectingLock ) {
							if( !m_hProcConnecting.ContainsKey( nProcId ) ) {
								m_hProcConnecting.Add( nProcId, new Mutex() );
							} else continue;
						}
						Mutex m = (Mutex)m_hProcConnecting[nProcId];
						m.WaitOne();
						if( m_hProcSock.ContainsKey( nProcId ) ) {
							m.ReleaseMutex();
							continue;
						}
						//Console.WriteLine( "Est: Connecting to {0}:{1}", yProcAddr.m_strHostName, yProcAddr.m_nPort );
						//sk.Connect( yProcAddr.m_strHostName, yProcAddr.m_nPort );
						SocketConnect( sk, yProcAddr.m_strHostName, yProcAddr.m_nPort );
						
						NetworkStream nsSock = new NetworkStream( sk );
						StreamWriter swSock = new StreamWriter( nsSock );
						StreamReader srSock = new StreamReader( nsSock );
						//Console.WriteLine( "{0}) sending YaccaiPort 50", m_nMyProcId );
						swSock.WriteLine( "YaccaiPort 50 " + m_nMyProcId );
						swSock.Flush();
						string strResponse50 = srSock.ReadLine();
						//Console.WriteLine( "{0}) received = {1}", m_nMyProcId, strResponse50 );
						if( strResponse50.Contains( "YaccaiPort 50" ) ) {
							lock( m_oProcSockLock ) {
								m_hProcSock.Add( nProcId, sk );

								Thread th_msgListening = new Thread( new ParameterizedThreadStart( this.MessageListening ) );
								th_msgListening.Name = "YaccaiMsgListening1_" + DateTime.Now.ToLongTimeString();
								th_msgListening.Start( nProcId );
							}
						} else {
							sk.Close();
						}
						m.ReleaseMutex();
						//TcpClient tClient = new TcpClient( yProcAddr.m_strHostName, yProcAddr.m_nPort );
					}
					
				}
				sw.WriteLine( "YaccaiPort 10" );
				sw.Flush();
				sw.Close();
				sr.Close();
				tcpClient.Close();
				return true;
			} catch( SocketException se ) {
				Console.WriteLine( "YaccaiError: " );
				Console.WriteLine( "In function EstablishConnection @ YMpi.cs:" );

				Console.WriteLine( se.Message );
				return false;
			}
		}

		public YaccaiVal GetCommunicator( ArrayList argLst, YEnv env )
		{
			if( !m_bAgent ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "GetCommunicator only allowed in agentclass" );
			}

			YaccaiVal commName, listener, retVal = null;
			if( argLst.Count != 3 ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid argument number for GetCommunicator" );
			}
			commName = (YaccaiVal)argLst[1];
			listener = (YaccaiVal)argLst[2];
			if( commName.m_type != YaccaiType.OBJECT ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid argument for GetCommunicator" );
			}
			YObject yObj = (YObject)commName.m_contents;
			object tContents = yObj.GetMemberEx( "2contents", env );
			if( tContents == null ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid argument for GetCommunicator" );
			}
			string strCommName = (string)tContents;
			if( m_hCommYComObj.ContainsKey( strCommName ) ) return (YaccaiVal)m_hCommYComObj[strCommName];
			else {
				Hashtable insVars = new Hashtable();
				ArrayList parentOjbs = new ArrayList();
				parentOjbs.Clear();
				insVars.Clear();
				insVars.Add( "Listener", listener );
				insVars.Add( "Name", NativeObjects.YStringObject( strCommName ) );
				parentOjbs.Add( NativeObjects.YObject() );
				YaccaiVal v = new YaccaiVal( YaccaiType.OBJECT, new YObject( (YClass)( NativeObjects.YCommunicatorClass().m_contents ), insVars, parentOjbs ) );
				m_hCommYComObj.Add( strCommName, v );
				retVal = v;
			}
			if( !m_bMsgSystemInited ) {
				InitMsgSystem();
			}
			return retVal;
		}

		private bool InitMsgSystem()
		{
			//Console.WriteLine( "Debug: InitMsgSystem ...: pma = {0}", m_strPortMgrAddr );
			Thread th = new Thread( new ThreadStart( ConnectionListeningThread ) );
			th.Name = "YaccaiInitMsgSystem";
			th.Start();
			//Console.WriteLine( "Debug: ConnectionListening thread started, calling Establish connection ..." );
			
			if( !EstablishConnections() ) {
				return false;
			}
			 
			m_bMsgSystemInited = true;
			return true;
		}
		private YaccaiVal DeliverMessage( YEnv env )
		{
			/*
			if( !m_msgArrivedEvent.WaitOne( 0, true ) ) {
				if( !m_bIdleWaitForMsg ) return new YaccaiVal( YaccaiType.VOID, null );
			}
			 * */
			if( !m_bAgent ) return new YaccaiVal( YaccaiType.VOID, null );
			//Console.WriteLine( "Agent{0}: DeliverMessage", m_nMyProcId );
			if( env.LookupClass( "Array" ).m_type == YaccaiType.NOT_FOUND ) {
				return new YaccaiVal( YaccaiType.VOID, null );
			}
			//Console.WriteLine( "Agent{0}: DeliverMessage, locking hProcRemoteEval...", m_nMyProcId );
			//lock( m_hProcRemoteEvalExpr ) {
			/*	
				foreach( int iiii in m_hProcRemoteEvalExpr.Keys ) {
					LinkedList<RemoteEvalExpr> reexl = m_hProcRemoteEvalExpr[iiii];
					if( reexl != null ) {
						//Console.WriteLine( "Agent{0}: locking RemoteList[{1}]", m_nMyProcId, iiii );
						lock( reexl ) {
							//Console.WriteLine( "Agent{0}: locked RemoteList[{1}]", m_nMyProcId, iiii );
							foreach( RemoteEvalExpr reex in reexl ) {
								ProcessReduceRequest( reex );
							}
							reexl.Clear();
						}
						//Console.WriteLine( "Agent{0}: unlocked RemoteList[{1}]", m_nMyProcId, iiii );
					}
				}
			//}
			*/ 
			YaccaiVal thisVal = env.Lookup( "2this" );
			YaccaiVal commObj = null;
			YObject yObj = null;
			YaccaiVal listenerVal = null;
			IntrinsicFunc iFunc = null;
			LinkedList<YMpiMsg> msgList = null;
			YMpiMsg msg = null;
			ArrayList argsLst = null;
			while( true ) {
				int nCount = 0;
				foreach( string strCommName in m_hCommYComObj.Keys ) {
					if( !m_hCommMsgList.ContainsKey( strCommName ) ) continue;
					msgList = (LinkedList<YMpiMsg>)m_hCommMsgList[strCommName];
					if( msgList.Count <= 0 ) continue;
					commObj = (YaccaiVal)m_hCommYComObj[strCommName];
					yObj = (YObject)commObj.m_contents;
					listenerVal = (YaccaiVal)yObj.m_instanceVars["Listener"];
					iFunc = (IntrinsicFunc)listenerVal.m_contents;
					lock( msgList ) {
						msg = msgList.First;
						msgList.RemoveFirst();
						nCount += msgList.Count;
					}
					// invoke MsgListener
					argsLst = new ArrayList();
					argsLst.Add( NativeObjects.YMsgObject( msg.m_strContents, msg.m_nSrc, msg.m_nDst, msg.m_strCommName ) );
					m_callStack.Push( new StackTraceItem( new SrcPos( "YMpi.cs", 0, 0 ), iFunc.m_strName ) );
					YEnv new_env = new YEnv( env.GetTopParent() );
					if( env.LookupClass("Array").m_type == YaccaiType.NOT_FOUND ) {
						env.PrintBindings();
						RuntimeError( new SrcPos( "YMpi.cs", 0, 0 ), "Class Array not found?" );
					}
					InvokeYaccaiMethod( iFunc.m_this, iFunc.m_strName, argsLst, new_env, new SrcPos( "YMpi.cs", 0, 0 ) );
					m_callStack.Pop();
				}
				if( !m_bIdleWaitForMsg ) {
					return new YaccaiVal( YaccaiType.VOID, null );
				} else {
					if( nCount <= 0 ) {
						//Console.WriteLine( "Wait for message..." );
						m_msgArrivedEvent.WaitOne();
						//Console.WriteLine( "msgArrived..." );
					}
				}
			}

		}
		public YaccaiVal Comm_Send( ArrayList argList,  YEnv env )
		{
			YaccaiVal commVal, dstVal, msgVal;
			
			if( argList.Count < 3 ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid number of argument for Send" );
			}
			commVal = (YaccaiVal)argList[0];
			dstVal = (YaccaiVal)argList[1];
			msgVal = (YaccaiVal)argList[2];

			if( dstVal.m_type != YaccaiType.INT ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid source value" );
			}
			object strObj = ( (YObject)( msgVal.m_contents ) ).GetMemberEx( "2contents", env );
			if( strObj == null ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid message value" );
			}
			string strMsg = (string)strObj;
			int nDst = (int)dstVal.m_contents;
			object commNameObj = ((YObject)( (YObject)( commVal.m_contents ) ).MemberLookup( "Name", env ).m_contents).GetMemberEx( "2contents", env );
			string strCommName = (string)commNameObj;
			int nSrc = m_nMyProcId;
			if( nDst == m_nMyProcId ) {
				YMpiMsg msg = new YMpiMsg( nSrc, nDst, strCommName, strMsg );
				LinkedList<YMpiMsg> msgList = null;
				lock( m_oCommMsgListLock ) {
					if( !m_hCommMsgList.ContainsKey( strCommName ) ) {
						msgList = new LinkedList<YMpiMsg>();
						m_hCommMsgList.Add( strCommName, msgList );
					} else msgList = (LinkedList<YMpiMsg>)m_hCommMsgList[strCommName];
				}
				
				lock( msgList ) {
					msgList.Add( msg );
				}
				return new YaccaiVal( YaccaiType.INT, 1 );
			}
			if( !m_hProcSock.ContainsKey( nDst ) ) {
				bool bConnected = false;
				do {
					bConnected = ConnectTo( nDst );
				} while( !bConnected );
			}
			while( true ) {
				lock( m_oProcSockLock ) {
					if( m_hProcSock.ContainsKey( nDst ) ) break;
				}
			}
			if( m_hProcSock.ContainsKey( nDst ) ) {
				Socket s = (Socket)m_hProcSock[nDst];
				//Console.WriteLine( "Agent{0}: locking socket[{1}]", m_nMyProcId, nDst );
				if( IsYaccaiMainThread() && m_bMainThreadExecutionTurn ) {
					m_bMainThreadExecutionTurn = false;
					m_mutexExecutionTurn.ReleaseMutex();
				}
				lock( s ) {
					//Console.WriteLine( "Agent{0}: locked socket[{1}]", m_nMyProcId, nDst );
					NetworkStream ns = new NetworkStream( s );
					StreamWriter sw = new StreamWriter( ns );
					string strToSend = strCommName + "!0!" + nSrc + "!" + nDst + "!" + strMsg + "!";
					//Console.WriteLine( "{0} sending to {1}: {2}", m_nMyProcId, nDst, strToSend );
					//Console.WriteLine( "Agent{0}: wait for CommSend.Send", m_nMyProcId );
					sw.WriteLine( strToSend );
					sw.Flush();
				}
				if(  IsYaccaiMainThread() && !m_bMainThreadExecutionTurn  ) {
					//Console.WriteLine( "Agent{0}: wait to complete CommSend", m_nMyProcId );
					m_mutexExecutionTurn.WaitOne();
					m_bMainThreadExecutionTurn = true;
				}				
				//Console.WriteLine( "Agent{0}: unlocked socket[{1}]", m_nMyProcId, nDst );
				return new YaccaiVal( YaccaiType.INT, 1 );
			}
			return new YaccaiVal( YaccaiType.INT, 0 );
		}

		public YaccaiVal Comm_Bcast( ArrayList argList, YEnv env )
		{
			if( argList.Count != 2 ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid number of argument for Bcast" );
			}
			YaccaiVal v1 = (YaccaiVal)argList[0];
			YaccaiVal v2 = (YaccaiVal)argList[1];
			if( v1.m_type != YaccaiType.OBJECT || v2.m_type != YaccaiType.OBJECT ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid type for Bcast" );
			}
			object commNameObj = ( (YObject)( (YObject)( v1.m_contents ) ).MemberLookup( "Name", env ).m_contents ).GetMemberEx( "2contents", env );
			string strCommName = (string)commNameObj;
			string strContents = (string)( (YObject)v2.m_contents ).GetMemberEx( "2contents", env );
			try {
				
				foreach( int nProcId in m_hProcSock.Keys ) {
					Socket sk = (Socket)m_hProcSock[nProcId];
					if( IsYaccaiMainThread() && m_bMainThreadExecutionTurn ) {
						m_bMainThreadExecutionTurn = false;
						m_mutexExecutionTurn.ReleaseMutex();
					}

					lock( sk ) {
						NetworkStream ns = new NetworkStream( sk );
						StreamWriter sw = new StreamWriter( ns );
						string strToSend = strCommName + "!1!" + m_nMyProcId + "!" + nProcId + "!" + strContents + "!";
						//Console.WriteLine( "Agent{0}: wait for Bcast.Send", m_nMyProcId );
						sw.WriteLine( strToSend );
						sw.Flush();
					}
					if(  IsYaccaiMainThread() && !m_bMainThreadExecutionTurn ) {
						//Console.WriteLine( "Agent{0}: wait to complete Bcast.Send", m_nMyProcId );
						m_mutexExecutionTurn.WaitOne();
						m_bMainThreadExecutionTurn = true;
						//Console.WriteLine( "Agent{0}: completed Bcast.Send", m_nMyProcId );
					}

				}
			} catch( Exception ex ) {
				Console.WriteLine( "In function CommBcast @ YMpi.cs:" );

				Console.WriteLine( ex.Message );
				return new YaccaiVal( YaccaiType.INT, 0 );
			}
			return new YaccaiVal( YaccaiType.INT, 1 );
		}

		public YaccaiVal Comm_Reduce( ArrayList args, YEnv env )
		{
			if( args.Count != 3 ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid argument number for Reduce" );
			}
			YaccaiVal v1, v2, v3;
			v1 = (YaccaiVal)args[0];
			v2 = (YaccaiVal)args[1];
			v3 = (YaccaiVal)args[2];
			if( v1.m_type != YaccaiType.OBJECT ||
				v2.m_type != YaccaiType.OBJECT ||
				v3.m_type != YaccaiType.OBJECT ) {
				return new YaccaiVal( YaccaiType.RUNTIME_ERROR, "Invalid argument type for Reduce" );
			}
			string strTargetData = (string)( (YObject)v2.m_contents ).GetMemberEx( "2contents", env );
			string strReduceOperator = (string)( (YObject)v3.m_contents ).GetMemberEx( "2contents", env );
			object commNameObj = ( (YObject)( (YObject)( v1.m_contents ) ).MemberLookup( "Name", env ).m_contents ).GetMemberEx( "2contents", env );
			string strCommName = (string)commNameObj;
			m_nMaxCollectiveIdentifier++;
			int nReduceId = m_nMaxCollectiveIdentifier;
			YaccaiVal opVal = ExprValFromString( strReduceOperator, env.GetTopParent() );
			ReduceResult reduceRet = new ReduceResult( opVal );
			reduceRet.m_nProcId = m_nMyProcId;
			m_hCollectiveIdReduceObj.Add( nReduceId, reduceRet );
			foreach( int nProcId in m_hProcSock.Keys ) {
				Socket sk = (Socket)m_hProcSock[nProcId];
				lock( reduceRet.m_pendingProcIdSet ) {
					reduceRet.m_pendingProcIdSet.Add( nProcId );
				}
				if( IsYaccaiMainThread() && m_bMainThreadExecutionTurn ) {
					m_bMainThreadExecutionTurn = false;
					m_mutexExecutionTurn.ReleaseMutex();
				}
				
				lock( sk ) {
					NetworkStream ns = new NetworkStream( sk );
					StreamWriter sw = new StreamWriter( ns );
					string strToSend = strCommName + "!2!" + m_nMyProcId + "!" + nProcId + "!" + 
						m_nMaxCollectiveIdentifier + "!" + strTargetData + "!" + strReduceOperator + "!";
					sw.WriteLine( strToSend );
					sw.Flush();
				}
				if( IsYaccaiMainThread() && !m_bMainThreadExecutionTurn ) {
					m_mutexExecutionTurn.WaitOne();
					m_bMainThreadExecutionTurn = true;
				}
				
			}
			
			reduceRet.m_bAllRequestSent = true;
			bool bReady = false;
			if( IsYaccaiMainThread() && m_bMainThreadExecutionTurn ) {
				m_bMainThreadExecutionTurn = false;
				m_mutexExecutionTurn.ReleaseMutex();
			}

			lock( reduceRet.m_pendingProcIdSet ) {
				if( reduceRet.m_pendingProcIdSet.Count == 0 ) bReady = true;
			}
			//Console.WriteLine( "Agent{0}: sent reduce request, wait for data", m_nMyProcId );
			if( !bReady ) {
				if( reduceRet.m_readyEvent.WaitOne() ) {
					bReady = true;
				}
			}
			//Console.WriteLine( "Agent{0}: reduce: received all data, wait turn", m_nMyProcId );
			if( IsYaccaiMainThread() && !m_bMainThreadExecutionTurn ) {
				m_mutexExecutionTurn.WaitOne();
				m_bMainThreadExecutionTurn = true;
			}
			//Console.WriteLine( "Agent{0}: applying reduce operator", m_nMyProcId );
			if( bReady ) {
				IntrinsicFunc iFunc = (IntrinsicFunc)opVal.m_contents;
				YaccaiVal curVal = ExprValFromString( strTargetData, env.GetTopParent() );
				foreach( int vId in reduceRet.m_valSet.Keys ) {
					YaccaiVal v = (YaccaiVal)reduceRet.m_valSet[vId];
					ArrayList iFuncArgs = new ArrayList();
					iFuncArgs.Add( curVal );
					if( v == null ) {
						Console.WriteLine( "Error: reduce parameter is null" );
						Environment.Exit( 0 );
					}
					iFuncArgs.Add( v );
					YEnv new_env = new YEnv( env.GetTopParent() );
					if( iFunc.m_bStatic ) {
						curVal = InvokeStaticMethod( iFunc.m_this, iFunc, iFuncArgs, new_env, new SrcPos( "YMpi.cs:Reduce", 0, 0 ) );
					} else {
						curVal = InvokeYaccaiMethod( iFunc.m_this, iFunc.m_strName, iFuncArgs, new_env, new SrcPos( "YMpi.cs:Reduce", 0, 0 ) );
					}
				}
				m_hCollectiveIdReduceObj.Remove( nReduceId );
				return curVal;
			}
			//Console.WriteLine( "Agent{0}: end reduction", m_nMyProcId );
			return new YaccaiVal( YaccaiType.NULL, null );
		}
	}
}
