using System;
using System.Collections;
using System.Text;
namespace Yaccai
{
	class YEnv
	{
		private Hashtable m_bindings;
		private YEnv m_parent;
		public YEnv()
		{
			m_parent = null;
			m_bindings = new Hashtable( 100, 0.6f );
			m_bindings.Clear();
		}
		public YEnv( YEnv parent )
		{
			m_parent = parent;
			m_bindings = new Hashtable( 100, 0.6f );
			m_bindings.Clear();
		}
		public YaccaiVal Lookup( string strName )
		{
			YaccaiVal yV = null;
			if( strName.Equals( "null" ) ) return new YaccaiVal( YaccaiType.NULL );
			if( m_bindings.ContainsKey( strName ) ) {
				return (YaccaiVal)m_bindings[strName];
			}
			// "this" should be changed into "2this" to avoid identifier "this" from user input
			if( m_bindings.ContainsKey( "2this" ) ) {
				YaccaiVal thisVal = (YaccaiVal)m_bindings["2this"];
				if( thisVal.m_type != YaccaiType.NOT_FOUND )  {
					if( thisVal.m_type == YaccaiType.OBJECT ) {
						YObject yObj = (YObject)thisVal.m_contents ;
						if( yObj == null ) return new YaccaiVal( YaccaiType.NOT_FOUND );
						yV = yObj.MemberLookup( strName, null );
						if( yV.m_type != YaccaiType.NOT_FOUND ) {
							return yV;
						}
					} else {
						if( thisVal.m_type == YaccaiType.CLASS ) {
							IYMemberSet memberSet = (IYMemberSet)thisVal.m_contents;
							yV = memberSet.MemberLookup( strName, this );
							if( yV.m_type != YaccaiType.NOT_FOUND ) return yV;
						}
					}
				}
			}
			if( m_parent != null ) {
				return m_parent.Lookup( strName );
			}
			return new YaccaiVal( YaccaiType.NOT_FOUND );
		}
		public void Set( string strName, YaccaiVal val )
		{
			YaccaiVal v = Lookup( strName );
			if( v == null || v.m_type == YaccaiType.NOT_FOUND ) {
				this.m_bindings.Add( strName, val );
			} else {
				v.m_accessType = val.m_accessType;
				v.m_contents = val.m_contents;
				v.m_type = val.m_type;
			}
		}
		public void Bind( string strName, YaccaiVal val )
		{
			if( m_bindings.ContainsKey( strName ) )
				m_bindings.Remove( strName );
			m_bindings.Add( strName, val );
		}
		public YEnv GetParent()
		{
			return m_parent;
		}
		public YaccaiVal LookupClass( string strName )
		{
			YaccaiVal retVal = null;
			if( m_bindings.ContainsKey( strName ) ) {
				retVal = (YaccaiVal)m_bindings[strName];
				if( retVal.m_type == YaccaiType.CLASS || retVal.m_type == YaccaiType.AGENT )
					return retVal;
			}
			if( this.m_parent != null ) return this.m_parent.LookupClass( strName );
			return new YaccaiVal( YaccaiType.NOT_FOUND, null );
		}
		public YEnv GetTopParent()
		{
			if( m_parent != null ) return m_parent.GetTopParent();
			return this;
		}
		public Hashtable GetBindings()
		{
			return m_bindings;
		}
		public void PrintBindings()
		{
			Console.Write( "[" );
			foreach( Object obj in m_bindings.Keys ) {
				String k = (String)obj;
				Console.Write( k + ", " );
			}
			if( m_parent != null ) m_parent.PrintBindings();
			Console.Write( "] " );
		}
	}
}
