using System;
using System.Collections;
using Antlr.Runtime;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Text.RegularExpressions;

namespace Yaccai
{
	using CallStack = Stack<StackTraceItem>;
	using StmtList = LinkedList<Stmt>;
	class StackTraceItem
	{
		public SrcPos m_srcPos;
		public string m_funcName;
		public StackTraceItem( SrcPos p, string strName )
		{
			m_srcPos = p;
			m_funcName = strName;
		}
	}

	class ExprEnvPair
	{
		public Expr m_expr;
		public YEnv m_env;
		public ExprEnvPair( Expr e, YEnv env )
		{
			m_expr = e;
			m_env = env;
		}
	}
	partial class YEvaluator
	{
		private CallStack m_callStack;
		private IntrinsicFunc m_mainFunc;
		private ArrayList m_strArgs;
		private Stack<int> m_stkBreakLevel;
		private Stack<int> m_stkContinueLevel;
		private Hashtable m_belief;
		private bool m_bAgent;
		private bool m_bMsgSystemInited;
		private Stack<bool> m_stkCheckCondition;
		private Stack<ExprEnvPair> m_stkMaintainCond;

		private ArrayList m_arrChildProcesses;
		private YEnv m_gYEnv;			// global environment
		private AutoResetEvent m_gYEnvReadyEvent;
		private bool m_bEvalRemote;
		
		// private bool m_bSuspendDeliverMessage;
		private Mutex m_mutexExecutionTurn;
		private bool m_bMainThreadExecutionTurn;
	
		public delegate YaccaiVal YDelFunc( ArrayList y_args, YEnv env );

		public YEvaluator( string[] args )
		{
			m_callStack = new Stack<StackTraceItem>();
			m_callStack.Clear();
			m_mainFunc = null;
			m_strArgs = new ArrayList();
			m_strArgs.Clear();
			m_stkContinueLevel = new Stack<int>();
			m_stkBreakLevel = new Stack<int>();
			m_stkBreakLevel.Clear();
			m_stkContinueLevel.Clear();
			m_belief = new Hashtable();
			m_belief.Clear();
			m_gYEnv = null;
			if( args != null ) {
				foreach( string str in args ) {
					m_strArgs.Insert( 0, str );
				}
			}
			Init();
		}
		public YEvaluator( ArrayList argList )
		{
			m_callStack = new Stack<StackTraceItem>();
			m_callStack.Clear();
			m_mainFunc = null;
			m_strArgs = new ArrayList();
			m_strArgs.Clear();
			m_stkContinueLevel = new Stack<int>();
			m_stkBreakLevel = new Stack<int>();
			m_stkBreakLevel.Clear();
			m_stkContinueLevel.Clear();
			m_belief = new Hashtable();
			m_belief.Clear();			
			m_strArgs = argList;
			m_bAgent = false;
			m_gYEnv = null;
			Init();
		}
		private void Init()
		{
			m_stkCheckCondition = new Stack<bool>();
			m_stkMaintainCond = new Stack<ExprEnvPair>();
			m_arrChildProcesses = new ArrayList();
			m_gYEnvReadyEvent = new AutoResetEvent( false );
			m_nIOConnectorPort = 11022;	// default port
			
			m_bEvalRemote = false;
			Thread.CurrentThread.Name = "YaccaiMainThread";
			m_mutexExecutionTurn = new Mutex( false );
			m_bMainThreadExecutionTurn = false;
		}
		private bool IsYaccaiMainThread()
		{
			if( Thread.CurrentThread.Name == null ) return false;
			return Thread.CurrentThread.Name.Equals( "YaccaiMainThread" );
		}
		private void AgentInit()
		{
			m_nMyPort = -1;
			m_strMyHost = "";
			m_listenerReadyEvent = new AutoResetEvent( false );
			m_hProcMsgQueue = new Hashtable();
			m_hCommMsgList = new Hashtable();
			m_oCommMsgListLock = new object();
			m_oProcConnectingLock = new object();
			m_oProcSockLock = new object();
			m_bMsgSystemInited = false;
			m_hCommYComObj = new Hashtable();
			m_msgArrivedEvent = new AutoResetEvent( false );
			m_hProcConnecting = new Hashtable();
			m_bIdleWaitForMsg = false;
			m_nMaxCollectiveIdentifier = 0;
			m_hCollectiveIdReduceObj = new Hashtable();
			m_gYEnv = null;
			if( !InitMsgSystem() ) {
				Console.WriteLine( "Could not init msg system" );
				Environment.Exit( 1 );
			}
		}
		public void SetPortMgrAddr( string strHost, int nPort )
		{
			m_strPortMgrAddr = strHost;
			m_nPortMgrPort = nPort;
		}
		private void PrintStackTrace( CallStack cStk )
		{
			CallStack.Enumerator en = cStk.GetEnumerator();
			StackTraceItem stkI;
			Console.WriteLine( "Stack trace: " );
			while( en.MoveNext() ) {
				stkI = en.Current;
				Console.WriteLine( "{0} : {1}", stkI.m_funcName, stkI.m_srcPos );
			}
		}

		public void RuntimeError( SrcPos p, string strMsg )
		{
			Console.WriteLine( "Runtime error at : {0}, line {1}, col {2}", p.m_strFileName,
								p.m_nLine, p.m_nCol );
			Console.WriteLine( "{0}", strMsg );
			PrintStackTrace( m_callStack );
			Environment.Exit( 1 );
		}

		public void RuntimeError( SrcPos p )
		{
			RuntimeError( p, "" );
		}
		private bool InstanceOfClass( YaccaiVal objVal, string strName, YEnv env )
		{
			switch( objVal.m_type ) {
				case YaccaiType.INT:
					if( strName.Equals( "int" ) ) return true;
					break;
				case YaccaiType.DOUBLE:
					if( strName.Equals( "double" ) ) return true;
					break;
				case YaccaiType.CHAR:
					if( strName.Equals( "char" ) ) return true;
					break;
				case YaccaiType.ARRAY:
					if( strName.Equals( "Array" ) ) return true;
					break;
				case YaccaiType.CLASS:
					if( strName.Equals( "class" ) ) return true;
					break;
				case YaccaiType.OBJECT: {
						YObject yObj = (YObject)objVal.m_contents;
						if( yObj.m_yclass.IsSubClassOf( strName, env ) ) return true;
					}
					break;
				default:
					break;
			}
			return false;
		}

		private ArrayList EvalExprYLinkedList( LinkedList<Expr> arrExpr, YEnv env )
		{
			ArrayList retLst = new ArrayList();
			retLst.Clear();
			YaccaiVal retVal = null;
			foreach( Expr ex in arrExpr ) {
				retVal = EvalExpr( ex, env );
				retLst.Add( retVal );
			}
			return retLst;
		}
		private ArrayList EvalExprList( ArrayList arrExpr, YEnv env )
		{
			ArrayList retLst = new ArrayList();
			retLst.Clear();
			YaccaiVal retVal = null;
			foreach( Expr ex in arrExpr ) {
				retVal = EvalExpr( ex, env );
				retLst.Add( retVal );
			}
			return retLst;
		}
		private YaccaiVal EvalExprFuncCall( Expr e, YEnv env )
		{
			// TODO
			ExprFuncCall fc = (ExprFuncCall)e.Contents;
			YaccaiVal fVal = EvalExpr( fc.m_f, env );
			ArrayList argsLst = EvalExprList( fc.m_args, env );
			YEnv new_env = new YEnv( env.GetTopParent() );
			YaccaiVal retVal = null;
			
			switch( fVal.m_type ) {
				case YaccaiType.CS_METHOD:
					CSharpMethod csMeth = (CSharpMethod)fVal.m_contents;
					m_callStack.Push( new StackTraceItem( e.Pos, csMeth.m_strMethodName ) );
					m_stkBreakLevel.Push( 0 );
					m_stkContinueLevel.Push( 0 );
					m_stkCheckCondition.Push( false );
					
					retVal = csMeth.m_csobj.InvokeMethod( csMeth.m_strMethodName, argsLst );
					m_stkCheckCondition.Pop();
					m_stkBreakLevel.Pop();
					m_stkContinueLevel.Pop();
					m_callStack.Pop();
					break;
				case YaccaiType.IFUNC:
					IntrinsicFunc insFunc = (IntrinsicFunc)fVal.m_contents;
					m_stkBreakLevel.Push( 0 );
					m_stkContinueLevel.Push( 0 );
					m_callStack.Push( new StackTraceItem( e.Pos, insFunc.m_strName ) );
					m_stkCheckCondition.Push( false );
					
					if( insFunc.m_this == null ) {
						new_env.Bind( "2this", env.Lookup( "2this" ) );
					} else {
						new_env.Bind( "2this", insFunc.m_this );
					}
					int i = 0;
					foreach( string strParamName in insFunc.m_paramList ) {
						new_env.Bind( strParamName, (YaccaiVal)argsLst[i] );
						i++;
					}
					retVal = EvalStmtList( insFunc.m_body, new_env );
					m_callStack.Pop();
					m_stkBreakLevel.Pop();
					m_stkContinueLevel.Pop();
					m_stkCheckCondition.Pop();
					if( retVal.m_type == YaccaiType.RETURN ) {
						if( retVal.m_contents != null ) {
							return (YaccaiVal)retVal.m_contents;
						} else return new YaccaiVal( YaccaiType.VOID, null );
					} else return retVal;
				case YaccaiType.NFUNC:
					NativeFunc ntFunc = (NativeFunc)fVal.m_contents;
					if( ntFunc.m_nParams != argsLst.Count + 1 ) {
						RuntimeError( e.Pos, "Invalid argument number for " + ntFunc.m_strName );
					}
					if( ntFunc.m_this != null ) {
						argsLst.Insert( 0, ntFunc.m_this );
					} else {
						argsLst.Insert( 0, env.Lookup( "2this" ) );
					}
					
					m_callStack.Push( new StackTraceItem( e.Pos, ntFunc.m_strName ) );
					m_stkCheckCondition.Push( false );
					
					retVal = ( new CSharpObject( typeof( YNativeFuncs ), null ) ).InvokeMethod( ntFunc.m_strName, argsLst );
					m_callStack.Pop();
					m_stkCheckCondition.Pop();
					if( retVal.m_type == YaccaiType.RUNTIME_ERROR ) {
						RuntimeError( e.Pos, (string)retVal.m_contents );
					}
					break;
				case YaccaiType.IDELEGATE:
					YaccaiDelegate yccDelegate = (YaccaiDelegate)fVal.m_contents;
					YDelFunc yDelFunc = yccDelegate.m_delegate;
					m_callStack.Push( new StackTraceItem( e.Pos, fc.m_f.ToString() ) );
					m_stkCheckCondition.Push( false );
					if( yccDelegate.m_this != null ) {
						argsLst.Insert( 0, yccDelegate.m_this );
					} else {
						argsLst.Insert( 0, env.Lookup( "2this" ) );
					}
					retVal = yDelFunc( argsLst, new YEnv( env.GetTopParent() ) );
					if( retVal.m_type == YaccaiType.RUNTIME_ERROR ) {
						RuntimeError( e.Pos, (string)retVal.m_contents );
					}
					m_callStack.Pop();
					m_stkCheckCondition.Pop();
					break;
				case YaccaiType.IPLAN:
					return EvalYPlan( fVal, argsLst, env, e.Pos  );
			}
			return retVal;
		}
		private YaccaiVal EvalYPlan( YaccaiVal pVal, ArrayList argList, YEnv env, SrcPos p )
		{
			
			IntrinsicPlan iPlan = (IntrinsicPlan)pVal.m_contents;
			
			m_callStack.Push( new StackTraceItem( p, iPlan.m_strName ) );
			m_stkBreakLevel.Push( 0 );
			m_stkContinueLevel.Push( 0 );
			
			YEnv new_env = new YEnv( env.GetTopParent() );
			YaccaiVal thisVal = env.Lookup( "2this" );
			if( thisVal.m_type != YaccaiType.NOT_FOUND ) {
				new_env.Bind( "2this", thisVal );
			}
			// bind
			if( iPlan.m_paramList == null ) iPlan.m_paramList = new ArrayList();
			if( argList.Count != iPlan.m_paramList.Count ) {
				RuntimeError( p, "Invalid argument for Plan execution" );
			}
			int i = 0;
			foreach( string strVarName in iPlan.m_paramList ) {
				new_env.Bind( strVarName, (YaccaiVal)argList[i] );
				i++;
			}
			YaccaiVal preCondVal = new YaccaiVal( YaccaiType.INT, 1 );
			if( iPlan.m_exprPreCond != null ) {
				preCondVal = EvalExpr( iPlan.m_exprPreCond, new_env );
			}
			YaccaiVal maintainCondVal = new YaccaiVal( YaccaiType.INT, 1 );
			YaccaiVal retVal = new YaccaiVal( YaccaiType.INT, 1 );
			if( preCondVal.IsFalseVal() ) {
				retVal = new YaccaiVal( YaccaiType.INT, 0 );
			} else {
				m_stkMaintainCond.Push( new ExprEnvPair( iPlan.m_exprMaintainCond, new_env ) );
				m_stkCheckCondition.Push( true );
				retVal = EvalStmtList( iPlan.m_body, new_env );
				m_stkCheckCondition.Pop();
				m_stkMaintainCond.Pop();
				
				/*
				foreach( Stmt st in iPlan.m_body ) {
					if( iPlan.m_exprMaintainCond != null ) {
						maintainCondVal = EvalExpr( iPlan.m_exprMaintainCond, new_env );
					}
					if( maintainCondVal.IsFalseVal() ) {
						retVal = new YaccaiVal( YaccaiType.INT, 0 );
						break;
					}
					retVal = EvalStmt( st, new_env );
					if( retVal.m_type == YaccaiType.RETURN ) {
						if( retVal.m_contents != null ) retVal = (YaccaiVal)retVal.m_contents;
						break;
					}
					if( retVal.m_type == YaccaiType.BREAK || retVal.m_type == YaccaiType.CONTINUE ) {
						retVal = new YaccaiVal( YaccaiType.VOID, null );
					}
					DeliverMessage( env );
				}
				*/ 
			}
			m_callStack.Pop();
			m_stkBreakLevel.Pop();
			m_stkContinueLevel.Pop();
			//m_stkCheckCondition.Pop();
			switch( retVal.m_type ) {
				case YaccaiType.RETURN:
					if( retVal.m_contents != null ) return (YaccaiVal)retVal.m_contents;
					else return new YaccaiVal( YaccaiType.VOID, null );
				default:
					return retVal;
			}	
		}
		private YaccaiVal EvalExpr( Expr e, YEnv env )
		{
			YaccaiVal retVal = null;
			string strName;
			try {
				switch( e.Type ) {
					case ExprType.LiteralChar:
						return new YaccaiVal( YaccaiType.CHAR, e.Contents );
					case ExprType.LiteralInt:
						return new YaccaiVal( YaccaiType.INT, e.Contents );
					case ExprType.LiteralDouble:
						return new YaccaiVal( YaccaiType.DOUBLE, e.Contents );
					case ExprType.LiteralString:
						string strContents = (string)e.Contents;
						strContents = strContents.Remove( 0, 1 );
						strContents = strContents.Remove( strContents.Length - 1, 1 );
						return NativeObjects.YStringObject( (string)strContents );
					case ExprType.Var:
						strName = (string)e.Contents;
						if( strName.Equals( "this" ) ) strName = "2this";
						retVal = env.Lookup( strName );
						if( retVal.m_type == YaccaiType.NOT_FOUND ) {
							RuntimeError( e.Pos, "Name '" + (string)e.Contents + "' not found" );
						}
						break;
					case ExprType.Assignment:
						return EvalExprAssignment( e, env );
					case ExprType.AttributeRef:
						return EvalExprAttributeRef( e, env );
					case ExprType.FuncCall:
						return EvalExprFuncCall( e, env );
					case ExprType.Operator:
						return EvalExprOperator( e, env );
					case ExprType.ObjectAllocation:
						return EvalExprNew( e, env );
					case ExprType.Subscript:
						return EvalExprSubscript( e, env );
					case ExprType.BeliefExpr:
						return EvalExprBelief( e, env );
					case ExprType.Create:
						return EvalExprCreate( e, env );
					case ExprType.Expression:
						return EvalExpr( (Expr)e.Contents, env );
					case ExprType.StopEvalExpr:
						return NativeObjects.YStringObject( ( (Expr)e.Contents ).ToString() );
					default:
						RuntimeError( e.Pos, "Unknown expression type" );
						break;
				}
			} catch( Exception exc ) {
				Console.WriteLine( exc.Message );
				PrintStackTrace( m_callStack );
				if( e != null ) Console.WriteLine( e.ToString() );
			}
			return retVal;
		}

		private YaccaiVal EvalExprBelief( Expr e, YEnv env )
		{
			ExprBelief eb = (ExprBelief)e.Contents;
			/*
			if( eb.ToString().Contains( "scanned" ) ) {
				Console.WriteLine( "line, scanned" );
			}
			if( m_belief.ContainsKey( "{dX, dY, dirt, weight, time}" ) ) {
				ArrayList arrItm = (ArrayList)m_belief["{dX, dY, dirt, weight, time}"];
				YaccaiVal yV = (YaccaiVal)arrItm[0];
				YObject yObj = (YObject)yV.m_contents;
				YaccaiVal v = (YaccaiVal)yObj.m_instanceVars["line"];
				if( v.m_type == YaccaiType.INT ) {
					int val = (int)v.m_contents;
					if( val == 3 ) {
						Console.WriteLine( "val = 3 " );
					}
				}
			}
			
			if( eb.m_type == BeliefExprType.Remove ) {
				Hashtable h = m_belief;
				Console.WriteLine( "removing..." );
			}
			**/
			switch( eb.m_type ) {
				case BeliefExprType.AddFact:
					return EvalExprBeliefAddFact( e, env );
				case BeliefExprType.Query:
					return EvalExprBeliefQuery( e, env );
				case BeliefExprType.Remove:
					return EvalExprBeliefRemove( e, env );
				default:
					RuntimeError( e.Pos, "TODO: implement other belief base operation" );
					break;
			}
			return null;
		}
		private YaccaiVal EvalExprBeliefAddFact( Expr e, YEnv env )
		{
			try {
				ExprBelief eb = (ExprBelief)e.Contents;
				YaccaiVal v = EvalExpr( eb.m_expr, env );
				if( v.m_type != YaccaiType.OBJECT ) {
					RuntimeError( eb.m_expr.Pos, "belief fact requires object" );
				}
				YClass yCls = ( (YObject)v.m_contents ).m_yclass;
				ArrayList itmLst = null;
				lock( m_belief ) {
					if( m_belief.ContainsKey( yCls.m_strName ) ) {
						itmLst = (ArrayList)m_belief[yCls.m_strName];
					} else {
						itmLst = new ArrayList();
						itmLst.Clear();
						m_belief[yCls.m_strName] = itmLst;
					}
					lock( itmLst ) {
						itmLst.Add( v );
					}
				}
				return new YaccaiVal( YaccaiType.VOID, null );
			} catch( Exception exc ) {
				Console.WriteLine( exc.Message );
				Console.WriteLine( exc.StackTrace );
				Environment.Exit( 1 );
			}
			return null;
		}
		private YaccaiVal EvalExprBeliefQuery( Expr e, YEnv env )
		{
			try {
				ExprBelief eb = (ExprBelief)e.Contents;
				string strClassName = eb.m_signature.ToString();
				ArrayList retLst = null;
				if( m_belief.ContainsKey( strClassName ) ) {
					retLst = (ArrayList)m_belief[strClassName];
				} else retLst = new ArrayList();
				ExprObjectAllocation eoa = new ExprObjectAllocation( "Array", new LinkedList<Expr>(), new LinkedList<Expr>() );
				Expr eToEval = new Expr( eoa, e.Pos );
				YaccaiVal retVal = EvalExpr( eToEval, env );

				lock( retLst ) {
					foreach( YaccaiVal v in retLst ) {
						ArrayList argLst = new ArrayList();
						argLst.Add( v );
						InvokeYaccaiMethod( retVal, "Add", argLst, env, e.Pos );
					}
				}
				if( eb.m_projection == null && eb.m_filterExpr == null ) {
					return retVal;
				}

				LinkedList<Expr> memberExpr = new LinkedList<Expr>();

				ArrayList resultList = new ArrayList();
				if( eb.m_projection != null ) {
					if( eb.m_projection.m_memberNameList != null ) {
						int i = 0;
						for( i = 0; i < eb.m_projection.m_memberNameList.Count; i++ ) {
							memberExpr.Add( new Expr( ExprType.Var, "null", e.Pos ) );
						}
					}
					YaccaiVal anItem = null;
					lock( retLst ) {
						foreach( YaccaiVal v in retLst ) {
							if( eb.m_projection.m_memberNameList != null ) {
								YaccaiVal retObj = EvalExprNew( new Expr( new ExprObjectAllocation( eb.m_projection.m_memberNameList, memberExpr ), e.Pos ), env );
								YObject yObj = (YObject)v.m_contents;
								foreach( string strFieldName in eb.m_projection.m_memberNameList ) {
									YaccaiVal fieldVal = yObj.MemberLookup( strFieldName, env );
									if( fieldVal.m_type == YaccaiType.NOT_FOUND ) {
										RuntimeError( e.Pos, "Member " + strFieldName + " not found" );
									}
									( (YObject)( retObj.m_contents ) ).m_instanceVars[strFieldName] = fieldVal;
								}
								anItem = retObj;
							} else {
								anItem = v;
							}
							if( eb.m_filterExpr != null ) {
								YEnv new_env = new YEnv( env );
								new_env.Bind( eb.m_projection.m_strObjName, anItem );
								YaccaiVal isResult = EvalExpr( eb.m_filterExpr, new_env );
								if( !isResult.IsFalseVal() ) resultList.Add( anItem );
							}
						}
					}
				} else {
					resultList = retLst;
				}
				ExprObjectAllocation eoa1 = new ExprObjectAllocation( "Array", new LinkedList<Expr>(), new LinkedList<Expr>() );
				Expr eToEval1 = new Expr( eoa1, e.Pos );
				YaccaiVal retVal1 = EvalExpr( eToEval1, env );
				lock( resultList ) {
					foreach( YaccaiVal v in resultList ) {
						ArrayList argLst = new ArrayList();
						argLst.Add( v );
						InvokeYaccaiMethod( retVal1, "Add", argLst, env, e.Pos );
					}
				}
				return retVal1;
			} catch( Exception exc ) {
				Console.WriteLine( exc.Message );
				Console.WriteLine( exc.StackTrace );
				Environment.Exit( 1 );
			}
			return null;
		}
		private YaccaiVal EvalExprBeliefRemove( Expr e, YEnv env )
		{
			ExprBelief eb = (ExprBelief)e.Contents;
			string strClassName = eb.m_signature.ToString();
			YaccaiVal retVal = CreateObjectFromYValues( "Array", new ArrayList(), env );
			YObject retObj = (YObject)retVal.m_contents;
			ArrayList candidateList = null;
			if( m_belief.ContainsKey( strClassName ) ) {
				candidateList = (ArrayList)m_belief[strClassName];
			} else return retVal;
		
			if( eb.m_projection == null && eb.m_filterExpr == null ) {
				lock( m_belief ) {
					m_belief.Remove( strClassName );
				}
				retObj.m_instanceVars["2contents"] = candidateList;
				retObj.m_instanceVars["Length"] = new YaccaiVal( YaccaiType.INT, candidateList.Count );
				return retVal;
			}
			if( eb.m_projection != null ) {
				ArrayList removeIndex = new ArrayList();
				ArrayList projectedCanList = null;
				if( eb.m_projection.m_memberNameList != null ) {
					Hashtable hNameValue = new Hashtable();
					projectedCanList = new ArrayList();
					foreach( string strFieldName in eb.m_projection.m_memberNameList ) {
						hNameValue.Add( strFieldName, null );
					}
					foreach( YaccaiVal v in candidateList ) {
						YObject yObj = (YObject)v.m_contents;
						foreach( string strFieldName in hNameValue.Keys ) {
							hNameValue[strFieldName] = yObj.MemberLookup( strFieldName, env );
						}
						projectedCanList.Add( CreateObjectFromYValues( hNameValue, env ) );
					}
				} else projectedCanList = candidateList;
				if( eb.m_filterExpr != null ) {
					int i = 0;
					for( i = 0; i < projectedCanList.Count; i++ ) {
						YaccaiVal v = (YaccaiVal)projectedCanList[i];
						YEnv new_env = new YEnv( env );
						
						new_env.Bind( eb.m_projection.m_strObjName, v );
						YaccaiVal bNeedRemove = EvalExpr( eb.m_filterExpr, new_env );
						if( !bNeedRemove.IsFalseVal() ) {
							removeIndex.Add( i );
						}
					}
				} else {
					for( int i = 0; i < projectedCanList.Count; i++ )
						removeIndex.Add( i );
				}
				ArrayList resultList = new ArrayList();
				foreach( int i in removeIndex ) {
					resultList.Add( projectedCanList[i] );
				}
				lock( candidateList ) {
					for( int i = removeIndex.Count - 1; i >= 0; i-- ) {
						candidateList.RemoveAt( (int)removeIndex[i] );
					}
				}
				retObj.m_instanceVars["2contents"] = resultList;
				retObj.m_instanceVars["Length"] = new YaccaiVal( YaccaiType.INT, resultList.Count );
				return retVal;
			}
			return null;
		}

		private YaccaiVal EvalExprSubscript( Expr e, YEnv env )
		{
			ExprSubscript esub = (ExprSubscript)e.Contents;
			YaccaiVal baseVal = EvalExpr( esub.m_base, env );
			YaccaiVal indexVal = EvalExpr( esub.m_index, env );
			if( baseVal.m_type != YaccaiType.OBJECT ) {
				RuntimeError( e.Pos, "Invalid subscript access" );
			}
			ArrayList argList = new ArrayList();
			argList.Add( indexVal );
			//Console.WriteLine( "Preparing to eval: {0}", e.ToString() );
			return InvokeYaccaiMethod( baseVal, "operatorSubscript", argList, env, e.Pos );
		}

		private YaccaiVal CreateYObject( YClass yClass, ArrayList paramList, YEnv env, SrcPos p )
		{
			ArrayList parentObjs = new ArrayList();
			parentObjs.Clear();
			if( yClass.m_parentList != null ) {
				foreach( string strClassName in yClass.m_parentList ) {
					YaccaiVal clsVal = env.LookupClass( strClassName );
					if( clsVal.m_type == YaccaiType.NOT_FOUND ) {
						RuntimeError( p, "Class " + strClassName + " not found (in parent list of " + yClass.m_strName + ")" );
					}
					parentObjs.Add( CreateYObject( (YClass)clsVal.m_contents, new ArrayList(), env, p ) );
				}
			}
			YEnv new_env = new YEnv( env.GetParent() );
			Hashtable hInstanceVars = new Hashtable();
			hInstanceVars.Clear();
			new_env.Bind( "2this", new YaccaiVal( YaccaiType.CLASS, yClass ) );
			foreach( Stmt st in yClass.m_stmtList ) {
				if( st.Kind == StmtKind.Expression ) {
					Expr st_e = (Expr)st.Contents;
					if( st_e.Type == ExprType.Assignment ) {
						ExprAssignment eass = (ExprAssignment)st_e.Contents;
						Expr varExp = eass.m_primaryList[0];
						
						string strVarName = (string)varExp.Contents;
						YaccaiAccessType yact = (YaccaiAccessType)yClass.m_memberVarsAccessType[strVarName];
						EvalStmt( st, new_env );
						YaccaiVal memberVal = new_env.Lookup( strVarName );
						memberVal.m_accessType = yact;
						hInstanceVars.Add( strVarName, memberVal );
					}
				}
			}
			YObject yObj = new YObject( yClass, hInstanceVars, parentObjs );
			string strConstructorName = yClass.m_strName + " " + paramList.Count;
			YaccaiVal retVal = new YaccaiVal( YaccaiType.OBJECT, yObj );
			if( yObj.MemberLookup( strConstructorName, env ).m_type != YaccaiType.NOT_FOUND ) {
				InvokeYaccaiMethod( retVal, strConstructorName, paramList, env, p );
			}
			return retVal;
		}
		private YaccaiVal CreateObjectFromYValues( string strClassName, ArrayList argValList, YEnv env )
		{
			// create object from YaccaiVal
			YaccaiVal cls = env.LookupClass( strClassName );
			if( cls.m_type == YaccaiType.NOT_FOUND ) {
				Console.WriteLine( "---- Bindings ---" );
				env.PrintBindings();
				Console.WriteLine( "-----End Bindings ---" );
				RuntimeError( null, "Class " + strClassName + " not found" );
			}
			YClass yCls = (YClass)cls.m_contents;
			YEnv new_env = new YEnv( env.GetTopParent() ); 
			ArrayList parentObjs = new ArrayList();
			parentObjs.Add( NativeObjects.YObject() );
			Hashtable instanceVars = new Hashtable();
			foreach( string strVarName in yCls.m_memberVarsAccessType.Keys ) {
				instanceVars.Add( strVarName, new YaccaiVal( YaccaiType.NULL, null, (YaccaiAccessType)yCls.m_memberVarsAccessType[strVarName] ) );
			}
			YObject yObj = new YObject( yCls, instanceVars, parentObjs );
			YaccaiVal retVal = new YaccaiVal( YaccaiType.OBJECT, yObj );
			new_env.Bind( "2this", retVal );
			EvalStmtList( yCls.m_stmtList, new_env );
			string constructorName = yCls.m_strName + " " + argValList.Count;
			InvokeYaccaiMethod( retVal, constructorName, argValList, new_env, new SrcPos( "native", 0, 0 ) );
			return retVal;
		}
		private YaccaiVal CreateObjectFromYValues( Hashtable memberNameVal, YEnv env )
		{
			
			LinkedList<string> strMemberList = new LinkedList<string>();
			
			foreach( string strName in memberNameVal.Keys ) {
				strMemberList.Add( strName );
			}
			strMemberList.Sort();
			YaccaiClassSignature ySign = new YaccaiClassSignature( strMemberList );
			string strClassName = ySign.ToString();
			YaccaiVal cls = env.LookupClass( strClassName );
			YClass yCls = null;
			if( cls.m_type == YaccaiType.NOT_FOUND ) {
				Hashtable memberAccTypes = new Hashtable();
				foreach( string strName in strMemberList ) {
					memberAccTypes.Add( strName, (object)YaccaiAccessType.Public );
				}
				ArrayList parentList = new ArrayList();
				parentList.Add( "Object" );
				yCls = new YClass( strClassName, parentList, new LinkedList<Stmt>(), new Hashtable(),
									new Hashtable(), memberAccTypes, YaccaiAccessType.Public );
				cls = new YaccaiVal( YaccaiType.CLASS, yCls );
			} else yCls = (YClass)cls.m_contents;
			ArrayList parentObjs = new ArrayList();
			parentObjs.Add( NativeObjects.YObject() );
			Hashtable instanceVars = new Hashtable();
			foreach( string strName in yCls.m_memberVarsAccessType.Keys ) {
				instanceVars.Add( strName, new YaccaiVal( YaccaiType.NULL, null, (YaccaiAccessType)yCls.m_memberVarsAccessType[strName] ) );
			}
			YObject yObj = new YObject( yCls, instanceVars, parentObjs );
			YaccaiVal retVal = new YaccaiVal( YaccaiType.OBJECT, yObj );

			foreach( string strName in memberNameVal.Keys ) {
				YaccaiVal v = (YaccaiVal)memberNameVal[strName];
				YaccaiVal v1 = (YaccaiVal)yObj.m_instanceVars[strName];
				v1.m_contents = v.m_contents;
				v1.m_type = v.m_type;
			}
			return retVal;
		}
		private YaccaiVal EvalExprNew( Expr e, YEnv env )
		{
			ExprObjectAllocation eoa = (ExprObjectAllocation)e.Contents;
			ArrayList exprValList;
			YaccaiVal retVal = null;
			if( eoa.m_strClassName != null && !eoa.m_strClassName.Equals( "" ) ) {
				YaccaiVal clsVal = env.LookupClass( eoa.m_strClassName );
				if( clsVal.m_type == YaccaiType.NOT_FOUND ) {
					Console.WriteLine( "--- Bindings --- " );
					env.PrintBindings();
					Console.WriteLine( "--- End Bindings---" );
					RuntimeError( e.Pos, "Class " + eoa.m_strClassName + " not found" );
				}
				YClass yCls = (YClass)clsVal.m_contents;
				if( eoa.m_memberList == null ) {
					// new constructor( expr, expr, ... )
					exprValList = EvalExprYLinkedList( eoa.m_exprList, env );
					retVal = CreateYObject( yCls, exprValList, env, e.Pos );
					if( eoa.m_bodyLst != null ) {	// eval initialization exprs
						YEnv new_env = new YEnv( env.GetParent() );
						new_env.Bind( "2this", retVal );
						EvalExprYLinkedList( eoa.m_bodyLst, new_env );
					}
					return retVal;
				}
			} else {
				// new { var = expr, var = expr, ... }
				ArrayList parentList = new ArrayList();
				ArrayList parentObjList = new ArrayList();
				
				parentList.Clear();
				parentList.Add( "Object" );
				string strClassName = "{";
				if( eoa.m_memberList != null && eoa.m_memberList.Count > 0 ) {
					ArrayList arrStrVarName = new ArrayList();
					arrStrVarName.Clear();
					foreach( string strVarName in eoa.m_memberList ) {
						arrStrVarName.Add( strVarName );
					}
					arrStrVarName.Sort();
					int i = 0;
					strClassName += arrStrVarName[0];
					for( i = 1; i < arrStrVarName.Count; i++ ) {
						strClassName += ", " + arrStrVarName[i];
					}
				}
				strClassName += "}";
				YClass yCls = null;
				YaccaiVal yClsVal = env.LookupClass( strClassName );
				if( yClsVal.m_type == YaccaiType.NOT_FOUND ) {
					yCls = new YClass( strClassName, parentList, new LinkedList<Stmt>(), new Hashtable(),
												new Hashtable(), new Hashtable(), YaccaiAccessType.Private );
					env.GetTopParent().Bind( strClassName, new YaccaiVal( YaccaiType.CLASS, yCls ) );
				} else {
					yCls = (YClass)yClsVal.m_contents;
				}
				
				parentObjList.Clear();
				parentObjList.Add( NativeObjects.YObject() );
				Hashtable instanceVars = new Hashtable();
				instanceVars.Clear();
				YObject yObj = new YObject( yCls, instanceVars, parentObjList);
				if( eoa.m_exprList != null && eoa.m_memberList != null && eoa.m_exprList.Count > 0 ) {
					YEnv new_env = new YEnv( env );
					new_env.Bind( "2this", new YaccaiVal( YaccaiType.OBJECT, yObj ) );
					int i = 0;
					foreach( Expr memberValExpr in eoa.m_exprList ) {
						YaccaiVal memberVal = EvalExpr( memberValExpr, new_env );
						YaccaiVal myMemberVal = new YaccaiVal( memberVal.m_type, memberVal.m_contents );
						myMemberVal.m_accessType = YaccaiAccessType.Public;
						YaccaiVal curVal = yObj.MemberLookup( eoa.m_memberList[i], new_env );
						if( curVal.m_type == YaccaiType.NOT_FOUND ) {
							yObj.m_instanceVars.Add( eoa.m_memberList[i], myMemberVal );
						} else {
							curVal.m_type = memberVal.m_type;
							curVal.m_contents = memberVal.m_contents;
						}
						i++;
					}
				}
				return new YaccaiVal( YaccaiType.OBJECT, yObj );
			}
			return null;
		}

		private ArrayList EvalOperatorArgumentList( ExprOperator expOp, YEnv env )
		{
			ArrayList retList = new ArrayList();
			foreach( Expr argE in expOp.m_args ) {
				YaccaiVal argVal = EvalExpr( argE, env );
				retList.Add( argVal );
			}
			return retList;
		}
		private YaccaiVal EvalExprOperator( Expr e, YEnv env )
		{
			ExprOperator expOp = (ExprOperator)e.Contents;
			
			switch( expOp.m_type ) {
				case OperatorType.PLUS:
					return EvalOperatorPlus( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.MINUS:
					return EvalOperatorMinus( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.MUL:
					return EvalOperatorMul( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.DIV:
					return EvalOperatorDiv( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.MOD:
					return EvalOperatorMod( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.EQ_EQ:
					return EvalOperatorEqEq( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.DIFF:
					return EvalOperatorDiff( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.GT:
					return EvalOperatorGt( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.GT_EQ:
					return EvalOperatorGtEq( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.LT:
					return EvalOperatorLt( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.LT_EQ:
					return EvalOperatorLtEq( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.UMINUS:
					return EvalOperatorUminus( EvalOperatorArgumentList( expOp, env ), env, e.Pos );
				case OperatorType.AND:
					return EvalOperatorAnd( expOp.m_args, env, e.Pos );
				case OperatorType.OR:
					return EvalOperatorOr( expOp.m_args, env, e.Pos );
				case OperatorType.NOT:
					return EvalOperatorNot( expOp.m_args, env, e.Pos );
				default:
					RuntimeError( e.Pos, "Operator not implemented yet" );
					return null;
			}
		}
		private void CheckAccessRight( YaccaiVal val, YaccaiVal thisVal, SrcPos p )
		{
			// TODO: raise exception here if invalid access occurred
		}
		private YaccaiVal EvalExprAttributeRef( Expr e, YEnv env )
		{
			ExprAttref attf = (ExprAttref)e.Contents;
			YaccaiVal baseVal = EvalExpr( attf.m_base, env );
			if( attf.m_field.Type != ExprType.Var ) {
				RuntimeError( attf.m_field.Pos, "Invalid attribute ref" );
			}
			string strFieldName = (string)attf.m_field.Contents;
			switch( baseVal.m_type ) {
				case YaccaiType.NULL:
					RuntimeError( e.Pos, "Null pointer exception" );
					break;
				case YaccaiType.CSOBJ:
					CSharpObject csObj = (CSharpObject)baseVal.m_contents;
					if( csObj.GetMember( strFieldName ) == null ) {
						RuntimeError( attf.m_field.Pos, "Member " +  strFieldName + " not found" );
					}
					// TODO: also return field (not method) here
					return new YaccaiVal( YaccaiType.CS_METHOD, new CSharpMethod( strFieldName, csObj ) );
				// TODO: other attribute ref: INT, DOUBLE, OBJECT, CLASS, CLASS_LIST
				case YaccaiType.OBJECT: 
				case YaccaiType.CLASS:
				case YaccaiType.AGENT:
					IYMemberSet yObj = (IYMemberSet)baseVal.m_contents;
					YaccaiVal memberInfo = yObj.MemberLookup( strFieldName, env );
					if( memberInfo.m_type == YaccaiType.NOT_FOUND ) {
						RuntimeError( e.Pos, "Member " + strFieldName + " not found" );
					}
					CheckAccessRight( memberInfo, env.Lookup( "2this" ), e.Pos );
					if( memberInfo.m_type == YaccaiType.IFUNC ||
						memberInfo.m_type == YaccaiType.IPLAN ||
						memberInfo.m_type == YaccaiType.NFUNC ||
						memberInfo.m_type == YaccaiType.IDELEGATE ) {

						YaccaiVal methInfo = memberInfo;

						switch( methInfo.m_type ) { // TODO: method_list
							case YaccaiType.IFUNC:
								IntrinsicFunc ifunc = (IntrinsicFunc)methInfo.m_contents;
								IntrinsicFunc newIfunc = new IntrinsicFunc( ifunc.m_strName, ifunc.m_paramList,
																ifunc.m_body, ifunc.m_bStatic, ifunc.m_pos );
								newIfunc.m_this = baseVal;
								methInfo = new YaccaiVal( YaccaiType.IFUNC, newIfunc, memberInfo.m_accessType );
								return methInfo;
							case YaccaiType.NFUNC:
								NativeFunc orgNFunc, retNFunc;
								orgNFunc = (NativeFunc)methInfo.m_contents;
								retNFunc = new NativeFunc( orgNFunc.m_strName, orgNFunc.m_nParams, baseVal );
								retNFunc.m_this = baseVal;
								methInfo = new YaccaiVal( YaccaiType.NFUNC, retNFunc, methInfo.m_accessType );
								return methInfo;
							case YaccaiType.IPLAN:
								IntrinsicPlan iplan = (IntrinsicPlan)methInfo.m_contents;
								IntrinsicPlan retIPlan = new IntrinsicPlan( iplan.m_strName, iplan.m_paramList,
															iplan.m_body, iplan.m_exprPreCond, iplan.m_exprMaintainCond,
															iplan.m_pos );
								retIPlan.m_this = baseVal;
								methInfo = new YaccaiVal( YaccaiType.IPLAN, retIPlan, methInfo.m_accessType );
								return methInfo;
							case YaccaiType.IDELEGATE:
								YaccaiDelegate yccDel = (YaccaiDelegate)memberInfo.m_contents;
								YaccaiDelegate retDelegate = new YaccaiDelegate( yccDel.m_delegate, baseVal );
								return new YaccaiVal( YaccaiType.IDELEGATE, retDelegate, YaccaiAccessType.Public );
							default:
								RuntimeError( e.Pos, "Invalid function type" );
								break;
						}
					} else {
						return memberInfo;
					}
					break;
				default:
					RuntimeError( e.Pos, "Invalid attribute reference: attribute '" + strFieldName + "'" );
					break;
				
			}
			RuntimeError( e.Pos, "TODO: implement other attribute ref here!" );
			return null;
		}
		private YaccaiVal InvokeStaticMethod( YaccaiVal cls, IntrinsicFunc func, ArrayList args, YEnv env, SrcPos pos )
		{
			if( !func.m_bStatic ) {
				RuntimeError( pos, "Static method required" );
			}
			if( func.m_paramList.Count != args.Count ) {
				RuntimeError( pos, "Method " + func.m_strName +
					" called with different number of arguments: required " +
					func.m_paramList.Count + ", provided " + args.Count );
			}
			YEnv new_env = new YEnv( env.GetTopParent() );
			int i = 0;
			foreach( string strVar in func.m_paramList ) {
				new_env.Bind( strVar, (YaccaiVal)args[i] );
				i++;
			}
			new_env.Bind( "2this", cls );
			m_stkBreakLevel.Push( 0 );
			m_stkContinueLevel.Push( 0 );
			m_callStack.Push( new StackTraceItem( pos, func.m_strName ) );
			m_stkCheckCondition.Push( false );
			YaccaiVal v = EvalStmtList( func.m_body, new_env );
			m_stkCheckCondition.Pop();
			m_stkBreakLevel.Pop();
			m_stkContinueLevel.Pop();
			m_callStack.Pop();
			if( v.m_type == YaccaiType.RETURN ) {
				if( v.m_contents != null ) return (YaccaiVal)v.m_contents;
				else return v;
			}
			return new YaccaiVal( YaccaiType.VOID );
		}
		private YaccaiVal InvokeYaccaiMethod( YaccaiVal obj, string strMethodName, ArrayList args, YEnv env, SrcPos pos )
		{
			
			YObject realObj = (YObject)obj.m_contents;
			YaccaiVal fVal = realObj.MethodLookup( strMethodName );
			if( fVal.m_type == YaccaiType.NOT_FOUND ) {
				RuntimeError( pos, "Method " + strMethodName + " not found in class " + realObj.m_yclass.m_strName );
			}
			if( fVal.m_type == YaccaiType.IFUNC ) {
				IntrinsicFunc iFunc = (IntrinsicFunc)fVal.m_contents;
				if( iFunc.m_paramList.Count != args.Count ) {
					RuntimeError( pos, "Method " + strMethodName +
						" called with different number of arguments: required " +
						iFunc.m_paramList.Count + ", provided " + args.Count );
				}
				YEnv new_env = new YEnv( env.GetTopParent() );
				int i = 0;
				// bind parameters
				foreach( string strVar in iFunc.m_paramList ) {
					new_env.Bind( strVar, (YaccaiVal)args[i] );
					i++;
				}
				// bind "this"
				new_env.Bind( "2this", obj );
				m_stkBreakLevel.Push( 0 );
				m_stkContinueLevel.Push( 0 );
				m_callStack.Push( new StackTraceItem( pos, strMethodName ) );
				m_stkCheckCondition.Push( false );
				YaccaiVal v = EvalStmtList( iFunc.m_body, new_env );
				m_stkCheckCondition.Pop();
				m_stkBreakLevel.Pop();
				m_stkContinueLevel.Pop();
				m_callStack.Pop();
				if( v.m_type == YaccaiType.RETURN ) {
					if( v.m_contents != null ) return (YaccaiVal)v.m_contents;
					else return v;
				}
			} else {
				if( fVal.m_type == YaccaiType.NFUNC ) {
					NativeFunc ntFunc = (NativeFunc)fVal.m_contents;
					ntFunc = new NativeFunc( ntFunc.m_strName, ntFunc.m_nParams, obj );
					ArrayList argsLst = args;
					if( ntFunc.m_nParams != argsLst.Count + 1 ) {
						RuntimeError( pos, "Invalid argument number for " + ntFunc.m_strName );
					}

					YaccaiVal retVal = null;
					argsLst.Insert( 0, ntFunc.m_this );
					m_callStack.Push( new StackTraceItem( pos, ntFunc.m_strName ) );
					m_stkCheckCondition.Push( false );
					retVal = ( new CSharpObject( typeof( YNativeFuncs ), null ) ).InvokeMethod( ntFunc.m_strName, argsLst );
					if( retVal.m_type == YaccaiType.RUNTIME_ERROR ) {
						RuntimeError( pos, (string)retVal.m_contents );
					}
					m_stkCheckCondition.Pop();
					m_callStack.Pop();
					return retVal;
				} else {
					if( fVal.m_type == YaccaiType.IDELEGATE ) {
						YDelFunc yDelFunc = ( (YaccaiDelegate)fVal.m_contents ).m_delegate;
						m_callStack.Push( new StackTraceItem( pos, yDelFunc.ToString() ) );
						m_stkCheckCondition.Push( false );
						args.Insert( 0, obj );
						YaccaiVal retVal = yDelFunc( args, new YEnv( env.GetTopParent() ) );
						if( retVal.m_type == YaccaiType.RUNTIME_ERROR ) {
							RuntimeError( pos, (string)retVal.m_contents );
						}
						m_stkCheckCondition.Pop();
						m_callStack.Pop();
						return retVal;
					}
				}
			}
			return new YaccaiVal( YaccaiType.VOID );
		}
		private YaccaiVal EvalExprAssignment( Expr e, YEnv env )
		{
			ExprAssignment ea = (ExprAssignment)e.Contents;
			YaccaiVal orTestVal = EvalExpr( ea.m_orTestExpr, env );
			if( ea.m_opList == null || ea.m_primaryList == null ) {
				RuntimeError( e.Pos, "Invalid assignment statement" );
			}
			int nCount = ea.m_opList.Count;
			if( ea.m_primaryList.Count != nCount ) {
				RuntimeError( e.Pos, "Invalid assignment statement: opList.size != primary.size" );
			}
			if( nCount == 0 ) return orTestVal;
			int i = nCount - 1;
			YaccaiVal v0 = orTestVal;
			while( i >= 0 ) {
				Expr e1 = ea.m_primaryList[i];
				YaccaiVal v1 = null;
				if( e1.Type == ExprType.Var ) {
					v1 = env.Lookup( (string)e1.Contents );
				} else {
					if( e1.Type == ExprType.AttributeRef )
						v1 = EvalExpr( e1, env );
					else v1 = EvalExpr( ( (ExprSubscript)e1.Contents ).m_base, env );
				}
				
				if( e1.Type == ExprType.Subscript || e1.Type == ExprType.AttributeRef ) {
					if( v1.m_type == YaccaiType.NOT_FOUND ) {
						RuntimeError( e1.Pos, "Member variable or array index not found" );
					}
				} else {
					if( e1.Type != ExprType.Var ) {
						RuntimeError( e1.Pos, "Lvalue required" );
					}
				}
				OperatorType opType = ea.m_opList[i];
				
				switch( opType ) {
					case OperatorType.EQ:
						if( v1.m_type == YaccaiType.NOT_FOUND ) {
							env.Set( (string)e1.Contents, new YaccaiVal( v0.m_type, v0.m_contents ) );
						} else {	
							if( v1.m_type == YaccaiType.IFUNC || v1.m_type == YaccaiType.IPLAN ||
								v1.m_type == YaccaiType.NFUNC ) {
								RuntimeError( e1.Pos, "Invalid Lvalue for assignment" );
							}
							if( e1.Type == ExprType.Subscript && v1.m_type == YaccaiType.NULL ) {
								RuntimeError( e1.Pos, "Null pointer exception" );
							}
							if( v1.m_type == YaccaiType.OBJECT ) {
								YObject realObj = (YObject)v1.m_contents;
								if( e1.Type == ExprType.Subscript ) {
									YaccaiVal indexVal = EvalExpr( ( (ExprSubscript)e1.Contents ).m_index, env );
									ArrayList scArgs = new ArrayList();
									scArgs.Add( indexVal );
									YaccaiVal subLVal = InvokeYaccaiMethod( v1, "operatorSubscriptLvalue", scArgs, env, e1.Pos );
									if( subLVal.m_type == YaccaiType.OBJECT ) {
										realObj = (YObject)subLVal.m_contents;
										if( realObj.MemberLookup( "operatorEq", env ).m_type != YaccaiType.NOT_FOUND ) {
											ArrayList args = new ArrayList();
											args.Add( v0 );
											v0 = InvokeYaccaiMethod( subLVal, "operatorEq", args, env, e1.Pos );
										} else {
											subLVal.m_type = v0.m_type;
											subLVal.m_contents = v0.m_contents;
										}
									} else {
										subLVal.m_type = v0.m_type;
										subLVal.m_contents = v0.m_contents;
									}
								} else {
									if( realObj.MethodLookup( "operatorEq" ).m_type != YaccaiType.NOT_FOUND ) {
										ArrayList args = new ArrayList();
										args.Add( v0 );
										v0 = InvokeYaccaiMethod( v1, "operatorEq", args, env.GetParent(), e1.Pos );
									} else {
										v1.m_contents = v0.m_contents;
										v1.m_type = v0.m_type;
									}
								}
							} else {
								v1.m_type = v0.m_type;
								v1.m_contents = v0.m_contents;
								v0 = v1;
							}
						}
						break;
					case OperatorType.BITWISE_AND_EQ:
						if( v1.m_type == YaccaiType.NOT_FOUND ) {
							RuntimeError( e1.Pos, "Identifier " + (string)e1.Contents + " not found" );
						}
						if( v1.m_type == YaccaiType.INT && v0.m_type == YaccaiType.INT ) {
							int v1Val = (int)v1.m_contents;
							int v0Val = (int)v0.m_contents;
							v1Val &= v0Val;
							v1.m_contents = (Object)v1Val;
							v0 = v1;
						} else {
							RuntimeError( e1.Pos, "Invalid arguments for operator &=" );
						}
						break;
					case OperatorType.PLUS_EQ:
						if( v1.m_type == YaccaiType.NOT_FOUND ) {
							RuntimeError( e1.Pos, "Identifier " + (string)e1.Contents + " not found" );
						}
						ArrayList plusArgList = new ArrayList();
						plusArgList.Clear();
						plusArgList.Add( v1 );
						plusArgList.Add( v0 );
						v0 = EvalOperatorPlus( plusArgList, env, e1.Pos );
						v1.m_contents = v0.m_contents;
						v1.m_type = v0.m_type;
						break;
					case OperatorType.MINUS_EQ:
						if( v1.m_type == YaccaiType.NOT_FOUND ) {
							RuntimeError( e1.Pos, "Identifier " + (string)e1.Contents + " not found" );
						}
						ArrayList minusArgList = new ArrayList();
						minusArgList.Clear();
						minusArgList.Add( v1 );
						minusArgList.Add( v0 );
						v0 = EvalOperatorMinus( minusArgList, env, e1.Pos );
						v1.m_contents = v0.m_contents;
						v1.m_type = v0.m_type;
						break;

					// TODO: add other operators here!
					default:
						RuntimeError( e1.Pos, "Unknown operator type" );
						break;
				}
				i--;
			}
			return v0;
		}
		private YaccaiVal EvalOperatorUminus( ArrayList argVals, YEnv env, SrcPos p )
		{
			if( argVals.Count != 1 ) {
				RuntimeError( p, "Operator - requires 1 argument" );
			}
			YaccaiVal v1 = (YaccaiVal)argVals[0];
			switch( v1.m_type ) {
				case YaccaiType.INT:
					return new YaccaiVal( YaccaiType.INT, (object)( -(int)v1.m_contents ) );
				case YaccaiType.DOUBLE:
					return new YaccaiVal( YaccaiType.DOUBLE, (object)( -(double)v1.m_contents ) );
				case YaccaiType.CHAR:
					RuntimeError( p, "Operator - not supported for CHAR" );
					break;
				case YaccaiType.OBJECT:
					return InvokeYaccaiMethod( v1, "operatorUminus", new ArrayList(), env, p );
				default:
					RuntimeError( p, "Operator - not supported for this type" );
					break;
			}
			return null;
		}
		private YaccaiVal EvalOperatorGt( ArrayList argVals, YEnv env, SrcPos p )
		{
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator > requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			YaccaiVal v1, v2;
			int retInt = 0;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			switch( v1.m_type ) {
				case YaccaiType.INT: {
						int v1Val = (int)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val > v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val > v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator >" );
							}
						}
					}
					break;
				case YaccaiType.DOUBLE: {
						double v1Val = (double)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val > v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val > v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator >" );
							}
						}
					}
					break;
				case YaccaiType.CHAR: {
						char v1Val = (char)v1.m_contents;
						if( v2.m_type != YaccaiType.CHAR ) {
							RuntimeError( p, "Invalid argument for operator >" );
						}
						char v2Val = (char)v2.m_contents;
						if( v1Val > v2Val ) retInt = 1;
					}
					break;
				case YaccaiType.OBJECT:
					newArgVals.Add( v2 );
					return InvokeYaccaiMethod( v1, "operatorGt", newArgVals, env, p );
				default:
					RuntimeError( p, "Operator > not supported for this type" );
					break;
			}
			YaccaiVal retVal = new YaccaiVal( YaccaiType.INT, retInt );
			return retVal;
		}

		private YaccaiVal EvalOperatorGtEq( ArrayList argVals, YEnv env, SrcPos p )
		{
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator >= requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			YaccaiVal v1, v2;
			int retInt = 0;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			switch( v1.m_type ) {
				case YaccaiType.INT: {
						int v1Val = (int)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val >= v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val >= v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator >=" );
							}
						}
					}
					break;
				case YaccaiType.DOUBLE: {
						double v1Val = (double)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val >= v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val >= v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator >=" );
							}
						}
					}
					break;
				case YaccaiType.CHAR: {
						char v1Val = (char)v1.m_contents;
						if( v2.m_type != YaccaiType.CHAR ) {
							RuntimeError( p, "Invalid argument for operator >=" );
						}
						char v2Val = (char)v2.m_contents;
						if( v1Val >= v2Val ) retInt = 1;
					}
					break;
				case YaccaiType.OBJECT: {
						newArgVals.Add( v2 ); 
						YaccaiVal v_Gt = InvokeYaccaiMethod( v1, "operatorGt", newArgVals, env, p );
						YaccaiVal v_Eq = InvokeYaccaiMethod( v1, "operatorEq", newArgVals, env, p );
						if( v_Gt.IsFalseVal() && v_Eq.IsFalseVal() ) retInt = 0;
						else retInt = 1;
					}
					break;
				default:
					RuntimeError( p, "Operator >= not supported for this type" );
					break;
			}
			YaccaiVal retVal = new YaccaiVal( YaccaiType.INT, retInt );
			return retVal;
		}


		private YaccaiVal EvalOperatorLt( ArrayList argVals, YEnv env, SrcPos p )
		{
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator < requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			YaccaiVal v1, v2;
			int retInt = 0;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			switch( v1.m_type ) {
				case YaccaiType.INT: {
						int v1Val = (int)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val < v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val < v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator <" );
							}
						}
					}
					break;
				case YaccaiType.DOUBLE: {
						double v1Val = (double)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val < v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val < v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator <" );
							}
						}
					}
					break;
				case YaccaiType.CHAR: {
						char v1Val = (char)v1.m_contents;
						if( v2.m_type != YaccaiType.CHAR ) {
							RuntimeError( p, "Invalid argument for operator <" );
						}
						char v2Val = (char)v2.m_contents;
						if( v1Val < v2Val ) retInt = 1;
					}
					break;
				case YaccaiType.OBJECT:
					newArgVals.Add( v2 );
					return InvokeYaccaiMethod( v1, "operatorLt", newArgVals, env, p );
				default:
					RuntimeError( p, "Operator < not supported for this type" );
					break;
			}
			YaccaiVal retVal = new YaccaiVal( YaccaiType.INT, retInt );
			return retVal;
		}

		private YaccaiVal EvalOperatorLtEq( ArrayList argVals, YEnv env, SrcPos p )
		{
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator <= requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			YaccaiVal v1, v2;
			int retInt = 0;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			switch( v1.m_type ) {
				case YaccaiType.INT: {
						int v1Val = (int)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val <= v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val <= v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator <= (INT, ?)" );
							}
						}
					}
					break;
				case YaccaiType.DOUBLE: {
						double v1Val = (double)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val <= v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val <= v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator <= (DOUBLE, ?)" );
							}
						}
					}
					break;
				case YaccaiType.CHAR: {
						char v1Val = (char)v1.m_contents;
						if( v2.m_type != YaccaiType.CHAR ) {
							RuntimeError( p, "Invalid argument for operator <= (CHAR, ?)" );
						}
						char v2Val = (char)v2.m_contents;
						if( v1Val <= v2Val ) retInt = 1;
					}
					break;
				case YaccaiType.OBJECT: {
						newArgVals.Add( v2 );
						YaccaiVal v_Lt = InvokeYaccaiMethod( v1, "operatorLt", newArgVals, env, p );
						YaccaiVal v_Eq = InvokeYaccaiMethod( v1, "operatorEq", newArgVals, env, p );
						if( v_Lt.IsFalseVal() && v_Eq.IsFalseVal() ) retInt = 0;
						else retInt = 1;
					}
					break;
				default:
					RuntimeError( p, "Operator <= not supported for this type" );
					break;
			}
			YaccaiVal retVal = new YaccaiVal( YaccaiType.INT, retInt );
			return retVal;
		}

		private YaccaiVal EvalOperatorEqEq( ArrayList argVals, YEnv env, SrcPos p )
		{
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator == requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			YaccaiVal v1, v2;
			int retInt = 0;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			switch( v1.m_type ) {
				case YaccaiType.INT: {
						int v1Val = (int)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val == v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( ((double)v1Val) == v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator ==" );
							}
						}
					}
					break;
				case YaccaiType.DOUBLE: {
						double v1Val = (double)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val == (double)v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val == v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator ==" );
							}
						}
					}
					break;
				case YaccaiType.CHAR: {
						char v1Val = (char)v1.m_contents;
						if( v2.m_type != YaccaiType.CHAR ) {
							RuntimeError( p, "Invalid argument for operator ==" );
						}
						char v2Val = (char)v2.m_contents;
						if( v1Val == v2Val ) retInt = 1;
					}
					break;
				case YaccaiType.NULL:
					if( v2.m_type == YaccaiType.NULL ) retInt = 1;
					break;
				case YaccaiType.OBJECT:
					if( v2.m_type == YaccaiType.NULL ) {
						retInt = 0;
						break;
					}
					if( v1.GetClass().GetMemberEx( "operatorEqEq", env ) != null ) {
						newArgVals.Add( v2 );
						return InvokeYaccaiMethod( v1, "operatorEqEq", newArgVals, env, p );
					} else {
						if( v1.m_contents == v2.m_contents ) retInt = 1;
					}
					break;
				default:
					RuntimeError( p, "Operator == not supported for this type" );
					break;
			}
			YaccaiVal retVal = new YaccaiVal( YaccaiType.INT, retInt );
			return retVal;
		}

		private YaccaiVal EvalOperatorDiff( ArrayList argVals, YEnv env, SrcPos p )
		{
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator != requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			YaccaiVal v1, v2;
			int retInt = 0;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			switch( v1.m_type ) {
				case YaccaiType.INT: {
						int v1Val = (int)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val != v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( ( (double)v1Val ) != v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator !=" );
							}
						}
					}
					break;
				case YaccaiType.DOUBLE: {
						double v1Val = (double)v1.m_contents;
						if( v2.m_type == YaccaiType.INT ) {
							int v2Val = (int)v2.m_contents;
							if( v1Val != (double)v2Val ) retInt = 1;
						} else {
							if( v2.m_type == YaccaiType.DOUBLE ) {
								double v2Val = (double)v2.m_contents;
								if( v1Val != v2Val ) retInt = 1;
							} else {
								RuntimeError( p, "Invalid argument for operator !=" );
							}
						}
					}
					break;
				case YaccaiType.CHAR: {
						char v1Val = (char)v1.m_contents;
						if( v2.m_type != YaccaiType.CHAR ) {
							RuntimeError( p, "Invalid argument for operator !=" );
						}
						char v2Val = (char)v2.m_contents;
						if( v1Val != v2Val ) retInt = 1;
					}
					break;
				case YaccaiType.NULL:
					if( v2.m_type != YaccaiType.NULL ) retInt = 1;
					break;
				case YaccaiType.OBJECT:
					if( v2.m_type == YaccaiType.NULL ) {
						retInt = 0;
						break;
					}
					newArgVals.Add( v2 );
					return InvokeYaccaiMethod( v1, "operatorDiff", newArgVals, env, p );
				default:
					RuntimeError( p, "Operator != not supported for this type" );
					break;
			}
			YaccaiVal retVal = new YaccaiVal( YaccaiType.INT, retInt );
			return retVal;
		}

		private YaccaiVal EvalOperatorPlus( ArrayList argVals, YEnv env, SrcPos p )
		{
			YaccaiVal v1, v2;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator + requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			if( v1.m_type == YaccaiType.INT ) {
				if( v2.m_type == YaccaiType.INT ) {
					return new YaccaiVal( YaccaiType.INT, (object)( (int)v1.m_contents + (int)v2.m_contents ) );
				} else {
					if( v2.m_type == YaccaiType.DOUBLE ) {
						return new YaccaiVal( YaccaiType.DOUBLE, (object)( (int)v1.m_contents + (double)v2.m_contents ) );
					} else {
						if( v2.m_type == YaccaiType.OBJECT ) {
							newArgVals.Add( NativeObjects.YStringObject( ( (int)v1.m_contents ).ToString() ) );
							newArgVals.Add( v2 );
							return EvalOperatorPlus( newArgVals, env, p );
						} else {
							RuntimeError( p, "Invalid argument for operator +" );
						}
					}
				}
			}
			if( v1.m_type == YaccaiType.DOUBLE ) {
				if( v2.m_type == YaccaiType.INT ) {
					return new YaccaiVal( YaccaiType.DOUBLE, (object)( (double)v1.m_contents + (int)v2.m_contents ) );
				} else {
					if( v2.m_type == YaccaiType.DOUBLE ) {
						return new YaccaiVal( YaccaiType.DOUBLE, (object)( (double)v1.m_contents + (double)v2.m_contents ) );
					} else {
						if( v2.m_type == YaccaiType.OBJECT ) {
							newArgVals.Add( NativeObjects.YStringObject( ( (double)v1.m_contents ).ToString() ) );
							return EvalOperatorPlus( newArgVals , env, p );
						} else {
							RuntimeError( p, "Invalid argument for operator +" );
						}
					}
				}
			}
			if( v1.m_type == YaccaiType.CHAR ) {
				if( v2.m_type == YaccaiType.INT ) {
					return new YaccaiVal( YaccaiType.CHAR, (object)( (char)( (int)v1.m_contents + (int)v2.m_contents ) ) );
				} else {
					if( v2.m_type == YaccaiType.CHAR ) {
						return NativeObjects.YStringObject( ( (char)v1.m_contents ).ToString() + ( (char)v2.m_contents ).ToString() );
					} else {
						RuntimeError( p, "Invalid argument for operator +" );
					}
				}
			}
			if( v1.m_type == YaccaiType.OBJECT ) {
				ArrayList argsLst = new ArrayList();
				argsLst.Add( v2 );
				return InvokeYaccaiMethod( v1, "operatorPlus", argsLst, env, p );
			}
			RuntimeError( p, "Invalid argument for operator +" );
			return new YaccaiVal( YaccaiType.VOID, null );
		}

		private YaccaiVal EvalOperatorMinus( ArrayList argVals, YEnv env, SrcPos p )
		{
			YaccaiVal v1, v2;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator - requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			if( v1.m_type == YaccaiType.INT ) {
				if( v2.m_type == YaccaiType.INT ) {
					return new YaccaiVal( YaccaiType.INT, (object)( (int)v1.m_contents - (int)v2.m_contents ) );
				} else {
					if( v2.m_type == YaccaiType.DOUBLE ) {
						return new YaccaiVal( YaccaiType.DOUBLE, (object)( (int)v1.m_contents - (double)v2.m_contents ) );
					} else {
						RuntimeError( p, "Invalid argument for operator -" );
					}
				}
			}
			if( v1.m_type == YaccaiType.DOUBLE ) {
				if( v2.m_type == YaccaiType.INT ) {
					return new YaccaiVal( YaccaiType.DOUBLE, (object)( (double)v1.m_contents - (int)v2.m_contents ) );
				} else {
					if( v2.m_type == YaccaiType.DOUBLE ) {
						return new YaccaiVal( YaccaiType.DOUBLE, (object)( (double)v1.m_contents - (double)v2.m_contents ) );
					} else {
						RuntimeError( p, "Invalid argument for operator -" );
					}
				}
			}
			if( v1.m_type == YaccaiType.CHAR ) {
				if( v2.m_type == YaccaiType.INT ) {
					return new YaccaiVal( YaccaiType.CHAR, (object)( (char)( (int)v1.m_contents - (int)v2.m_contents ) ) );
				} else {
					RuntimeError( p, "Invalid argument for operator -" );
				}
			}
			if( v1.m_type == YaccaiType.OBJECT ) {
				ArrayList argsLst = new ArrayList();
				argsLst.Add( v2 );
				return InvokeYaccaiMethod( v1, "operatorMinus", argsLst, env, p );
			}
			RuntimeError( p, "Invalid argument for operator -" );
			return new YaccaiVal( YaccaiType.VOID, null );
		}

		private YaccaiVal EvalOperatorMul( ArrayList argVals, YEnv env, SrcPos p )
		{
			YaccaiVal v1, v2;
			
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator * requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			if( v1.m_type == YaccaiType.INT ) {
				if( v2.m_type == YaccaiType.INT ) {
					return new YaccaiVal( YaccaiType.INT, (object)( (int)v1.m_contents * (int)v2.m_contents ) );
				} else {
					if( v2.m_type == YaccaiType.DOUBLE ) {
						return new YaccaiVal( YaccaiType.DOUBLE, (object)( (int)v1.m_contents * (double)v2.m_contents ) );
					} else {
						RuntimeError( p, "Invalid argument for operator *" );
					}
				}
			}
			if( v1.m_type == YaccaiType.DOUBLE ) {
				if( v2.m_type == YaccaiType.INT ) {
					return new YaccaiVal( YaccaiType.DOUBLE, (object)( (double)v1.m_contents * (int)v2.m_contents ) );
				} else {
					if( v2.m_type == YaccaiType.DOUBLE ) {
						return new YaccaiVal( YaccaiType.DOUBLE, (object)( (double)v1.m_contents * (double)v2.m_contents ) );
					} else {
						RuntimeError( p, "Invalid argument for operator *" );
					}
				}
			}
			if( v1.m_type == YaccaiType.CHAR ) {
				if( v2.m_type == YaccaiType.INT ) {
					string strRet = "";
					for( int i = 0; i < (int)v2.m_contents; i++ ) strRet += ( (char)v1.m_contents ).ToString();
					return NativeObjects.YStringObject( strRet );
				} else {
					RuntimeError( p, "Invalid argument for operator *" );
				}
			}
			if( v1.m_type == YaccaiType.OBJECT ) {
				ArrayList argsLst = new ArrayList();
				argsLst.Add( v2 );
				return InvokeYaccaiMethod( v1, "operatorMul", argsLst, env, p );
			}
			RuntimeError( p, "Invalid argument for operator *" );
			return new YaccaiVal( YaccaiType.VOID, null );
		}
		private YaccaiVal EvalOperatorDiv( ArrayList argVals, YEnv env, SrcPos p )
		{
			YaccaiVal v1, v2;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator / requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			if( v1.m_type == YaccaiType.INT ) {
				if( v2.m_type == YaccaiType.INT ) {
					if( (int)v2.m_contents == 0 ) {
						RuntimeError( p, "Division by zero" );
					}
					return new YaccaiVal( YaccaiType.INT, (object)( (int)v1.m_contents / (int)v2.m_contents ) );
				} else {
					if( v2.m_type == YaccaiType.DOUBLE ) {
						return new YaccaiVal( YaccaiType.DOUBLE, (object)( (int)v1.m_contents / (double)v2.m_contents ) );
					} else {
						RuntimeError( p, "Invalid argument for operator /" );
					}
				}
			}
			if( v1.m_type == YaccaiType.DOUBLE ) {
				if( v2.m_type == YaccaiType.INT ) {
					if( (int)v2.m_contents == 0 ) RuntimeError( p, "Division by zero" );
					return new YaccaiVal( YaccaiType.DOUBLE, (object)( (double)v1.m_contents / (int)v2.m_contents ) );
				} else {
					if( v2.m_type == YaccaiType.DOUBLE ) {
						if( (double)v2.m_contents == 0.0 ) RuntimeError( p, "Division by zero" );
						return new YaccaiVal( YaccaiType.DOUBLE, (object)( (double)v1.m_contents / (double)v2.m_contents ) );
					} else {
						RuntimeError( p, "Invalid argument for operator /" );
					}
				}
			}
			if( v1.m_type == YaccaiType.OBJECT ) {
				ArrayList argsLst = new ArrayList();
				argsLst.Add( v2 );
				return InvokeYaccaiMethod( v1, "operatorDiv", argsLst, env, p );
			}
			RuntimeError( p, "Invalid argument for operator /" );
			return new YaccaiVal( YaccaiType.VOID, null );
		}

		private YaccaiVal EvalOperatorMod( ArrayList argVals, YEnv env, SrcPos p )
		{
			YaccaiVal v1, v2;
			ArrayList newArgVals = new ArrayList();
			newArgVals.Clear();
			if( argVals.Count != 2 ) {
				RuntimeError( p, "Operator % requires 2 arguments (provided: " + argVals.Count + ")" );
			}
			v1 = (YaccaiVal)argVals[0];
			v2 = (YaccaiVal)argVals[1];
			if( v1.m_type == YaccaiType.INT ) {
				if( v2.m_type == YaccaiType.INT ) {
					if( (int)v2.m_contents == 0 ) {
						RuntimeError( p, "Division by zero" );
					}
					return new YaccaiVal( YaccaiType.INT, (object)( (int)v1.m_contents % (int)v2.m_contents ) );
				}
			}
			RuntimeError( p, "Invalid argument for operator %" );
			return new YaccaiVal( YaccaiType.VOID, null );
		}

		private YaccaiVal EvalOperatorAnd( ArrayList argExpr, YEnv env, SrcPos p )
		{
			Expr e1, e2;
			if( argExpr.Count != 2 ) {
				RuntimeError( p, "Operator && requires 2 arguments (provided: " + argExpr.Count + ")" );
			}
			e1 = (Expr)argExpr[0];
			e2 = (Expr)argExpr[1];
			YaccaiVal v1, v2;
			v1 = EvalExpr( e1, env );
			if( v1.IsFalseVal() ) return new YaccaiVal( YaccaiType.INT, 0 );
			v2 = EvalExpr( e2, env );
			if( v2.IsFalseVal() ) return new YaccaiVal( YaccaiType.INT, 0 );
			return new YaccaiVal( YaccaiType.INT, 1 );
		}

		private YaccaiVal EvalOperatorOr( ArrayList argExpr, YEnv env, SrcPos p )
		{
			Expr e1, e2;
			if( argExpr.Count != 2 ) {
				RuntimeError( p, "Operator || requires 2 arguments (provided: " + argExpr.Count + ")" );
			}
			e1 = (Expr)argExpr[0];
			e2 = (Expr)argExpr[1];
			YaccaiVal v1, v2;
			v1 = EvalExpr( e1, env );
			if( !v1.IsFalseVal() ) return new YaccaiVal( YaccaiType.INT, 1 );
			v2 = EvalExpr( e2, env );
			if( !v2.IsFalseVal() ) return new YaccaiVal( YaccaiType.INT, 1 );
			return new YaccaiVal( YaccaiType.INT, 0 );
		}

		private YaccaiVal EvalOperatorNot( ArrayList argExpr, YEnv env, SrcPos p )
		{
			Expr e1;
			if( argExpr.Count != 1 ) {
				RuntimeError( p, "Operator ! requires 2 arguments (provided: " + argExpr.Count + ")" );
			}
			e1 = (Expr)argExpr[0];
			YaccaiVal v1;
			v1 = EvalExpr( e1, env );
			if( v1.IsFalseVal() ) return new YaccaiVal( YaccaiType.INT, 1 );
			return new YaccaiVal( YaccaiType.INT, 0 );
		}
		
		private YaccaiVal EvalClassDecl( Stmt st, YEnv env )
		{
			ClassDeclStmt cds = (ClassDeclStmt)st.Contents;
			cds.m_inheritanceList.Add( "Object" );
			Hashtable staticVars = new Hashtable();
			Hashtable funcTab = new Hashtable();
			Hashtable memberVarsAccType = new Hashtable();
			StmtList stmList = new StmtList();
			bool bMainFunc = false;
			staticVars.Clear();
			funcTab.Clear();
			stmList.Clear();
			memberVarsAccType.Clear();
			foreach( Stmt stm in cds.m_stmtList ) {
				if( stm.Kind == StmtKind.MethodDeclaration ) {
					YaccaiVal yMeth = EvalMethodDecl( stm, env );
					IntrinsicFunc intrf = (IntrinsicFunc)yMeth.m_contents;
					if( intrf.m_strName.Equals( cds.m_className ) ) {
						int nArgs = intrf.m_paramList.Count;
						string strConstructorName = intrf.m_strName + " " + nArgs;
						funcTab.Add( strConstructorName, new YaccaiVal( YaccaiType.IFUNC, intrf ) );
					} else {
						funcTab.Add( intrf.m_strName, yMeth );
					}
					if( intrf.m_strName.Equals( "Main" ) ) {
						m_mainFunc = intrf;
						bMainFunc = true;
					}
				} else {
					if( stm.Kind == StmtKind.MemberVarDeclaration ) {
						EvalMemberVarDecl( stm, env, staticVars, stmList, memberVarsAccType );
					} else {
						// for test plan execution
						if( stm.Kind == StmtKind.PlanDeclaration ) {
							YaccaiVal planVal = EvalPlanDecl( stm, env );
							IntrinsicPlan iPlan = (IntrinsicPlan)planVal.m_contents;
							funcTab.Add( iPlan.m_strName, planVal );
						} else {
							RuntimeError( stm.Pos, "Invalid statement: only Method declaration or Member var declaration permitted here!" );
						}
					}
				}
			}
			YClass ycls = new YClass( cds.m_className, cds.m_inheritanceList,
										 stmList, staticVars, funcTab, memberVarsAccType, cds.m_accessType );
			ycls.m_strFileName = st.Pos.m_strFileName;
			YaccaiVal v = new YaccaiVal( YaccaiType.CLASS, ycls );
			foreach( YaccaiVal methVal in ycls.m_funcTable.Values ) {
				if( methVal.m_type == YaccaiType.IFUNC ) {
					IntrinsicFunc iFunc = (IntrinsicFunc)methVal.m_contents;
					iFunc.m_this = v;
				}
			}
			YaccaiVal v1 = env.Lookup( cds.m_className );
			if( v1.m_type != YaccaiType.NOT_FOUND ) {
				if( v1.m_type == YaccaiType.CLASS ) {
					YClass v1Cls = (YClass)v1.m_contents;
					if( v1Cls.m_strFileName.Equals( st.Pos.m_strFileName ) ) {
						RuntimeError( st.Pos, "Duplicate class declaration: class name = " + cds.m_className );
					}
					LinkedList<YaccaiVal> retLst = new LinkedList<YaccaiVal>();
					retLst.Clear();
					retLst.Add( v1 );
					retLst.Add( v );
					env.Set( cds.m_className, new YaccaiVal( YaccaiType.CLASS_LIST, retLst ) );
				}
				if( v1.m_type == YaccaiType.CLASS_LIST ) {
					LinkedList<YaccaiVal> clsList = (LinkedList<YaccaiVal>)v1.m_contents;
					if( clsList != null ) {
						clsList.Add( v );
					}
				}
			} else {
				env.Set( cds.m_className, v );
			}
			if( bMainFunc ) m_mainFunc.m_this = v;
			return v;
		}
		private YaccaiVal EvalMethodDecl( Stmt st, YEnv env )
		{
			MethodDeclStmt mds = (MethodDeclStmt)st.Contents;
			IntrinsicFunc intrf = new IntrinsicFunc( mds.m_strName, mds.m_paramList, mds.m_stmtList, mds.m_bStatic, st.Pos );
			YaccaiAccessType yat = mds.m_accessType;
			return new YaccaiVal( YaccaiType.IFUNC, intrf, yat );
		}
		private void EvalMemberVarDecl( Stmt st, YEnv env, Hashtable staticVars, StmtList stmList, Hashtable memberVarsAccType )
		{
			MemberVarDeclStmt mvs = (MemberVarDeclStmt)st.Contents;
			YaccaiVal yv;
			if( mvs.m_bStatic ) {
				foreach( MemberVarExprPair p in mvs.m_pairList ) {
					if( p.m_expr != null ) {
						yv = EvalExpr( p.m_expr, env );
						yv.m_accessType = mvs.m_accessType;
						staticVars.Add( p.m_strName, yv );
					} else {
						staticVars.Add( p.m_strName, new YaccaiVal( YaccaiType.NULL, null, mvs.m_accessType ) );
					}
				}
			} else {
				foreach( MemberVarExprPair p in mvs.m_pairList ) {
					memberVarsAccType.Add( p.m_strName, mvs.m_accessType );
					LinkedList<Expr> priList = new LinkedList<Expr>();
					priList.Clear();
					priList.Add( new Expr( ExprType.Var, p.m_strName, st.Pos ) );
					LinkedList<OperatorType> opLst = new LinkedList<OperatorType>();
					opLst.Clear();
					opLst.Add( OperatorType.EQ );
					if( p.m_expr == null ) p.m_expr = new Expr( ExprType.Var, "null", st.Pos );
					Stmt asStm = new Stmt( new Expr( new ExprAssignment( priList, opLst, p.m_expr ), st.Pos ) );
					stmList.Add( asStm );
				}
			}
		}
		private YaccaiVal EvalPlanDecl( Stmt st, YEnv env )
		{
			PlanDeclStmt pds = (PlanDeclStmt)st.Contents;
			IntrinsicPlan intp = new IntrinsicPlan( pds.m_strName, pds.m_paramList, pds.m_stmtList,
													pds.m_preCond, pds.m_maintainCond, st.Pos );
			YaccaiVal v = new YaccaiVal( YaccaiType.IPLAN, intp, pds.m_accessType );
			return v;
		}
		private YaccaiVal EvalAgentClassDecl( Stmt st, YEnv env )
		{
			AgentClassDeclStmt acd = (AgentClassDeclStmt)st.Contents;
			Hashtable funcTab = new Hashtable();
			//Hashtable planTab = new Hashtable();
			Hashtable memberVarsAccType = new Hashtable();
			StmtList stmList = new StmtList();
			funcTab.Clear();
			//planTab.Clear();
			stmList.Clear();
			memberVarsAccType.Clear();
			acd.m_inheritanceList.Add( "Object" );
			foreach( Stmt stm in acd.m_stmtList ) {
				switch( stm.Kind ) {
					case StmtKind.MethodDeclaration:
						YaccaiVal yMeth = EvalMethodDecl( stm, env );
						IntrinsicFunc intrf = (IntrinsicFunc)yMeth.m_contents;

						if( intrf.m_strName.Equals( acd.m_agentClassName ) ) {
							string strConstructorName = intrf.m_strName;
							strConstructorName += " " + intrf.m_paramList.Count;
							funcTab.Add( strConstructorName, yMeth );
						} else {
							funcTab.Add( intrf.m_strName, yMeth );
						}
						break;
					case StmtKind.PlanDeclaration:
						YaccaiVal pv = EvalPlanDecl( stm, env );
						IntrinsicPlan itrp = (IntrinsicPlan)pv.m_contents;
						funcTab.Add( itrp.m_strName, pv );
						break;
					case StmtKind.MemberVarDeclaration:
						MemberVarDeclStmt mvds = (MemberVarDeclStmt)stm.Contents;
						if( mvds.m_bStatic ) {
							RuntimeError( st.Pos, "Static variable is not allowed for Agent" );
						}
						EvalMemberVarDecl( stm, env, null, stmList, memberVarsAccType );
						break;
					default:
						RuntimeError( stm.Pos, "Invalid statement: only Method, Plan, MemberVar declarations allowed here" );
						break;
				}
			}
			YClass yCls = new YClass( acd.m_agentClassName, acd.m_inheritanceList, stmList, null, funcTab, memberVarsAccType, acd.m_accessType );
			yCls.m_bAgent = true;
			yCls.m_strFileName = st.Pos.m_strFileName;
			YaccaiVal v = new YaccaiVal( YaccaiType.CLASS, yCls, acd.m_accessType );
			YaccaiVal v1 = env.Lookup( acd.m_agentClassName );
			if( v1.m_type != YaccaiType.NOT_FOUND ) {
				if( v1.m_type == YaccaiType.CLASS ) {
					YClass v1Cls = (YClass)v1.m_contents;
					if( v1Cls.m_strFileName.Equals( yCls.m_strFileName ) ) {
						RuntimeError( st.Pos, "Duplicate class declaration: class name = " + acd.m_agentClassName );
					} else {
						LinkedList<YaccaiVal> clsList = new LinkedList<YaccaiVal>();
						clsList.Clear();
						clsList.Add( v1 );
						clsList.Add( v );
						env.Set( acd.m_agentClassName, new YaccaiVal( YaccaiType.CLASS_LIST, clsList, YaccaiAccessType.Public ) );
					}
				}
				if( v1.m_type == YaccaiType.CLASS_LIST ) {
					LinkedList<YaccaiVal> clsList = (LinkedList<YaccaiVal>)v1.m_contents;
					if( clsList != null ) {
						clsList.Add( v );
					}
				}
			} else {
				env.Set( acd.m_agentClassName, v );
			}
			return v;
		}
		private void IncContinueOrBreakLevel( Stack<int> stkLevel )
		{
			if( stkLevel.Count == 0 ) return;
			lock( stkLevel ) {
				int n = stkLevel.Pop();
				n++;
				stkLevel.Push( n );
			}
		}

		private void DecContinueOrBreakLevel( Stack<int> stkLevel )
		{
			if( stkLevel.Count == 0 ) return;
			lock( stkLevel ) {
				int n = stkLevel.Pop();
				if( n > 0 )
					n--;
				stkLevel.Push( n );
			}
		}

		private YaccaiVal EvalStmtList( StmtList stmList, YEnv env )
		{
			YaccaiVal v = null;
			bool bLoop;
			
			foreach( Stmt st in stmList ) {
				bLoop = ( st.Kind == StmtKind.For || st.Kind == StmtKind.ForEach ||
					st.Kind == StmtKind.DoWhile || st.Kind == StmtKind.While );
				if( bLoop ) IncContinueOrBreakLevel( m_stkContinueLevel );
				if( bLoop || st.Kind == StmtKind.Switch ) IncContinueOrBreakLevel( m_stkBreakLevel );
				v = EvalStmt( st, env );
				if( bLoop ) DecContinueOrBreakLevel( m_stkContinueLevel );
				if( bLoop || st.Kind == StmtKind.Switch ) DecContinueOrBreakLevel( m_stkBreakLevel );
				if( v.m_type == YaccaiType.BREAK ) {
					return v;
				}
				if( v.m_type == YaccaiType.CONTINUE ) {
					return v;
				}
				if( v.m_type == YaccaiType.RETURN ) {
					return v;
				}
			}
			return new YaccaiVal( YaccaiType.VOID, null );
		}
		public void ExecuteFileInput( string strFileName )
		{
			StmtList stmList = ParseFileInput( strFileName );
			if( stmList == null ) return;
			YEnv env = new YEnv();
			EvalFileInput( stmList, env );
			EvalFileInput( ParseFileInput( "Array.yac" ), env );
			EvalFileInput( ParseFileInput( "ColOps.yac" ), env );	
			RegisterNativeObjects( env );
			if( m_mainFunc == null ) {
				Console.WriteLine( "Error: Main function not found" );
				Environment.Exit( 1 );
			}
			ArrayList argLst = new ArrayList();
			argLst.Clear();
			int i;
			string str;
			for( i = 0; i < m_strArgs.Count; i++ ) {
				str = (string)m_strArgs[i];
				argLst.Add( new YaccaiVal( YaccaiType.STRING, str ) );
			}
			YaccaiVal yArgsVal = CreateYObject( NativeObjects.YArrayClass().GetClass(), new ArrayList(), env.GetTopParent(), new SrcPos( strFileName, 0, 0 ) );
			foreach( YaccaiVal v in argLst ) {
				ArrayList argLst1 = new ArrayList();
				argLst1.Add( v );
				InvokeYaccaiMethod( yArgsVal, "Add", argLst1, env, new SrcPos( strFileName, 0, 0 ) );
			}

			YEnv localEnv = new YEnv( env );
			localEnv.Bind( "2this", m_mainFunc.m_this );
			if( m_mainFunc.m_paramList != null ) {
				if( m_mainFunc.m_paramList.Count > 1 ) {
					RuntimeError( m_mainFunc.m_pos, "Main function with appropriate arguments not found" );
				}
				if( m_mainFunc.m_paramList.Count == 1 ) {
					localEnv.Set( (string)m_mainFunc.m_paramList[0], yArgsVal );
				}
			}
			
			m_callStack.Push( new StackTraceItem( m_mainFunc.m_pos, m_mainFunc.m_strName ) );
			m_stkBreakLevel.Push( 0 );
			m_stkContinueLevel.Push( 0 );
			m_stkCheckCondition.Push( false );
			EvalStmtList( m_mainFunc.m_body, localEnv );
			m_stkCheckCondition.Pop();
			m_callStack.Pop();
			m_stkBreakLevel.Pop();
			m_stkContinueLevel.Pop();
			bool bAllExited = false;
			while( !bAllExited ) {
				bAllExited = true;
				foreach( Process p in m_arrChildProcesses ) {
					if( !p.HasExited ) {
						bAllExited = false;
						p.WaitForExit();
					}
				}
			}
		}
		public StmtList ParseFileInput( string strFileName )
		{
			FileLookup fl = new FileLookup();
			string strParseFileName = fl.LookupFile( strFileName );
			if( strParseFileName == null ) {
				Console.WriteLine( "File not found: {0}", strFileName );
				return null;
			} else {
				YaccaiLexer lex = new YaccaiLexer( new Antlr.Runtime.ANTLRFileStream( strParseFileName ) );
				CommonTokenStream tks = new CommonTokenStream( lex );
				YaccaiParser p = new YaccaiParser( tks );
				p.StrInputFileName = strParseFileName;
				StmtList lst = p.translation_unit();
				if( p.IsSuccessParse() ) {
					return lst;
				} else {
					Console.WriteLine( "Parse failed" );
					Environment.Exit( 1 );
				}
			}
			return null;
		}
		public void EvalFileInput( StmtList stmList, YEnv env )
		{
			foreach( Stmt st in stmList ) {
				switch( st.Kind ) {
					case StmtKind.ClassDeclaration:
						EvalClassDecl( st, env );
						break;
					case StmtKind.AgentClassDeclaration:
						EvalAgentClassDecl( st, env );
						break;
					case StmtKind.Include:
						EvalStmtInclude( st, env );
						break;
				}
			}
		}
		private void EvalStmtInclude( Stmt st, YEnv env )
		{
			string strFileName = (string)st.Contents;
			strFileName.Remove( 0, 1 );
			strFileName.Remove( strFileName.Length - 1, 1 );
			LinkedList<Stmt> stmList = ParseFileInput( strFileName );
			EvalFileInput( stmList, env );
		}
		private YaccaiVal EvalStmtIf( Stmt st, YEnv env )
		{
			IfStmt ifs = (IfStmt)st.Contents;
			YaccaiVal condVal = EvalExpr( ifs.m_expr, env );
			YaccaiVal retVal = null;
			if( !condVal.IsFalseVal() ) {
				retVal = EvalStmt( ifs.m_trueBranch, env );
			} else {
				if( ifs.m_falseBranch != null )
					retVal = EvalStmt( ifs.m_falseBranch, env );
			}
			if( retVal != null ) {
				if( retVal.m_type == YaccaiType.CONTINUE ) {
					if( m_stkContinueLevel.Peek() > 0 ) return retVal;
				}
				if( retVal.m_type == YaccaiType.BREAK ) {
					if( m_stkBreakLevel.Peek() > 0 ) return retVal;
				}
				if( retVal.m_type == YaccaiType.RETURN ) {
					return retVal;
				}
			}
			return new YaccaiVal( YaccaiType.VOID, null );
		}
		private YaccaiVal EvalStmtFor( Stmt st, YEnv env )
		{
			ForStmt fStm = (ForStmt)st.Contents;
			//IncContinueOrBreakLevel( m_stkBreakLevel );
			//IncContinueOrBreakLevel( m_stkContinueLevel );
			EvalExprYLinkedList( fStm.m_initExpr, env );
			ArrayList condValList = EvalExprYLinkedList( fStm.m_conditionExpr, env );
			bool isFalseCond = false;
			foreach( YaccaiVal v in condValList ) {
				if( v.IsFalseVal() ) {
					isFalseCond = true;
					break;
				}
			}
			while( !isFalseCond ) {
				YaccaiVal stmRetVal = EvalStmt( fStm.m_stmt, env );
				if( stmRetVal.m_type == YaccaiType.RETURN ) {
					//DecContinueOrBreakLevel( m_stkBreakLevel );
					//DecContinueOrBreakLevel( m_stkContinueLevel );
					return stmRetVal;
				}
				if( stmRetVal.m_type == YaccaiType.BREAK ) {
					//DecContinueOrBreakLevel( m_stkBreakLevel );
					//DecContinueOrBreakLevel( m_stkContinueLevel );
					return new YaccaiVal( YaccaiType.VOID, null );
				}
				EvalExprYLinkedList( fStm.m_updateExpr, env );
				condValList = EvalExprYLinkedList( fStm.m_conditionExpr, env );
				isFalseCond = false;
				foreach( YaccaiVal v in condValList ) {
					if( v.IsFalseVal() ) {
						isFalseCond = true;
						break;
					}
				}
			}
			//DecContinueOrBreakLevel( m_stkBreakLevel );
			//DecContinueOrBreakLevel( m_stkContinueLevel );
			return new YaccaiVal( YaccaiType.VOID, null );
		}
		private YaccaiVal EvalStmtWhile( Stmt st, YEnv env )
		{
			WhileStmt wStm = (WhileStmt)st.Contents;
			//IncContinueOrBreakLevel( m_stkBreakLevel );
			//IncContinueOrBreakLevel( m_stkContinueLevel );
			YaccaiVal condVal = EvalExpr( wStm.m_expr, env );
			YaccaiVal retVal = null;
			while( !condVal.IsFalseVal() ) {
				retVal = EvalStmt( wStm.m_stmt, env );
				if( retVal.m_type == YaccaiType.RETURN ) {
					//DecContinueOrBreakLevel( m_stkBreakLevel );
					//DecContinueOrBreakLevel( m_stkContinueLevel );
					return retVal;
				}
				if( retVal.m_type == YaccaiType.BREAK ) {
					//DecContinueOrBreakLevel( m_stkBreakLevel );
					//DecContinueOrBreakLevel( m_stkContinueLevel );
					return new YaccaiVal( YaccaiType.VOID, null );
				}
				condVal = EvalExpr( wStm.m_expr, env );
			}
			//DecContinueOrBreakLevel( m_stkBreakLevel );
			//DecContinueOrBreakLevel( m_stkContinueLevel );
			
			return new YaccaiVal( YaccaiType.VOID, null );
		}

		private YaccaiVal EvalStmtDoWhile( Stmt st, YEnv env )
		{
			DoWhileStmt dws = (DoWhileStmt)st.Contents;
			YaccaiVal condVal = null;
			YaccaiVal retVal = null;
			// do not need new_env because a new env will be created when evaluating block of statement
			do {
				retVal = EvalStmt( dws.m_stmt, env );
				if( retVal.m_type == YaccaiType.RETURN ) return retVal;
				if( retVal.m_type == YaccaiType.BREAK ) return new YaccaiVal( YaccaiType.VOID, null );
				condVal = EvalExpr( dws.m_expr, env );
			} while( !condVal.IsFalseVal() );
			return new YaccaiVal( YaccaiType.VOID, null );
		}

		private YaccaiVal EvalStmtForeach( Stmt st, YEnv env )
		{
			ForEachStmt fest = (ForEachStmt)st.Contents;
			YaccaiVal exprVal = EvalExpr( fest.m_expr, env );
			YaccaiVal yEnumerator = InvokeYaccaiMethod( exprVal, "GetEnumerator", new ArrayList(), env, fest.m_expr.Pos );
			InvokeYaccaiMethod( yEnumerator, "Reset", new ArrayList(), env, fest.m_expr.Pos );
			YaccaiVal bHasNext = InvokeYaccaiMethod( yEnumerator, "MoveNext", new ArrayList(), env, fest.m_expr.Pos );
			// this new env's parent should be env, not env.parent
			YEnv new_env = new YEnv( env );
			new_env.Bind( fest.m_strIdentifier, new YaccaiVal( YaccaiType.NULL, null ) );
			while( !bHasNext.IsFalseVal() ) {
				YaccaiVal curVal = InvokeYaccaiMethod( yEnumerator, "Current", new ArrayList(), env, fest.m_expr.Pos );
				new_env.Set( fest.m_strIdentifier, curVal );
				YaccaiVal retVal = EvalStmt( fest.m_stmt, new_env );
				switch( retVal.m_type ) {
					case YaccaiType.RETURN:
						return retVal;
					case YaccaiType.BREAK:
						return new YaccaiVal( YaccaiType.VOID, null );
					default:
						break;
				}
				bHasNext = InvokeYaccaiMethod( yEnumerator, "MoveNext", new ArrayList(), env, fest.m_expr.Pos );
			}
			return new YaccaiVal( YaccaiType.VOID, null );
		}

		public YaccaiVal EvalStmtSwitch( Stmt st, YEnv env )
		{
			SwitchStmt sstm = (SwitchStmt)st.Contents;
			YaccaiVal exprVal = EvalExpr( sstm.m_expr, env );
			if( exprVal.m_type != YaccaiType.INT &&
				exprVal.m_type != YaccaiType.DOUBLE &&
				exprVal.m_type != YaccaiType.CHAR ) {
				RuntimeError( sstm.m_expr.Pos, "Switch statement does not allow this type of expression" );
			}
			int i = 0;
			Stmt aStm;
			YaccaiVal labelVal;
			ArrayList argsList = new ArrayList();
			if( sstm.m_stmtList.Count == 0 ) return new YaccaiVal( YaccaiType.VOID, null );

			for( i = 0; i < sstm.m_stmtList.Count; i++ ) {
				aStm = sstm.m_stmtList[i];
				if( i == 0 && aStm.Kind != StmtKind.LabelDeclaration ) {
					RuntimeError( aStm.Pos, "Label declaration required for Switch statement" );
				}
				if( aStm.Kind == StmtKind.LabelDeclaration ) {
					LabelDeclStmt lds = (LabelDeclStmt)aStm.Contents;
					argsList.Clear();
					argsList.Add( exprVal );
					if( lds.m_expr == null ) break;
					labelVal = EvalExpr( lds.m_expr, env );
					argsList.Add( labelVal );
					YaccaiVal compareResult = EvalOperatorEqEq( argsList, env, lds.m_expr.Pos );
					if( !compareResult.IsFalseVal() ) break;
				}
			}
			if( i >= sstm.m_stmtList.Count ) return new YaccaiVal( YaccaiType.VOID, null );
			int j;
			StmtList stmList = new StmtList();
			for( j = i + 1; j < sstm.m_stmtList.Count; j++ ) {
				aStm = sstm.m_stmtList[j];
				stmList.Add( aStm );
			}
			YaccaiVal retVal = EvalStmtList( stmList, env );
			if( retVal.m_type == YaccaiType.RETURN ) return retVal;
			if( retVal.m_type == YaccaiType.BREAK ) return new YaccaiVal( YaccaiType.VOID, null );
			return new YaccaiVal( YaccaiType.VOID, null );
		}
		private int GetRegexMaxGroupNum( string strPattern )
		{
			Regex r = new Regex( strPattern );
			int[] arrGNums = r.GetGroupNumbers();
			int nMax = 0;
			foreach( int gNum in arrGNums ) {
				if( gNum > nMax ) nMax = gNum;
			}
			if( Environment.OSVersion.Platform.ToString().Contains( "Unix" ) ) {
				if( nMax > 0 ) {
					return nMax + 1;
				}
			}
			return nMax;
		}
		private Expr ExprFromString( string strExpr )
		{
			YaccaiLexer lex = new YaccaiLexer( new Antlr.Runtime.ANTLRStringStream( strExpr ) );
			CommonTokenStream tks = new CommonTokenStream( lex );
			YaccaiParser yParser = new YaccaiParser( tks );
			yParser.StrInputFileName = "YaccaiEvaluator.cs";
			Expr yExp = null;
			try {
				yExp = yParser.expression();
			} catch( Exception ex ) {
				Console.WriteLine( ex.Message );
				RuntimeError( new SrcPos( "YaccaiEvaluator.cs:ExprFromString", 1, 0 ), "Parse failed" );
			}
			if( yParser.IsSuccessParse() ) {
				if( yExp == null ) {
					RuntimeError( new SrcPos( "YaccaiEvaluator.cs:ExprFromString", 2, 0 ), "Parse failed" );
				}
				return yExp;
			}
			return null;
		}
		// for debug
		private void PrintBinding( string strPattern, Hashtable h, Match aMatch, Hashtable hGroupIdVar )
		{
			Console.WriteLine( "pattern = {0}", strPattern );
			Console.WriteLine( "MaxIndex = {0}", GetRegexMaxGroupNum( strPattern ) );
			Console.WriteLine();
			int i;
			for( i = 0; i < aMatch.Groups.Count; i++ ) {
				Console.WriteLine( "{0} => {1}", i, aMatch.Groups[i].ToString() );
			}
			Console.Write( "\n{" );
			foreach( int j in hGroupIdVar.Keys ) {
				Console.Write( "{0} => {1}, ", j, hGroupIdVar[j] ); 
			}
			Console.WriteLine( "}" );
			Console.Write( "{" );
			foreach( string strKey in h.Keys ) {
				Console.Write( "{0} => {1}, ", strKey, h[strKey] );
			}
			Console.WriteLine( "}" );

		}
		private YaccaiVal EvalStmtMatch( Stmt st, YEnv env )
		{
			MatchStmt mStmt = (MatchStmt)st.Contents;
			//Console.WriteLine( mStmt.ToString() );
			YaccaiVal v = EvalExpr( mStmt.m_subjectExpr, env );
			if( !InstanceOfClass( v, "String", env ) ) {
				RuntimeError( mStmt.m_subjectExpr.Pos, "Invalid argument type for match expression (string required)" );
			}
			string strVal = (string)( (YObject)v.m_contents ).GetMemberEx( "2contents", env );
			if( mStmt.m_targetList.Count <= 0 ) return new YaccaiVal( YaccaiType.VOID, null );
			string strPattern = "";
			string strPartialPattern = "";
			Hashtable hGroupIndexVar = new Hashtable();
			YaccaiMatchPair mPair;
			int i, j;
			int nGroupNum = 1;

			for( i = 0; i < mStmt.m_targetList.Count; i++ ) {
				mPair = mStmt.m_targetList[i];
				LinkedList<Expr> mExprList = mPair.m_exprList;
				strPattern = "^";
				int nIndex = 0;
				hGroupIndexVar.Clear();
				YEnv new_env = new YEnv( env );
				bool bMatched = true;
				nGroupNum = 1;
				Expr ePat;
				for( j = 0; j < mExprList.Count; j++ ) {
					ePat = mExprList[j];
					ExprRegexExpr rEPat = (ExprRegexExpr)ePat.Contents;
					YaccaiVal rEPatVal = null;
					if( rEPat.m_expr != null ) {
						bool bConstant = true;
						if( rEPat.m_expr.Type == ExprType.StopEvalExpr ) {
							bConstant = false;
							Expr exprToBind = (Expr)rEPat.m_expr.Contents;
							if( exprToBind == null || exprToBind.Type != ExprType.Var ) {
								RuntimeError( rEPat.m_expr.Pos, "Invalid expression for 'match'" );
							}
							string strBindVarName = (string)( (Expr)rEPat.m_expr.Contents ).Contents;
							rEPatVal = new_env.Lookup( strBindVarName );
							if( rEPatVal.m_type == YaccaiType.NOT_FOUND ) {
								rEPatVal = NativeObjects.YStringObject( @"\w" );
								//Console.WriteLine( "----> " + rEPat.m_expr.Contents.ToString() );
								new_env.Bind( rEPat.m_expr.Contents.ToString(), rEPatVal );
								hGroupIndexVar.Add( nGroupNum, strBindVarName );
							} else {
								string strVarName = strBindVarName;
								if( new_env.GetBindings().ContainsKey( "? regex " + strVarName ) ) {
									rEPatVal = NativeObjects.YStringObject( (string)new_env.GetBindings()["? regex " + strVarName] );
									bConstant = true;
								} else {
									//hGroupIndexVar.Add( GetRegexMaxGroupNum( strPattern ) + 1, strBindVarName );
									hGroupIndexVar.Add( nGroupNum, strBindVarName );
								}
							}
						} else {
							if( rEPat.m_expr.Type == ExprType.Var ) {
								if( new_env.GetBindings().ContainsKey( "? regex " + (string)rEPat.m_expr.Contents ) ) {
									rEPatVal = NativeObjects.YStringObject( (string)new_env.GetBindings()["? regex " + (string)rEPat.m_expr.Contents] );
								} else rEPatVal = EvalExpr( rEPat.m_expr, env );
							} else {
								rEPatVal = EvalExpr( rEPat.m_expr, env );
							}
						}

						switch( rEPatVal.m_type ) {
							case YaccaiType.INT:
								if( bConstant ) strPartialPattern = "(" + Regex.Escape( ( (int)rEPatVal.m_contents ).ToString() ) + "\\s*)";
								else strPartialPattern = @"((\+|-)?\d+\s*)";
								break;
							case YaccaiType.DOUBLE:
								if( bConstant ) strPartialPattern = "(" + Regex.Escape( ( (double)rEPatVal.m_contents ).ToString() ) + "\\s*)";
								else strPartialPattern = @"((\+|-)?\d*\.\d*\s*)";
								break;
							case YaccaiType.CHAR:
								if( bConstant ) strPartialPattern = "(" + Regex.Escape( ( (char)rEPatVal.m_contents ).ToString() ) + "\\s*)";
								else strPartialPattern = @"(\S\s*)";
								break;
							case YaccaiType.OBJECT:
								if( !InstanceOfClass( rEPatVal, "String", env ) ) {
									RuntimeError( rEPat.m_expr.Pos, "Only String is allowed for matching" );
								}
								string strPatVal = (string)( (YObject)rEPatVal.m_contents ).GetMemberEx( "2contents", env );
								if( bConstant ) {
									strPartialPattern = "(" + Regex.Escape( strPatVal ) + @"\s*)";
								} else strPartialPattern = @"(\w+\s*)";
								break;
							default:
								RuntimeError( rEPat.m_expr.Pos, "Invalid argument type for Match" );
								break;
						}
					} else strPattern += ".";
					string strSpace = "";
					if( nIndex > 0 ) strSpace = "";
					switch( rEPat.m_opType ) {
						case RegexOpType.Star:
							strPattern += strSpace + strPartialPattern + "*";
							break;
						case RegexOpType.Question:
							strPattern += strSpace + strPartialPattern + "?";
							break;
						case RegexOpType.Plus:
							strPattern += strSpace + strPartialPattern + "+";
							break;
						default:
							strPattern += strSpace + strPartialPattern;
							break;
					}
					nIndex++;
					
					//Console.WriteLine( "Matching: pattern = {0}, val = {1}", strVal, strPattern );
					if( !Regex.IsMatch( strVal, strPattern ) ) {
						//Console.WriteLine( "Match failed: pattern = {0}, val = {1}", strVal, strPattern );
						bMatched = false;
						break;
					}
					Match aMatch = Regex.Match( strVal, strPattern );
					nGroupNum = aMatch.Groups.Count;
					foreach( int gIndex in hGroupIndexVar.Keys ) {
						string strVarName = (string)hGroupIndexVar[gIndex];
						string strValue = aMatch.Groups[gIndex].ToString();
						strValue = strValue.TrimEnd();
						strValue = strValue.TrimStart();
						if( !new_env.GetBindings().ContainsKey( "? regex " + strVarName ) ) {
							new_env.GetBindings().Add( "? regex " + strVarName, strValue );
						}
					}
					//PrintBinding( strPattern, new_env.GetBindings(), aMatch, hGroupIndexVar );
				}
				if( !bMatched ) continue;
				strPattern += "$";
				//Console.WriteLine( "Matching: pattern = {0}, val = {1}", strVal, strPattern );
				if( Regex.IsMatch( strVal, strPattern ) ) {
					foreach( int gIndex in hGroupIndexVar.Keys ) {
						string strVarName = (string)hGroupIndexVar[gIndex];
						string strValue = (string)new_env.GetBindings()["? regex " + strVarName];
						YaccaiVal v_to_bind = new_env.Lookup( strVarName );
						switch( v_to_bind.m_type ) {
							case YaccaiType.INT:
								//Console.WriteLine( "Int32.Parse( {0} ), strVarName = {1},\n pattern = {2}", strValue, strVarName, strPattern );
								try {
									v_to_bind.m_contents = Int32.Parse( strValue );
								} catch( Exception ex ) {
									Console.WriteLine( ex.Message );
									Console.WriteLine( "strValue = {0}, strVal = {1}", strValue, strVal );
									Console.WriteLine( "expr = {0}", mStmt.m_subjectExpr );
									Console.WriteLine( mStmt );
									Environment.Exit( 1 );
								}
								break;
							case YaccaiType.DOUBLE:
								v_to_bind.m_contents = Double.Parse( strValue );
								break;
							case YaccaiType.CHAR:
								v_to_bind.m_contents = Char.Parse( strValue );
								break;
							case YaccaiType.OBJECT:
								if( InstanceOfClass( v, "String", new_env ) ) {
									if( InstanceOfClass( v_to_bind, "String", new_env ) ) {
										( (YObject)v_to_bind.m_contents ).m_instanceVars["2contents"] = strValue;
									} else {
										v_to_bind.m_contents = NativeObjects.YStringObject( strValue );
									}
								}
								break;
							default:
								break;
						}
					}
					foreach( string strVarName in hGroupIndexVar.Values ) {
						string strRegexVarName = "? regex " + strVarName;
						if( new_env.GetBindings().ContainsKey( strRegexVarName ) ) {
							new_env.GetBindings().Remove( strRegexVarName );
						}
					}
					YaccaiVal retVal = EvalStmt( mPair.m_stmt, new_env );
					return retVal;
				} 
			}
			return new YaccaiVal( YaccaiType.VOID, null );
		}

		public YaccaiVal EvalStmt( Stmt st, YEnv env )
		{
			m_bIdleWaitForMsg = false;
			bool bCheckCondition = false;

			if( IsYaccaiMainThread() && !m_bMainThreadExecutionTurn ) {
				//if( !m_bMainThreadExecutionTurn ) {
					//Console.WriteLine( "Agent{0}[{1}]: wait to execute stmt", m_nMyProcId, Thread.CurrentThread.ManagedThreadId );
					m_mutexExecutionTurn.WaitOne();
					m_bMainThreadExecutionTurn = true;
				//}
				DeliverMessage( env );
			}
			
			if( m_stkCheckCondition.Count > 0 ) bCheckCondition = m_stkCheckCondition.Peek();
			if( bCheckCondition ) {
				/*
				if( env.LookupClass( "Array" ).m_type == YaccaiType.NOT_FOUND ) {
					Console.WriteLine( "Error: Array not found??" );
					env.PrintBindings();
					Console.WriteLine();
					Console.WriteLine( st );
					RuntimeError( new SrcPos( "YEvaluator.cs", 0, 0 ), "Array not found????" );
				}
				*/
				

				//Console.WriteLine( st );
				ExprEnvPair eepair = m_stkMaintainCond.Peek();
				Expr condExp = eepair.m_expr;
				if( condExp != null ) {
					//Console.WriteLine( "Press enter ... " );
					//Console.ReadLine();
					YaccaiVal condVal = EvalExpr( condExp, eepair.m_env );
					if( condVal.IsFalseVal() ) {
						if( IsYaccaiMainThread() ) {
							m_bMainThreadExecutionTurn = false;
							m_mutexExecutionTurn.ReleaseMutex();
						}
						return new YaccaiVal( YaccaiType.RETURN, new YaccaiVal( YaccaiType.INT, 0 ) );
					}
					
				}
			}
			YaccaiVal retVal = null;
			switch( st.Kind ) {
				case StmtKind.Break:
					if( m_stkBreakLevel.Peek() > 0 ) {
						retVal = new YaccaiVal( YaccaiType.BREAK, null, YaccaiAccessType.Public );
					} else {
						Console.WriteLine( "Agent{0}: Error while executing {1}", m_nMyProcId, st );
						Console.WriteLine( "Agent{2}: stkBreakLevel.Peed() = {0}, count = {1}, evalRemote = {3}", m_stkBreakLevel.Peek(), m_stkBreakLevel.Count, m_nMyProcId, m_bEvalRemote );
						/*
						Stack<int>.Enumerator en = m_stkBreakLevel.GetEnumerator();
						while( en.MoveNext() ) {
							Console.Write( en.Current + " " );
						}
						Console.WriteLine( "]" );
						*/
						RuntimeError( st.Pos, "Break statement not allowed here" );
					}
					break;
				case StmtKind.Continue:
					if( m_stkContinueLevel.Peek() > 0 ) {
						retVal = new YaccaiVal( YaccaiType.CONTINUE, null, YaccaiAccessType.Public );
					} else {
						RuntimeError( st.Pos, "Continue statement not allowed here" );
					}
					break;
				case StmtKind.Block:
					retVal = EvalStmtList( (StmtList)st.Contents, new YEnv( env ) );
					break;
				case StmtKind.Expression:
					retVal = EvalExpr( (Expr)st.Contents, env );
					break;
				case StmtKind.Return:
					if( st.Contents != null ) {
						Expr retExpr = (Expr)st.Contents;
						if( retExpr == null ) {
							retVal = new YaccaiVal( YaccaiType.RETURN, null );
						} else {
							YaccaiVal retVal1 = EvalExpr( retExpr, env );
							retVal = new YaccaiVal( YaccaiType.RETURN, retVal1 );
						}
					} else retVal = new YaccaiVal( YaccaiType.RETURN, null );
					break;
				case StmtKind.If:
					retVal = EvalStmtIf( st, env );
					break;
				case StmtKind.For:
					retVal = EvalStmtFor( st, env );
					break;
				case StmtKind.ForEach:
					retVal = EvalStmtForeach( st, env );
					break;
				case StmtKind.While:
					retVal = EvalStmtWhile( st, env );
					break;
				case StmtKind.DoWhile:
					retVal = EvalStmtDoWhile( st, env );
					break;
				case StmtKind.Switch:
					retVal = EvalStmtSwitch( st, env );
					break;
				case StmtKind.Match:
					retVal = EvalStmtMatch( st, env );
					break;
				case StmtKind.LabelDeclaration:
					retVal = new YaccaiVal( YaccaiType.VOID, null );
					break;
				default:
					RuntimeError( st.Pos, "Invalid statement" );
					break;
			}
			if( IsYaccaiMainThread() && m_bMainThreadExecutionTurn ) {
				m_bMainThreadExecutionTurn = false;
				m_mutexExecutionTurn.ReleaseMutex();
			}
			return retVal;
		}
		public YaccaiVal EvalExprCreate( Expr e, YEnv env )
		{
			ExprCreate ec = (ExprCreate)e.Contents;
			ArrayList argVals = EvalExprYLinkedList( ec.m_exprList, env );
			
			if( argVals.Count < 1 ) {
				RuntimeError( e.Pos, "Agent ID required" );
			}
			string strHostname = null;
			if( ec.m_exprHostname != null ) {
				YaccaiVal hostVal = EvalExpr( ec.m_exprHostname, env );
				YObject yObj = (YObject)hostVal.m_contents;
				strHostname = (string)yObj.GetMemberEx( "2contents", env );
			} else strHostname = "localhost";
			/*
			if( strHostname.Contains( "@" ) ) {
				string[] arrHostAddr = strHostname.Split( new char[] { '@' } );
				strHostname = arrHostAddr[1];
			}
			 * */
			string strPath = null;
			if( strHostname.Contains( ":" ) ) {
				string[] arrHostPath = strHostname.Split( new char[] { ':' } );
				strHostname = arrHostPath[0];
				strPath = arrHostPath[1];
			}
			YaccaiVal idVal = (YaccaiVal)argVals[0];
			argVals.RemoveAt( 0 );
			if( idVal.m_type != YaccaiType.INT ) {
				RuntimeError( e.Pos, "Agent ID must be integer" );
			}
			int nAgentId = (int)idVal.m_contents;

			YaccaiVal agentVal = env.LookupClass( ec.m_agentClassName );
			if( agentVal.m_type == YaccaiType.NOT_FOUND ) {
				RuntimeError( e.Pos, "Agent class " + ec.m_agentClassName + " not found" );
			}
			YClass agentCls = (YClass)agentVal.m_contents;
			FileLookup fl = new FileLookup();
			string strFullPath = Path.GetFullPath( fl.LookupFile( agentCls.m_strFileName ) );
			if( strPath == null ) strPath = strFullPath;
			else {
				strPath += "" + Path.DirectorySeparatorChar + Path.GetFileName( strFullPath );
			}
			ArrayList strArgVals = new ArrayList();
			foreach( YaccaiVal v in argVals ) {
				strArgVals.Add( v.Marshall().ToString() );
			}
			int nRank = nAgentId;
			/*
			TcpClient tClient = new TcpClient( m_strPortMgrAddr, m_nPortMgrPort );
			NetworkStream ns = tClient.GetStream();
			StreamWriter sw = new StreamWriter( ns );
			StreamReader sr = new StreamReader( ns );
			if( nAgentId < 0 ) {
				sw.WriteLine( "YaccaiPort 200 -1 " + strHostname );
			} else {
				sw.WriteLine( "YaccaiPort 210 " + nAgentId + " " + strHostname );
			}
			sw.Flush();
			
			string retStr = sr.ReadLine();
			string[] arrRet;
			if( retStr.Contains( "YaccaiPort 200" ) ) {
				arrRet = Regex.Split( retStr, @"\s+" );
				if( arrRet.Length < 3 ) {
					Console.WriteLine( "Invalid reply from PortMgr" );
					sw.WriteLine( "YaccaiPort 10" );
					sw.Flush();
					sw.Close();
					tClient.Close();
					return new YaccaiVal( YaccaiType.INT, -1 );
				}
				nRank = Int32.Parse( arrRet[2] );
			} else {
				if( retStr.Contains( "YaccaiPort 210" ) ) {
					nRank = nAgentId;
				} else {
					if( retStr.Contains( "YaccaiPort 211" ) ) {
						sw.WriteLine( "YaccaiPort 10" );
						sw.Flush();
						tClient.Close();
						RuntimeError( e.Pos, "Agent ID " + nAgentId + " already existed" );
					}
				}
			}
			sw.WriteLine( "YaccaiPort 10" );
			
			sw.Flush();
			tClient.Close();
			*/ 
			Process p = new Process();
			p.StartInfo.UseShellExecute = false;
			if( Environment.OSVersion.ToString().Contains( "Unix" ) ) {
				p.StartInfo.FileName = "ssh";
				p.StartInfo.Arguments = strHostname + " ";
			} else {
				p.StartInfo.FileName = "YIOConnector.exe";
				p.StartInfo.Arguments = strHostname + " " + m_nIOConnectorPort + " ";
			}
			p.StartInfo.Arguments += "yaccai.exe ";
			p.StartInfo.Arguments += "-rank " + nRank + " ";
			p.StartInfo.Arguments += "-pma " + m_strPortMgrAddr + " ";
			p.StartInfo.Arguments += "-pmp " + m_nPortMgrPort + " ";
			p.StartInfo.Arguments += "-yiop " + m_nIOConnectorPort + " "; 
			p.StartInfo.Arguments += "-m " + ec.m_agentClassName + " ";
			p.StartInfo.Arguments += strPath;
			foreach( string strArg in strArgVals ) {
				p.StartInfo.Arguments += " " + strArg;
			}
			//Console.WriteLine( p.StartInfo.FileName + " " +  p.StartInfo.Arguments );
			m_arrChildProcesses.Add( p );
			p.Start();
			return new YaccaiVal( YaccaiType.INT, p.Id );
		}

		public void ExecuteAgent( string strFileName, string strAgentClsName, ArrayList args )
		{
			// TODO: fix this method
			LinkedList<Stmt> stmList = ParseFileInput( strFileName );
			m_stkCheckCondition.Push( false );
			if( stmList == null ) return;
			YEnv env = new YEnv();
			EvalFileInput( stmList, env );
			EvalFileInput( ParseFileInput( "Array.yac" ), env );
			EvalFileInput( ParseFileInput( "ColOps.yac"), env );
			RegisterNativeObjects( env );
			ArrayList parentObj = new ArrayList();
			parentObj.Add( NativeObjects.YObject() );
			YaccaiVal clsVal = env.LookupClass( strAgentClsName );
			if( clsVal.m_type == YaccaiType.NOT_FOUND ) {
				Console.WriteLine( "Error: class " + strAgentClsName + " not found" );
				Environment.Exit( 1 );
			}
			m_bAgent = true;
			AgentInit();
			YClass yCls = (YClass)clsVal.m_contents;
			ArrayList real_args = new ArrayList();
			real_args.Add( new YaccaiVal( YaccaiType.INT, m_nMyProcId ) );
			int i = 0;
			for( i = 0; i < args.Count; i++ ) {
				real_args.Add( NativeObjects.YStringObject( (string)args[i] ) );
			}
			YaccaiVal agentVal = CreateYObject( yCls, real_args, env, new SrcPos( "ExecuteAgent:CreateYObject", 0, 0 ) );
			env.Bind( "2this", agentVal );
			m_gYEnv = env;
			m_gYEnvReadyEvent.Set();
			YaccaiVal actPlan = env.Lookup( "act" );
			if( actPlan.m_type == YaccaiType.NOT_FOUND ) {
				Console.WriteLine( "Error: act plan not found" );
				Environment.Exit( 1 );
			}
			IntrinsicPlan iPlan = (IntrinsicPlan)actPlan.m_contents ;
			EvalYPlan( actPlan, new ArrayList(), env, iPlan.m_pos );

			while( true ) {
				m_bIdleWaitForMsg = true;
				DeliverMessage( env );
			}
			// m_stkCheckCondition.Pop();
			//InvokeYaccaiMethod( agentVal, "act", new ArrayList(), env, new SrcPos( "Native", 0, 0 ) );
		}

	}

}
