using System;
using System.Collections;
using System.Reflection;
namespace Yaccai
{
	using StmtList = LinkedList<Stmt>;
	enum YaccaiType { INT, DOUBLE, CHAR, STRING, CONTINUE, BREAK, NOT_FOUND, VOID, NULL, OBJECT,
						NFUNC, IFUNC, CLASS, ARRAY, AGENT, CSOBJ, CLASS_LIST, IPLAN, RETURN, CS_METHOD,
						RUNTIME_ERROR, IDELEGATE };
	interface IYMemberSet
	{
		YaccaiVal MemberLookup( string strName, YEnv env );
		object GetMemberEx( string strName, YEnv env );
	}
	class YClass	:	IYMemberSet
	{
		public string m_strName;	// class name
		public string m_strFileName;
		public YaccaiAccessType m_accessType;
		public ArrayList m_parentList;
		public LinkedList<Stmt> m_stmtList;
		public Hashtable m_staticVars;
		public Hashtable m_memberVarsAccessType;
		public Hashtable m_funcTable;
		public bool m_bAgent;
		public YClass( string clsName, ArrayList parentList, LinkedList<Stmt> stmList, 
						Hashtable staticVars, Hashtable funcTab, Hashtable memberVarAccType, YaccaiAccessType yct )
		{
			m_strName = clsName;
			m_parentList = parentList;
			m_stmtList = stmList;
			m_funcTable = funcTab;
			m_staticVars = staticVars;
			m_accessType = yct;
			m_strFileName = "{Yaccai Native Class}";
			m_memberVarsAccessType = memberVarAccType;
			m_bAgent = false;
		}
		public YaccaiVal MemberLookup( string strName, YEnv env )
		{
			if( m_staticVars != null ) {
				if( m_staticVars.ContainsKey( strName ) ) return (YaccaiVal)m_staticVars[strName];
			}
			if( m_funcTable.ContainsKey( strName ) ) return (YaccaiVal)m_funcTable[strName];
			foreach( string className in m_parentList ) {
				YaccaiVal yV = env.LookupClass( className );
				if( yV.m_type != YaccaiType.NOT_FOUND ) {
					IYMemberSet yMbs = (IYMemberSet)yV.m_contents;
					YaccaiVal v = yMbs.MemberLookup( strName, env );
					if( v.m_type != YaccaiType.NOT_FOUND ) return v;
				}
			}
			return new YaccaiVal( YaccaiType.NOT_FOUND, null );
		}
		public object GetMemberEx( string strName, YEnv env )
		{
			if( m_staticVars.ContainsKey( strName ) ) return (object)m_staticVars[strName];
			if( m_funcTable.ContainsKey( strName ) ) return (object)m_funcTable[strName];
			foreach( string className in m_parentList ) {
				YaccaiVal yV = env.Lookup( className );
				if( yV.m_type != YaccaiType.NOT_FOUND ) {
					IYMemberSet yMbs = (IYMemberSet)yV.m_contents;
					object v = yMbs.GetMemberEx( strName, env );
					if( v != null ) return v;
				}
			}
			return null;
		}
		public bool IsSubClassOf( string strName, YEnv env )
		{
			if( m_strName.Equals( strName ) ) return true;
			foreach( string strClassName in m_parentList ) {
				YaccaiVal yV = env.LookupClass( strClassName );
				if( yV.m_type != YaccaiType.NOT_FOUND ) {
					if( yV.m_type == YaccaiType.CLASS ) {
						YClass yCls = (YClass)yV.m_contents;
						if( yCls.IsSubClassOf( strName, env ) ) return true;
					}
				}
			}
			return false;
		}
	}

	class YObject : IYMemberSet
	{
		public YClass m_yclass;
		public Hashtable m_instanceVars;
		public ArrayList m_parentObjs;
		public YObject( YClass cls, Hashtable instanceVars, ArrayList parentObjs )
		{
			m_yclass = cls;
			m_instanceVars = instanceVars;
			m_parentObjs = parentObjs;
		}
		public YaccaiVal MemberLookup( string strName, YEnv env )
		{
			if( m_instanceVars.Contains( strName ) ) {
				return (YaccaiVal)m_instanceVars[strName];
			}
			if( !m_yclass.m_bAgent ) {
				if( m_yclass.m_staticVars.ContainsKey( strName ) ) {
					return (YaccaiVal)m_yclass.m_staticVars[strName];
				}
			}
			if( m_yclass.m_funcTable.Contains( strName ) ) {
				return (YaccaiVal)m_yclass.m_funcTable[strName];
			}
			YaccaiVal yV = null;
			foreach( YaccaiVal yObjVal in m_parentObjs ) {
				YObject yObj = (YObject)yObjVal.m_contents;
				yV = yObj.MemberLookup( strName, null );
				if( yV.m_type != YaccaiType.NOT_FOUND )
					return yV;
			}
			return new YaccaiVal( YaccaiType.NOT_FOUND );
		}
		public object GetMemberEx( string strName, YEnv env )
		{
			if( m_instanceVars.Contains( strName ) ) {
				return m_instanceVars[strName];
			}
			if( !m_yclass.m_bAgent ) {
				if( m_yclass.m_staticVars.ContainsKey( strName ) ) {
					return m_yclass.m_staticVars[strName];
				}
			}
			if( m_yclass.m_funcTable.Contains( strName ) ) {
				return m_yclass.m_funcTable[strName];
			}
			object yV = null;
			foreach( YaccaiVal yObjVal in m_parentObjs ) {
				YObject yObj = (YObject)yObjVal.m_contents;
				yV = yObj.GetMemberEx( strName, null );
				if( yV != null )
					return yV;
			}
			return null;
		}

		public YaccaiVal MethodLookup( string strName )
		{
			if( m_yclass.m_funcTable.Contains( strName ) )
				return (YaccaiVal)m_yclass.m_funcTable[strName];
			YaccaiVal yV = null;
			foreach( YaccaiVal yVal in m_parentObjs ) {
				YObject yObj = (YObject)yVal.m_contents;
				yV = yObj.MethodLookup( strName );
				if( yV.m_type != YaccaiType.NOT_FOUND )
					return yV;
			}
			return new YaccaiVal( YaccaiType.NOT_FOUND, null, YaccaiAccessType.Public );
		}
	}
	class YArray
	{
		public ArrayList m_arrData;
		public YArray()
		{
			m_arrData = new ArrayList();
			m_arrData.Clear();
		}
		public YArray( YaccaiVal[] arrVal )
		{
			int i = 0;
			if( arrVal != null ) {
				for( i = 0; i < arrVal.Length; i++ )
					m_arrData.Add( arrVal[i] );
			}
		}
		public YArray( ArrayList lst )
		{
			m_arrData = lst;
		}
	}
	class IntrinsicFunc
	{
		public string m_strName;
		public ArrayList m_paramList;
		public StmtList m_body;
		public SrcPos m_pos;
		public YaccaiVal m_this;
		public bool m_bStatic;
		public IntrinsicFunc( string strName, ArrayList paramList, StmtList body, SrcPos p )
		{
			m_strName = strName;
			m_paramList = paramList;
			m_body = body;
			m_pos = p;
			m_bStatic = false;
			m_this = null;
		}
		public IntrinsicFunc( string strName, ArrayList paramList, StmtList body, bool bStatic, SrcPos p )
		{
			m_strName = strName;
			m_paramList = paramList;
			m_body = body;
			m_pos = p;
			m_bStatic = bStatic;
			m_this = null;
		}
	}

	class IntrinsicPlan : IntrinsicFunc
	{
		public Expr m_exprPreCond;
		public Expr m_exprMaintainCond;
		
		public IntrinsicPlan( string strName, ArrayList paramList, StmtList body, Expr preCond,
								Expr maintainCond, SrcPos p ) : 
			base( strName, paramList, body, p )
		{
			m_exprMaintainCond = maintainCond;
			m_exprPreCond = preCond;
			m_this = null;
		}
	}

	class YaccaiDelegate
	{
		public YEvaluator.YDelFunc m_delegate;
		public YaccaiVal m_this;
		public YaccaiDelegate( YEvaluator.YDelFunc f, YaccaiVal thisVal )
		{
			m_delegate = f;
			m_this = thisVal;
		}
		public YaccaiDelegate( YEvaluator.YDelFunc f )
		{
			m_delegate = f;
			m_this = null;
		}
	}
	class NativeFunc
	{
		public string m_strName;
		public int m_nParams;
		public YaccaiVal m_this;
		public NativeFunc( string strName, int nParams, YaccaiVal thisVal )
		{
			m_strName = strName;
			m_nParams = nParams;
			m_this = thisVal;
		}
	}

	class CSharpObject
	{
		public Type m_type;
		public object m_contents;
		public CSharpObject( Type t, object c )
		{
			m_type = t;
			m_contents = c;
		}
		public MemberInfo GetMember( string strName )
		{
			MemberInfo[] arrTypes = m_type.GetMember( strName );
			if( arrTypes == null || arrTypes.Length == 0 ) return null;
			return arrTypes[0];
		}
		public YaccaiVal InvokeMethod( string strMethodName, ArrayList arrArg )
		{
			ArrayList realArgs = new ArrayList();
			realArgs.Clear();
			foreach( YaccaiVal yVal in arrArg ) {
				realArgs.Add( yVal.Marshall() );
			}
			Type[] arrTypes = new Type[realArgs.Count];
			object[] argToPass = null;
			int i = 0;
			if( m_type.FullName.Equals( "System.Console" ) ) {
				if( realArgs.Count <= 1 ) {
					foreach( object obj in realArgs ) {
						arrTypes[i] = obj.GetType();
						i++;
					}
					argToPass = realArgs.ToArray();
				} else {
					argToPass = new object[2];
					arrTypes = new Type[2];
					object[] tmpArr = new object[realArgs.Count - 1];
					arrTypes[0] = "".GetType();
					arrTypes[1] = tmpArr.GetType();
					argToPass[0] = realArgs[0];
					for( i = 1; i < realArgs.Count; i++ ) {
						tmpArr[i - 1] = realArgs[i];
					}
					argToPass[1] = tmpArr;
				}
			} else {
				i = 0;
				foreach( object obj in realArgs ) {
					arrTypes[i] = obj.GetType();
					i++;
				}
				argToPass = realArgs.ToArray();
			}
			MethodInfo methInfo = m_type.GetMethod( strMethodName, arrTypes );
			if( methInfo.ReturnType.FullName.Equals( "System.Void" ) ) {
				m_type.InvokeMember( strMethodName, BindingFlags.Default | BindingFlags.InvokeMethod,
									null, m_contents, argToPass );
				return new YaccaiVal( YaccaiType.VOID, null );
			}
			object retObj = m_type.InvokeMember( strMethodName, BindingFlags.Default | BindingFlags.InvokeMethod,
												 null, m_contents, argToPass );
			return YaccaiVal.UnMarshall( retObj );
		}
	}
	class CSharpMethod
	{
		public string m_strMethodName;
		public CSharpObject m_csobj;
		public CSharpMethod( string name, CSharpObject obj )
		{
			m_strMethodName = name;
			m_csobj = obj;
		}
	}

	class NativeObjects
	{
		private static YaccaiVal m_yObjectClass = null;
		private static YaccaiVal m_yStringClass = null;
		private static YaccaiVal m_yArrayClass = null;
		private static YaccaiVal m_yEnvironmentClass = null;
		private static YaccaiVal m_yCommunicatorClass = null;
		private static YaccaiVal m_yMsgClass = null;
		private static YaccaiVal m_yPeerClass = null;
		public static YEvaluator m_yEvaluator = null;
	
		public NativeObjects()
		{
		}
		public static YaccaiVal YObjectClass()
		{
			// return Object class
			if( m_yObjectClass != null ) return m_yObjectClass;
			YClass yc = new YClass( "Object", new ArrayList(), new StmtList(), new Hashtable(), new Hashtable(), new Hashtable(), YaccaiAccessType.Public );
			yc.m_funcTable.Clear();
			
			yc.m_parentList.Clear();
			yc.m_stmtList.Clear();
			yc.m_staticVars.Clear();
			YaccaiVal v = new YaccaiVal( YaccaiType.CLASS, yc, YaccaiAccessType.Public );
			m_yObjectClass = v;
			yc.m_funcTable.Add( "Object 0", new YaccaiVal( YaccaiType.NFUNC, new NativeFunc( "ObjectObject_0", 1, v ) ) );
			yc.m_funcTable.Add( "operatorEqEq", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.NtObjectOperatorEqEq ), YaccaiAccessType.Public ) );
			yc.m_funcTable.Add( "operatorDiff", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.NtObjectOperatorDiff ), YaccaiAccessType.Public ) );
			return m_yObjectClass;
		}
		public static YaccaiVal YObject()
		{
			// return an instance of Object class
			
			Hashtable insVars = new Hashtable();
			ArrayList parentOjbs = new ArrayList();
			parentOjbs.Clear();
			insVars.Clear();
			YaccaiVal v = new YaccaiVal( YaccaiType.OBJECT, new YObject( (YClass)(YObjectClass().m_contents), insVars, parentOjbs ) );
			return v;
		}

		public static YaccaiVal YStringClass()
		{
			// return String class
			if( m_yStringClass != null ) return m_yStringClass;
			ArrayList parentList = new ArrayList();
			parentList.Clear();
			parentList.Add( "Object" );
			Hashtable funcTab = new Hashtable();
			funcTab.Add( "operatorPlus", new YaccaiVal( YaccaiType.NFUNC, new NativeFunc( "StringOperatorPlus", 2, null ) ) );
			funcTab.Add( "Equals", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.Nt_String_Equals ) ) );
			funcTab.Add( "Split", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.Nt_String_Split ) ) );
			funcTab.Add( "ToInt", new YaccaiVal( YaccaiType.NFUNC, new NativeFunc( "StringToInt", 1, null ) ) );
			funcTab.Add( "ToDouble", new YaccaiVal( YaccaiType.NFUNC, new NativeFunc( "StringToDouble", 1, null ) ) );
			// TODO: register String's constructors, methods here!
			YClass yCls = new YClass( "String", parentList, new StmtList(), new Hashtable(), funcTab, new Hashtable(), YaccaiAccessType.Public );
			m_yStringClass = new YaccaiVal( YaccaiType.CLASS, yCls, YaccaiAccessType.Public );
			return m_yStringClass;
		}
		public static YaccaiVal YStringObject( string str )
		{
			// return an instance of String class
			Hashtable insVars = new Hashtable();
			ArrayList parentOjbs = new ArrayList();
			parentOjbs.Clear();
			insVars.Clear();
			insVars.Add( "2contents", str );
			parentOjbs.Add( YObject() );
			YaccaiVal v = new YaccaiVal( YaccaiType.OBJECT, new YObject( (YClass)( YStringClass().m_contents ), insVars, parentOjbs ) );
			return v;
		}
		public static YaccaiVal YArrayClass()
		{
			// return Array class
			if( m_yArrayClass != null ) return m_yArrayClass;
			ArrayList parentList = new ArrayList();
			parentList.Clear();
			parentList.Add( "Object" );
			Hashtable funcTab = new Hashtable();
			Hashtable memberVarsAccType = new Hashtable();
			memberVarsAccType.Add( "2contents", null );
			memberVarsAccType.Add( "Length", YaccaiAccessType.Public );
			funcTab.Add( "operatorSubscript", new YaccaiVal( YaccaiType.NFUNC, new NativeFunc( "ArrayOperatorSubscript", 2, null ) ) );
			funcTab.Add( "operatorSubscriptLvalue", new YaccaiVal( YaccaiType.NFUNC, new NativeFunc( "ArrayOperatorSubscriptLvalue", 2, null ) ) );
			funcTab.Add( "Array 0", new YaccaiVal( YaccaiType.NFUNC, new NativeFunc( "ArrayArray_0", 1, null ) ) );
			funcTab.Add( "Array 1", new YaccaiVal( YaccaiType.NFUNC, new NativeFunc( "ArrayArray_1", 2, null ) ) );
			funcTab.Add( "Add", new YaccaiVal( YaccaiType.NFUNC, new NativeFunc( "ArrayAdd", 2, null ) ) );
			ArrayList getEnParamLst = new ArrayList();
			getEnParamLst.Clear();
			LinkedList<Expr> opNewParamList = new LinkedList<Expr>();
			opNewParamList.Add( new Expr( ExprType.Var, "this", new SrcPos( "YaccaiValue.cs (native)", 306, 100 ) ) );
			LinkedList<Stmt> getEnStmtList = new LinkedList<Stmt>();
			getEnStmtList.Add( new Stmt( StmtKind.Return, new SrcPos( "native", 0, 0 ) , new Expr( new ExprObjectAllocation( "ArrayEnumerator", opNewParamList, new LinkedList<Expr>() ), new SrcPos( "native", 0, 0 ) ) ) );
			funcTab.Add( "GetEnumerator", new YaccaiVal( YaccaiType.IFUNC, new IntrinsicFunc( "GetEnumerator", 
															getEnParamLst, getEnStmtList, new SrcPos( "native", -1, -1 ) ) ) );
			// TODO: register Array's constructors, methods here!
			YClass yCls = new YClass( "Array", parentList, new StmtList(), new Hashtable(), funcTab, new Hashtable(), YaccaiAccessType.Public );
			m_yArrayClass = new YaccaiVal( YaccaiType.CLASS, yCls, YaccaiAccessType.Public );
			return m_yArrayClass;
		}
		public static YaccaiVal YEnviromentClass()
		{
			if( m_yEnvironmentClass != null ) return m_yEnvironmentClass;
			ArrayList parentList = new ArrayList();
			parentList.Clear();
			parentList.Add( "Object" );
			Hashtable funcTab = new Hashtable();
			Hashtable memberVarsAccType = new Hashtable();
			funcTab.Add( "GetCommunicator", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.GetCommunicator ), YaccaiAccessType.Public ) );
			funcTab.Add( "GetRank", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.NtGetRank ), YaccaiAccessType.Public ) );
			funcTab.Add( "Exit", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.NtEnvironment_Exit ), YaccaiAccessType.Public ) );
			funcTab.Add( "Sleep", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.NtSleep ), YaccaiAccessType.Public ) );
			funcTab.Add( "EvalExpr", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.NtEvalExpr ), YaccaiAccessType.Public ) );
			YClass yCls = new YClass( "Environment", parentList, new StmtList(), new Hashtable(), funcTab, new Hashtable(), YaccaiAccessType.Public );
			m_yEnvironmentClass = new YaccaiVal( YaccaiType.CLASS, yCls, YaccaiAccessType.Public );
			return m_yEnvironmentClass;		
		}
		public static YaccaiVal YCommunicatorClass()
		{
			if( m_yCommunicatorClass != null ) return m_yCommunicatorClass;
			ArrayList parentList = new ArrayList();
			parentList.Clear();
			parentList.Add( "Object" );
			Hashtable funcTab = new Hashtable();
			funcTab.Add( "Send", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.Comm_Send ), YaccaiAccessType.Public ) );
			funcTab.Add( "Bcast", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.Comm_Bcast ), YaccaiAccessType.Public ) );
			funcTab.Add( "Reduce", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.Comm_Reduce ), YaccaiAccessType.Public ) );
			
			Hashtable memberVarsAccType = new Hashtable();
			memberVarsAccType.Add( "Name", YaccaiAccessType.Public );
			memberVarsAccType.Add( "Listener", YaccaiAccessType.Public );
			YClass yCls = new YClass( "Communicator", parentList, new StmtList(), new Hashtable(), funcTab, new Hashtable(), YaccaiAccessType.Public );
			m_yCommunicatorClass = new YaccaiVal( YaccaiType.CLASS, yCls, YaccaiAccessType.Public );
			return m_yCommunicatorClass;
		}
		public static YaccaiVal YMsgClass()
		{
			if( m_yMsgClass != null ) return m_yMsgClass;
			ArrayList parentList = new ArrayList();
			parentList.Clear();
			parentList.Add( "Object" );
			Hashtable funcTab = new Hashtable();
			Hashtable memberVarsAccType = new Hashtable();
			memberVarsAccType.Add( "Value", YaccaiAccessType.Public );
			memberVarsAccType.Add( "Src", YaccaiAccessType.Public );
			memberVarsAccType.Add( "Dst", YaccaiAccessType.Public );
			memberVarsAccType.Add( "Communicator", YaccaiAccessType.Public );
			YClass yCls = new YClass( "YMessage", parentList, new StmtList(), new Hashtable(), funcTab, new Hashtable(), YaccaiAccessType.Public );
			m_yMsgClass = new YaccaiVal( YaccaiType.CLASS, yCls, YaccaiAccessType.Public );
			return m_yMsgClass;
		}
		public static YaccaiVal YMsgObject( string strVal, int nSrc, int nDst, string strCommName )
		{
			// return an instance of YMsgClass
			Hashtable insVars = new Hashtable();
			ArrayList parentOjbs = new ArrayList();
			parentOjbs.Clear();
			insVars.Clear();
			insVars.Add( "Value", YStringObject( strVal ) );
			insVars.Add( "Src", new YaccaiVal( YaccaiType.INT, nSrc ) );
			insVars.Add( "Dst", new YaccaiVal( YaccaiType.INT, nDst ) );
			insVars.Add( "Communicator", YStringObject( strCommName ) );
			parentOjbs.Add( YObject() );
			YaccaiVal v = new YaccaiVal( YaccaiType.OBJECT, new YObject( (YClass)( YMsgClass().m_contents ), insVars, parentOjbs ) );
			return v;
		}

		public static YaccaiVal YPeerClass()
		{
			if( m_yPeerClass != null ) return m_yPeerClass;
			ArrayList parentList = new ArrayList();
			parentList.Clear();
			parentList.Add( "Object" );
			Hashtable funcTab = new Hashtable();
			Hashtable memberVarsAccType = new Hashtable();
			memberVarsAccType.Add( "LocalHost", YaccaiAccessType.Public );
			memberVarsAccType.Add( "LocalPort", YaccaiAccessType.Public );
			memberVarsAccType.Add( "Host", YaccaiAccessType.Public );
			memberVarsAccType.Add( "Port", YaccaiAccessType.Public );
			memberVarsAccType.Add( "2Sock", YaccaiAccessType.Public );
			funcTab.Add( "Peer 2", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.Nt_Peer_Peer ), YaccaiAccessType.Public ) );
			funcTab.Add( "Send", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.Nt_Peer_Send ), YaccaiAccessType.Public ) );
			funcTab.Add( "Recv", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.Nt_Peer_Recv ), YaccaiAccessType.Public ) );
			funcTab.Add( "Close", new YaccaiVal( YaccaiType.IDELEGATE, new YaccaiDelegate( m_yEvaluator.Nt_Peer_Close ), YaccaiAccessType.Public ) );

			YClass yCls = new YClass( "Peer", parentList, new StmtList(), new Hashtable(), funcTab, new Hashtable(), YaccaiAccessType.Public );
			m_yPeerClass = new YaccaiVal( YaccaiType.CLASS, yCls, YaccaiAccessType.Public );
			return m_yPeerClass;
		}

		public static YaccaiVal YPeerObject( string strRemoteHost, int nRemotePort, string strLocalHost, int nLocalPort )
		{
			// return an instance of YPeerClass
			Hashtable insVars = new Hashtable();
			ArrayList parentOjbs = new ArrayList();
			parentOjbs.Clear();
			insVars.Clear();
			insVars.Add( "Host", YStringObject( strRemoteHost ) );
			insVars.Add( "Port", new YaccaiVal( YaccaiType.INT, nRemotePort ) );
			insVars.Add( "LocalHost", new YaccaiVal( YaccaiType.INT, nLocalPort ) );
			insVars.Add( "LocalPort", YStringObject( strLocalHost ) );
			parentOjbs.Add( YObject() );
			YaccaiVal v = new YaccaiVal( YaccaiType.OBJECT, new YObject( (YClass)( YMsgClass().m_contents ), insVars, parentOjbs ) );
			return v;
		}
	}
	class YaccaiVal
	{
		public Object m_contents;
		public YaccaiType m_type;
		public YaccaiAccessType m_accessType;
		public YaccaiVal( YaccaiType t )
		{
			m_type = t;
			m_contents = null;
			m_accessType = YaccaiAccessType.Private;
		}
		public YaccaiVal( YaccaiType t, Object contents )
		{
			m_contents = contents;
			m_type = t;
			m_accessType = YaccaiAccessType.Public;
		}
		public YaccaiVal( YaccaiType t, Object contents, YaccaiAccessType yt )
		{
			m_contents = contents;
			m_type = t;
			m_accessType = yt;
		}
		public object Marshall()
		{
			if( m_type != YaccaiType.OBJECT ) {
				return m_contents;
			}
			YObject obj = (YObject)m_contents;
			string strClassName = obj.m_yclass.m_strName;
			switch( strClassName ) {
				case "Object":
					return obj;
				case "String":
					return ( (string)obj.m_instanceVars["2contents"] );
				default:
					return obj;
			}
		}
		public static YaccaiVal UnMarshall( object obj )
		{
			string strName = obj.GetType().FullName;
			YaccaiType retType = YaccaiType.NULL;
			switch( strName ) {
				case "System.Int32":
					retType = YaccaiType.INT;
					break;
				case "System.Double":
					retType = YaccaiType.DOUBLE;
					break;
				case "System.String":
					return NativeObjects.YStringObject( (string)obj );
				case "System.Char":
					retType = YaccaiType.CHAR;
					break;
				case "Yaccai.YaccaiVal":
					return (YaccaiVal)obj;
				case "Yaccai.YObject":
					retType = YaccaiType.OBJECT;
					break;
				default:
					if( !strName.Contains( "Yaccai.CSharpObject" ) )
						obj = new CSharpObject( obj.GetType(), obj );
					retType = YaccaiType.CSOBJ;
					break;
			}
			return new YaccaiVal( retType, obj );
		}
		public YClass GetClass()
		{
			switch( this.m_type ) {
				case YaccaiType.OBJECT:
					YObject yObj = (YObject)this.m_contents;
					return yObj.m_yclass;
				case YaccaiType.CLASS:
				case YaccaiType.AGENT:
					return (YClass)this.m_contents;
				default:
					return null;
				
			}
		}
		public bool IsFalseVal()
		{
			switch( m_type ) {
				case YaccaiType.INT:
					return ( ( (int)m_contents ) == 0 );
				case YaccaiType.DOUBLE:
					return ( ( (double)m_contents ) == 0.0 );
				case YaccaiType.NULL:
					return true;
				case YaccaiType.OBJECT:
					return false;
				/// TODO
				default:
					return false;
			}
		}
		public override string ToString()
		{
			switch( m_type ) {
				case YaccaiType.CHAR:
					return ( (char)m_contents ).ToString();
				case YaccaiType.INT:
					return ( (int)m_contents ).ToString();
				case YaccaiType.DOUBLE:
					return ( (double)m_contents ).ToString();
				case YaccaiType.NULL:
					return "null";
				default:
					return "YaccaiVal";
			}
		}
	}
}
